/*  Given a list of W-X pairs, we want to select the Xs in random
    order.  w_member(X, WXs) picks an X from WXs according to its
    weight.  If backtracked into, it selects from the remaining
    elements.
*/
/*
w_member(X, WXs) :-             % pick X randomly from WXs
    WXs = [W-_|WX1],            % fail if WXs is empty
    w_sum(WX1, W, Sum),         % add up all the weights
    w_select(X, WXs, Sum).
*/

r_member(X, Xs) :-              % Random member
    length(Xs, N),
    N > 0,
    random(Big), R is Big mod N,
    r_find(Xs, R, X).

r_find([X|Xs], R, A) :-
    (   R =:= 0 -> A = X
    ;   R1 is R - 1,
        r_find(Xs, R1, A)
    ).



w_member(X, WXs) :-             % Weighted random member
    WXs = [W-Z|WX1],
    (   WX1 = [] -> X = Z
    ;   w_sum(WX1, W, Sum),
        random(Big), R is Big mod Sum,
        w_find(WXs, R, X)
    ).

w_find([W-X|WXs], R, A) :-
    (   R < W -> A = X
    ;   R1 is R - W,
        w_find(WXs, R1, A)
    ).


w_sum([], Sum, Sum).
w_sum([W-_|WXs], Sum0, Sum) :-
    Sum1 is Sum0 + W,
    w_sum(WXs, Sum1, Sum).

/*
w_select(X, WXs, Sum) :-
    (    / * If WXs has only one element * /
        WXs = [_ - Z]
    ->   / * then we must select it * /
        X = Z
    ;    / * Otherwise we generate a random number 0 <= R < Sum * /
        random(Big), R is Big mod Sum,
         / * let WXs = B <> [W-Z] <> A * /
         / * where 0 <= sum-weights(B) <= R < W + sum-weights(B) * /
         / * let WXr = B <> A * /
        w_sel(W, Z, WXs, WXr, R),
        (    / * First try Z as the answer * /
            X = Z
        ;    / * and if that fails, try again with WXr for WXs * /
            Sum1 is Sum - W,
            w_select(X, WXr, Sum1)
        )
    ).

w_sel(W, X, [WX|WXs], WXr, R) :-
    WX = V-Z,
    (   V > R -> W = V, X = Z, WXr = WXs
    ; / * V =< R * /  R1 is R - V, WXr = [WX|WXn], w_sel(W, X, WXs, WXn, R1)
    ).
*/

/*  <X>_head(Clause Number, Dictionary, Semantics, Limit, Prefix).
    <X>_rule(Clause Number, Dictionary, Semantics, Limit0, Limit, S0, S).
*/

e(X, L) :-
    e([], X, L, _, L, []).


e(D, X, N0, N, S0, S) :-
    findall(W-K, e_head(K, W, D, X, N0, S0), Ks),
    w_member(K, Ks),
    e_rule(K, D, X, N0, N, S0, S).

e_head(1, 2, _, let(I,_,_), [_,_,_,_,_|_], [w(let),w(I),w(be)|_]).
e_head(2, 1, _, if(_,_,_,_,_), [_,_,_,_,_,_|_], [w(if)|_]).
e_head(3, 20, _, _, _, [T|_]) :- e_filt(T).

e_filt(X) :- var(X), !.
e_filt(+).
e_filt(-).
e_filt(w(_)).
e_filt(n(_)).
e_filt('(').

e_rule(1, D, let(I,V,E), [_,_,_,_,_|N0], N) -->
        [w(let), w(I), w(be)], {id(I)},
        e(D, V, N0, N1),
        [w(in)],
        e([I|D], E, N1, N).
e_rule(2, D, if(A,R,B,T,E), [_,_,_,_,_,_,_|N0], N) -->
        [w(if)],
        s(D, A, N0, N1),
        [r(R)], {rel(R)},
        s(D, B, N1, N2),
        [w(then)],
        e(D, T, N2, N3),
        [w(else)],
        e(D, E, N3, N).
e_rule(3, D, E, N0, N) -->
        s(D, E, N0, N).

s(D, X, N0, N, S0, S) :-
    findall(W-K, s_head(K, W, D, X, N0, S0), Ks),
    w_member(K, Ks),
    s_rule(K, D, X, N0, N, S0, S).

s_head(1, 3, _, +(_,_), [_,_|_], [T|_]  ) :- s_filt(T).
s_head(2, 3, _, -(_,_), [_,_|_], [T|_]  ) :- s_filt(T).
s_head(3, 1, _, _,      [_|_],   [+,T|_]) :- s_filt(T).
s_head(4, 1, _, _,      [_|_],   [-,T|_]) :- s_filt(T).
s_head(5,10, _, _,      _,       [T|_]  ) :- s_filt(T).

s_filt(X) :- var(X), !.
s_filt(w(_)).
s_filt(n(_)).
s_filt('(').

s_rule(1, D, +(A,B), [_,_|N0], N) -->
        s(D, A, N0, N1),
        [+],
        t(D, B, N1, N).
s_rule(2, D, -(A,B), [_,_|N0], N) -->
        s(D, A, N0, N1),
        [-],
        t(D, B, N1, N).
s_rule(3, D, B, [_|N0], N) -->
        [+],
        t(D, B, N0, N).
s_rule(4, D, -(n(0),B), [_|N0], N) -->
        [-],
        t(D, B, N0, N).
s_rule(5, D, B, N0, N) -->
        t(D, B, N0, N).

t(D, X, N0, N, S0, S) :-
    findall(W-K, t_head(K, W, D, X, N0, S0), Ks),
    w_member(K, Ks),
    t_rule(K, D, X, N0, N, S0, S).

t_head(1, 3, _, *(_,_), [_,_|_], [T|_]) :- s_filt(T).
t_head(2, 1, _, /(_,_), [_,_|_], [T|_]) :- s_filt(T).
t_head(3,10, _, _,      _,       [T|_]) :- s_filt(T).

t_rule(1, D, *(A,B), [_,_|N0], N) -->
        t(D, A, N0, N1),
        [*],
        f(D, B, N1, N).
t_rule(2, D, /(A,B), [_,_|N0], N) -->
        t(D, A, N0, N1),
        [/],
        f(D, B, N1, N).
t_rule(3, D, B, N0, N) -->
        f(D, B, N0, N).

f(D, X, N0, N, S0, S) :-
    findall(W-K, f_head(K, W, D, X, N0, S0), Ks),
    w_member(K, Ks),
    f_rule(K, D, X, N0, N, S0, S).

f_head(1, 1, _, ^(_,_), [_,_|_], [T|_]) :- s_filt(T).
f_head(2, 9, _, _,      _,       [T|_]) :- s_filt(T).

f_rule(1, D, ^(A,B), [_,_|N0], N) -->
        p(D, A, N0, N1),
        [^],
        p(D, B, N1, N).
f_rule(2, D, A, N0, N) -->
        p(D, A, N0, N).

p(D, X, N0, N, S0, S) :-
    findall(W-K, p_head(K, W, D, X, N0, S0), Ks),
    w_member(K, Ks),
    p_rule(K, D, X, N0, N, S0, S).

p_head(1, 5, _, n(N), _, [n(N)|_]).
p_head(2, 5, D, i(I), _, [w(I)|_]) :- nonvar(D), D = [_|_].
p_head(3, 1, _, _,    [_,_|_], ['(',T|_]) :- p_filt(T).

p_filt(X) :- var(X), !.
p_filt(w(let)).
p_filt(w(if)).
p_filt(+).
p_filt(-).
p_filt(w(_)).
p_filt(n(_)).
p_filt('(').

p_rule(1, _, n(N), N0, N0) -->
        [n(N)], {num(N)}.
p_rule(2, D, i(I), N0, N0) -->
        [w(I)], {id(D, I)}.
p_rule(3, D, E, [_,_|N0], N) -->
        ['('],
        e(D, E, N0, N),
        [')'].



/* parsing:
num(N) :- number(N), N >= 0.
id(I)  :- atom(I).
rel(<).
rel(>).
rel(=).
rel(>=).
rel(=<).
rel(/=).
*/
num(N) :-
    (   var(N) ->
        r_member(N, [0,1,2,5,10])
    ;   integer(N),
        N >= 0
    ).

rel(R) :-
    (   var(R) ->
        r_member(R, [<,>,=,>=,=<,/=])
    ;   atom(R),
        member(R, [<,>,=,>=,=<,/=]),
        !
    ).

id(I) :-
    (   var(I) ->
        r_member(I, [a,b,c])
    ;   atom(I)
    ).

id(D, I) :-
    (   var(I) ->
        r_member(I, D)
    ;   member(I, D), 
        !
    ).


pr([]) :-
    nl.
pr([X|Xs]) :-
    put(32),
    pr1(X),
    pr(Xs).

pr1(w(X)) :- !, write(X).
pr1(n(X)) :- !, write(X).
pr1(r(X)) :- !, write(X).
pr1(X)    :-    write(X).

go :-
    ( e(_, L) -> put(46)
      % pr(L) 
    ; pr([failed]) ),
    go.

rgo :-
    repeat,
    (  e(_, L) -> put(46)
      % pr(L) 
      ; pr([failed])
    ),
    fail.

