%   File   : STDIO.PL
%   Author : K. De Bosschere
%   Created: March 22, 1994
%   Purpose: Read and Write Standard Prolog Text in BinProlog

% -------------- initialisation of user_input and user_output -----

is_prolog(binprolog_with_cparser).

% by defining this, macro-expansion is not performed on put_code and
% friends

default_input :- 
   open(user_input, read, Stdin), 
   saved(Stdin, S2),
   (def(alias, user_input, S2); set(alias, user_input, S2)), !,
   (def(current_stream, read, S2); set(current_stream, read, S2)), !.
default_output :- 
   open(user_output, write, Stdout), 
   saved(Stdout, S2),
   (def(alias, user_output, S2); set(alias, user_output, S2)), !,
   (def(current_stream, write, S2); set(current_stream, write, S2)), !.

init_io :- 
  default_input, default_output.

% ------------------------------------------------------------------

current_input(Stream) :-
  val(current_stream,read,Stream).

current_output(Stream) :-
  val(current_stream,write,Stream).

set_input(Stream_or_Alias) :-
  alias(Stream_or_Alias,Stream),
  saved(Stream,Saved_Stream), 
  set(current_stream,read,Saved_Stream).

set_output(Stream_or_Alias) :-
  alias(Stream_or_Alias,Stream),
  saved(Stream,Saved_Stream), 
  set(current_stream,write,Saved_Stream).

open(File, Mode, Stream) :-
  open(File, Mode, Stream, []).

open(File, Mode, S, Options) :-
  S = '$stream'(Stream,File,Mode,Options),
  io_options(Mode,Mode1),
  iso_open_stream(Mode1,File,Stream),
  open_options(Options,S).

open_options([],_).
open_options([type(text)|L],S) :- open_options(L,S).
open_options([type(binary)|L],S) :- open_options(L,S).
open_options([reposition(true)|L],S) :- open_options(L,S).
open_options([reposition(false)|L],S) :- open_options(L,S).
open_options([eof_action(error)|L],S) :- open_options(L,S).
open_options([eof_action(eof_code)|L],S) :- open_options(L,S).
open_options([eof_action(reset)|L],S) :- open_options(L,S).
open_options([alias(A)|L],S) :- saved(S,S2),      %--- create aliases
                                (def(alias, A, S2); set(alias,A,S2)), !,
                                open_options(L,S).

close(Stream_or_Alias) :-
  close(Stream_or_Alias,[]).

close(Stream_or_Alias, _) :-
  alias(Stream_or_Alias,'$stream'(Stream,_File,Mode,_)),
  iso_close_stream(Mode,Stream).

flush_output :-
  current_output(Stream),
  flush_output(Stream).

flush_output(Stream_or_Alias) :-
  alias(Stream_or_Alias, _Stream),
  true. /* iso_write automatically flushes all buffers */

stream_property('$stream'(_,File,_,_),file_name(File)).
stream_property('$stream'(_,_,Mode,_),mode(Mode)).
stream_property('$stream'(_,_,read,_),input).
stream_property('$stream'(_,_,write,_),output).
stream_property('$stream'(_,_,_,Options),alias(A)) :- 
   member(alias(A),Options).
stream_property('$stream'(Stream,_,_,_Options),position(P)) :- 
   iso_lseek(Stream,0,1,P).
stream_property(Stream,end_of_stream(at))   :- at_end_of_stream(Stream), !.
stream_property(Stream,end_of_stream(past)) :- at_end_of_stream(Stream), !.
stream_property(_Stream,end_of_stream(not))  :- !.           % default
stream_property('$stream'(_,_,_,Options),eof_action(A)) :-
   member(eof_action(A),Options) -> true ; A = eof_code.
stream_property('$stream'(_,_,_,Options),reposition(B)) :- 
   member(reposition(B),Options).
stream_property('$stream'(_,_,_,Options),type(T)) :-
   member(type(T),Options) -> true; T=text.

at_end_of_stream :-
  current_input(Stream), at_end_of_stream(Stream).

at_end_of_stream(Stream_or_Alias) :- 
  alias(Stream_or_Alias,'$stream'(Stream,_,_,_)), 
  iso_eof(Stream).

set_stream_position(Stream_or_Alias, Position) :- 
  alias(Stream_or_Alias,'$stream'(Stream,_,_,_)), 
  iso_lseek(Stream,Position,0,_).

io_options(read, r).
io_options(write,w).
io_options(append,a).


get_char(Char) :-
  current_input(Stream), get_char(Stream,Char).

get_char(Stream_or_Alias,Char) :- 
  get_byte(Stream_or_Alias,Byte),
  (Byte= -1, Char=end_of_file; name(Char,[Byte])), !.

get_code(Code) :-
  current_input(Stream), get_code(Stream,Code).

get_code(Stream_or_Alias,Code) :- 
  alias(Stream_or_Alias,'$stream'(Stream,_,_,_)), iso_get_byte(Stream,Code).

peek_char(Char) :-
  current_input(Stream), peek_char(Stream, Char).

peek_char(Stream, Char) :-
  peek_byte(Stream,Byte), 
  (Byte= -1, Char=end_of_file; name(Char,[Byte])). 

peek_code(Code) :-
  current_input(Stream), peek_code(Stream,Code).

peek_code(Stream_or_Alias,Code) :-
  alias(Stream_or_Alias,'$stream'(Stream,_,_,_)), iso_peek_byte(Stream,Code).

put_char(Char) :-
  current_output(Stream), put_char(Stream,Char).

put_char(Stream_or_Alias,Char) :- 
  name(Char,[Byte]), put_byte(Stream_or_Alias,Byte).

put_code(Code) :-
  current_output(Stream), put_code(Stream,Code).

put_code(Stream_or_Alias,Code) :- 
  alias(Stream_or_Alias,'$stream'(Stream,_,_,_)), iso_put_byte(Stream,Code).

nl :- put_byte(10).

nl(Stream_or_Alias) :- 
  put_byte(Stream_or_Alias,10).

get_byte(Byte) :-
  current_input(Stream), get_byte(Stream,Byte).

get_byte(Stream_or_Alias,Byte) :-
  alias(Stream_or_Alias,'$stream'(Stream,_,_,_)), iso_get_byte(Stream,Byte).

peek_byte(Byte) :-
  current_input(Stream), peek_byte(Stream,Byte).

peek_byte(Stream_or_Alias,Byte) :-
  alias(Stream_or_Alias,'$stream'(Stream,_,_,_)), iso_peek_byte(Stream,Byte).

put_byte(Byte) :-
  current_output(Stream), put_byte(Stream,Byte).

put_byte(Stream_or_Alias,Byte) :-
  alias(Stream_or_Alias,'$stream'(Stream,_,_,_)), iso_put_byte(Stream,Byte).

alias(Stream_or_Alias, Stream) :-
  val(alias, Stream_or_Alias, Stream), !.
alias('$stream'(Stream,File,Mode,Options),
      '$stream'(Stream,File,Mode,Options)).


%------------ basic input/output primitives -------------


read(T) :- 
  current_input(S), read(S,T).
read(S,T) :- 
  read_term(S,T,[]).  

read_term(T,O) :- 
  current_input(S), read_term(S,T,O). 
read_term(Stream_or_Alias,T,O) :-
  alias(Stream_or_Alias,'$stream'(S,_,_,_)),
  nonvar(O),
  read_options(O,0,0,0,O1,O2,O3,0,R), 
  iso_read_term(S,options(R,O1,O2,O3),T). 

read_options([],I1,I2,I3,I1,I2,I3,A,A).
read_options([variables(Vars)|L],_,I2,I3,O1,O2,O3,A,R) :- 
   !, A2 is A \/ 1, read_options(L,Vars,I2,I3,O1,O2,O3,A2,R).
read_options([variable_names(VN_list)|L],I1,_,I3,O1,O2,O3,A,R) :- 
   !, A2 is A \/ 2, read_options(L,I1,VN_list,I3,O1,O2,O3,A2,R).
read_options([singletons(VN_list)|L],I1,I2,_,O1,O2,O3,A,R) :- 
   !, A2 is A \/ 4, read_options(L,I1,I2,VN_list,O1,O2,O3,A2,R).
read_options([_|L],I1,I2,I3,O1,O2,O3,A,R) :- 
   write('error in read option'),
   read_options(L,I1,I2,I3,O1,O2,O3,A,R).

write_term(T,O) :- 
  current_output(S),
  write_term(S,T,O).
write(T) :- 
  current_output(Stream),
  write_term(Stream,T,[numbervars(true)]).
write(Stream_or_Alias,T) :- 
  write_term(Stream_or_Alias,T,[numbervars(true)]).

writeq(T) :- 
  current_output(S),
  write_term(S,T,[quoted(true),numbervars(true)]).
writeq(S,T) :- 
  write_term(S,T,[quoted(true),numbervars(true)]).

write_canonical(T) :- 
  current_output(S), 
  write_term(S,T,[quoted(true),ignore_ops(true)]).
write_canonical(S,T) :- 
  write_term(S,T,[quoted(true),ignore_ops(true)]).

write_term(Stream_or_Alias,T,O) :-
  alias(Stream_or_Alias,'$stream'(S,_,_,_)),
  nonvar(O),
  write_options(O,0,BP),
  iso_write_term(S,BP,T).

% convert to bit pattern

write_options([],A,A).
write_options([quoted(B)|L],A,R) :- 
   !, (B==true -> A2 is A \/ 1; A2 is A /\ -2), write_options(L,A2,R).
write_options([ignore_ops(B)|L],A,R) :- 
   !, (B==true -> A2 is A \/ 2; A2 is A /\ -3), write_options(L,A2,R).
write_options([numbervars(B)|L],A,R) :- 
   !, (B==true -> A2 is A \/ 4; A2 is A /\ -5), write_options(L,A2,R).
write_options([_|L],A,R) :- 
   write('Error in write options'), write_options2(L,A,R).


% ----------------------- for compatibility ----------------------

generic_write(T,writeq):-write_canonical(T).

see(user) :- !, see(user_input).
see(F) :- 
    (val('$see',F,S); 
     (open(F,read,S2), saved(S2,S), def('$see',F,S))), !, set_input(S).

tell(user) :- !, tell(user_output).
tell(F) :- 
    (val('$tell',F,S); 
     (open(F,write,S2), saved(S2,S), def('$tell',F,S))), !, set_output(S).

seen :- current_input('$stream'(_,user_input,_,_)), !.
seen :- current_input(S), close(S), S='$stream'(_,F,_,_),
        rm('$see',F), see(user_input).

told :- current_input('$stream'(_,user_output,_,_)), !.
told :- current_output(S), close(S), S='$stream'(_,F,_,_),
        rm('$tell',F), tell(user_output).

seeing(F) :- current_input('$stream'(_,F,_,_)).

telling(F) :- current_output('$stream'(_,F,_,_)).

find_file(user,user).   % added for compatibility
find_file(File,NewFile):-atom(File),!,
        file_search_path(Path),
        file_extension_list(Ext),
        see_a_file(Path,File,Ext,NewFile), !.  
find_file(library(File),NewFile):-atom(File),!,
        file_library(Path,Ext),
        see_a_file(Path,File,Ext,NewFile), !. 
find_file(File,_):-
        user_error('bad file name',File).

see_a_file(Prefs,File,Sufs,NewFile):-
        make_file_name(Prefs,File,Sufs,NewFile),
        open(NewFile, read, S, []), close(S).

putbyte(C) :- put_code(C).

print(X):-current_output(S), write_term(S,X,[]).
display(X):-current_output(S), write_term(S,X,[ignore_ops(true)]).

portable_print(X) :- print(X).
portable_write(X) :- write(X).
portable_writeq(X) :- writeq(X).
portable_display(X) :- display(X).

switch_vars([],[]).
switch_vars([X=Y|L1],[Y=X|L2]) :- switch_vars(L1,L2).

prolog_r_term(A,V) :- current_input(S), read_term(S,A,[variable_names(V)]).
prolog_read(X):-prolog_r_term(X,_).

ttyin(top_read_term(T,W)) :- !,
  read_term(user_input,T,[variable_names(V)]), switch_vars(V,W). 

ttyin(get0(A)) :- !,
  get_code(user_input,A).

ttyin(ReadOp):- 
  seeing(F), see(user), ReadOp, seen, see(F).

ttyout(writeq(T)) :- !,
  alias(user_output,'$stream'(S,_,_,_)),
  iso_write_term(S,1,T). 
ttyout(WriteOp):-
  current_output(Out),
  set_output(user_output),
  WriteOp,flush_output,
  set_output(Out).

ttyprin(X) :- write(user_output, X).
ttyprint(X) :- write(user_output, X), nl(user_output).
ttywrite(X):-ttyprin(X).
ttywriteln(X):-ttyprint(X).
ttynl:-nl(user_output).
ttyput(X):-ttyout(put_code(X)).


portray_clause(C):-pp_clause0(C),fail.
portray_clause(_).

pp_clause0(:-(Body)) :- !,
	nl,
	numbervars(Body, 0, _),
	l_clauses(Body, 0, 2, 8). % 0!='' in sicstus,sb..
pp_clause0((Pred:-Body)) :-
	numbervars(Pred+Body, 0, _),
	portable_writeq(Pred),
	l_clauses(Body, 0, 2, 8), !.
pp_clause0((Pred)) :-
	pp_clause0((Pred:-true)).


l_clauses((A,B), L, R, D) :- !,
	l_clauses(A, L, 1, D), !,
	l_clauses(B, 1, R, D).
l_clauses(true, _L, 2, _D) :- !,[P]=".",
	put_code(P), nl.
l_clauses((A;B), L, R, D) :- !,
	l_magic(fail, L, D),
	l_magic((A;B), 0, 2, D),
	l_magic_nl(R, '.').

l_clauses((A->B), L, R, D) :- !,
	l_clauses(A, L, 5, D), !,
	l_clauses(B, 5, R, D).
l_clauses(Goal, L, R, D) :-
	l_magic(Goal, L, D),
	portable_writeq(Goal),
	l_magic_nl(R,'.').

l_magic(!,    0, _D) :- !,
	fast_write(' :- ').
l_magic(!,    1, _D) :- !,
	fast_write(',  ').
l_magic(_Goal, 0, D) :- !,
	fast_write(' :- '),
	nl, tab(D).
l_magic(_Goal, 1, D) :- !, [Char]=",",
	put_code(Char),
	nl, tab(D).
l_magic(_Goal, 3, _D) :- !,
	fast_write('(   ').
l_magic(_Goal, 4, _D) :- !,
	fast_write(';   ').
l_magic(_Goal, 5, D) :- !,
	fast_write(' ->'),
	nl, tab(D).
l_magic(_Goal, Key, D) :-
	atom(Key),
	fast_write((':- ')), fast_write(Key),
	nl, tab(D).

l_magic_nl(2, C) :- !, fast_write(C),nl.
l_magic_nl(_, _).

l_magic((A;B), L, R, D) :- !,
	l_magic(A, L, 1, D), !,
	l_magic(B, 1, R, D).
l_magic(Conj,  L, R, D) :-
	E is D+8,
	M is L+3,
	l_clauses(Conj, M, 1, E),
	nl, tab(D),
	l_magic2(R, ')' ).

l_magic2(2, C) :- !, fast_write(C).
l_magic2(_, _).


peval_io(_,_):-fail.

read_clause(C_dcg) :- read(C),std_expand_term(C,C_dcg).
fast_write(C) :- write(C).

read_with_singletons(C,Vs,Ss) :-
  current_input(S),
  read_term(S,C,[singletons(Sing),variable_names(VN)]),
  singletonnames(Sing,Ss),
  termvarlist(VN,Vs).

singletonnames([],[]).
singletonnames([_X=N|L1],[N|L2]) :- singletonnames(L1,L2).

termvarlist([],[]).
termvarlist([X=N|L1],[var(N,X,_)|L2]) :- termvarlist(L1,L2).

warn_singletons([S|Ss],Vs,C):-quiet(Q),Q=<1,
   warn_singletons1(C,[S|Ss],Vs),
   fail.
warn_singletons(_,_,_).

%warn_singletons(_Ss,_Vs,_C).

warn_singletons1(':-'(_),_,_):-!.
warn_singletons1(C,Ss,Vs):-
  add_true(C,(H:-_)),
  seeing_at(Byte),
  melt_varnames(Vs),
  ttyout((
    print(Ss),print('** warning singleton_variables, byte=>'(Byte)),nl,
    print('=> '),print(H),fast_write(':-...'),nl,
    nl
  )),
  !.

melt_varnames([]).
melt_varnames([var(V,V,_)|Xs]) :- melt_varnames(Xs).
