% generic client/server predicates

default_host(Host):-is_dynamic(host(_)),host(Host),!.
default_host(Host):-assumed(host(Host)),!.
default_host(localhost).

default_port(Port):-is_dynamic(port(_)),port(Port),!.
default_port(Port):-assumed(port(Port)),!.
default_port(7777).

client(Query,Answer):-
  default_host(Host),
  default_port(Port),
  client(Host,Port,Query,Answer).

server(Interactor,Port,Timout,Fuel):-
  quietmes(1,'starting server on port'(Port)),
  new_server(Port,Server),
  server_loop(Interactor,Server,Timout,Fuel),
  close_socket(Server),
  quietmes(1,'server done on port'(Port)).

server_loop(Interactor,Server,Timout,Fuel):-
  for(I,1,Fuel),
    quietmes(1,'serving client'(I)),
    interact(Interactor,Server,Timout),
  fail.
server_loop(_,_,_,_).

interact(Interactor,Server,Timout):-
  % blocks here until one of a max of 64 waiting clients connect
  new_service(Server,Timout,Service), 
  sock_read_term(Service,Input),
  call(Interactor,Input,Output),
  sock_write_term(Service,Output),
  close_socket(Service),
  !,
  fail.
interact(Interactor,Server,Timout):-
  errmes(bad_interaction,server(Interactor,Server,Timout)).

client(Host,Port,Question,Answer):-
  new_client(Host,Port,Connection),
  sock_write_term(Connection,Question),
  sock_read_term(Connection,Result),
  close_socket(Connection),
  !,
  Answer=Result.
client(Host,Port,Question,_):-
  errmes(bad_interaction,client(Host,Port,Question)).

sock_read_term(Socket,Term):-
  sock_readln(Socket,String),
  quietmes(1,'Read'(String)),
  read_term_from_chars(String,Term).

sock_write_term(Socket,Term):-
  swrite(Term,String),
  quietmes(1,'Written'(String)),
  sock_writeln(Socket,String).

% BinProlog Linda server

linda_server:-default_port(Port),linda_server(Port).

linda_server(Port):-
  Fuel is 1<<28,
  server(linda_answer,Port,0,Fuel).

linda_answer(cin(X),R):-linda_cin(X,R).
linda_answer(out(X),R):-linda_out(X,R).
linda_answer(rd(X),R):-linda_rd(X,R).
linda_answer(all(X),Xs):-linda_facts(X,X,Xs).
linda_answer(all(X,G),Xs):-linda_collect(X,G,Xs).
linda_answer(run(G),R):-G->R=G;R=no.
linda_answer(run(X,G),R):-G->R=X;R=no.

linda_cin(X,R):-retract1(X),!,R=X.
linda_cin(_,no).

linda_rd(X,R):-clause(X,true),!,R=X.
linda_rd(_,no).

linda_out(no,no):-!.
linda_out(yes,no):-!.
linda_out(X,yes):-assert(X).

linda_collect(X,G,Xs):-nonvar(G),!,findall(X,G,Xs).
linda_collect(X,G,Xs):-linda_facts(X,G,Xs).

linda_facts(X,G,Xs):-findall(X,clause(G,true),Xs).

% Linda clients for Java and BinProlog Linda servers
% -- to be merged once both servers will provide similar functions

% Linda operations with Java server
java_out(X):-client(out(X),yes).
java_in(X):-client(in(X),X).
java_all(X,Xs):-client(all(X),Xs).
java_rd(X):-java_all(X,[X|_]).
java_cin(X):-java_rd(X),java_in(X).

% Linda operations with BinProlog server
bp_out(X):-client(out(X),yes).
bp_in(X):-bp_cin(X)->true;wait_for_in(8,X). % to be redone with pthreads!
bp_all(X,Xs):-client(all(X),Xs).
bp_rd(X):-client(rd(X),X).
bp_cin(X):-client(cin(X),X).

bp_all(X,G,Xs):-client(run(X,G),Xs).
bp_run(G):-client(run(G),G).            % runs G on server!
bp_run(X,G):-client(run(X,G),G).

wait_for_in(Max,X):-for(I,1,Max),sleep(I),bp_cin(X),!.

/*
% simple benchmarks

jbm(N):-for(I,1,N),out(a(I)),fail.
jbm(_).

pbm(N):-for(I,1,N),bp_out(a(I)),fail.
pbm(_).

bm(G):-
  statistics,
  ctime(T1),G,ctime(T2),T is T2-T1,
  statistics,
  write(time(T)),nl.

jbm:-bm(jbm(1000)).

pbm:-bm(pbm(1000)).
*/
