% interface

linda_server:-default_port(Port),linda_server(Port).

linda_server(Port):-
  Fuel is 1<<28,
  server(serve_answer,Port,0,Fuel).

linda_out(X):-default_lang(java),!,java_out(X).
linda_out(X):-bp_out(X).

linda_in(X):-default_lang(java),!,java_in(X).
linda_in(X):-bp_in(X).

linda_all(X,Xs):-default_lang(java),!,java_all(X,Xs).
linda_all(X,Xs):-bp_all(X,Xs).

linda_rd(X):-default_lang(java),!,java_rd(X).
linda_rd(X):-bp_rd(X).

linda_cin(X):-default_lang(java),!,java_cin(X).
linda_cin(X):-bp_cin(X).

linda_host_id(S):-asserted(server_id(X)),!,S=X.
linda_host_id(S):-client(id(R),R),assert(server_id(R)),S=R.

linda_fork_server(Pid):-default_port(Port),linda_fork_server(Port,Pid).

linda_fork_server(Port,Pid):-
  (free_port(Port)->true;errmes('port busy',Port)),
  unix_fork(Pid),
  ( Pid=:=0->linda_server(Port),halt % child: a server
  ; sleep(1),default_host(Host),              % parent
    Id=server_id(server(bp,Host,Port)),
    assert(Id),
    linda_out(Id)
  ).

% generic client/server predicates

default_host(X):-default(host(X),X=localhost).

default_port(X):-default(port(X),X=7777).

default_lang(Lang):-linda_host_id(server(Lang,_,_)).

client(Query,Answer):-
  default_host(Host),
  default_port(Port),
  client(Host,Port,Query,Answer).

free_port(Port):-new_server(Port,NewServer),close_socket(NewServer).

server(Interactor,Port,Timout,Fuel):-
  quietmes(1,'starting server on port'(Port)),
  new_server(Port,NewServer),
  server_loop(Interactor,NewServer,Timout,Fuel),
  close_socket(NewServer),
  quietmes(1,'server done on port'(Port)).

server_loop(Interactor,Server,Timout,Fuel):-
  for(I,1,Fuel),
    quietmes(1,'serving client'(I)),
    interact(Interactor,Server,Timout),
  fail.
server_loop(_,_,_,_).

interact(Interactor,Server,Timout):-
  % blocks here until one of a max of 64 waiting clients connect
  new_service(Server,Timout,Service), 
  sock_read_term(Service,Input),
  call(Interactor,Input,Output),
  sock_write_term(Service,Output),
  close_socket(Service),
  !,
  fail.
interact(Interactor,Server,Timout):-
  errmes(bad_interaction,server(Interactor,Server,Timout)).

client(Host,Port,Question,Answer):-
  new_client(Host,Port,Connection),
  sock_write_term(Connection,Question),
  sock_read_term(Connection,Result),
  close_socket(Connection),
  !,
  Answer=Result.
client(Host,Port,Question,_):-
  errmes(bad_interaction,client(Host,Port,Question)).

sock_read_term(Socket,Term):-
  sock_readln(Socket,String),
  quietmes(1,'Read'(String)),
  read_term_from_chars(String,Term).

sock_write_term(Socket,Term):-
  swrite(Term,String),
  quietmes(1,'Written'(String)),
  sock_writeln(Socket,String).

% BinProlog Linda server operations

serve_answer(cin(X),R):-serve_cin(X,R).
serve_answer(out(X),R):-serve_out(X,R).
serve_answer(rd(X),R):-serve_rd(X,R).
serve_answer(all(X),Xs):-serve_facts(X,X,Xs).
serve_answer(all(X,G),Xs):-serve_collect(X,G,Xs).
serve_answer(run(G),R):-G->R=G;R=no.
serve_answer(run(X,G),R):-G->R=X;R=no.
serve_answer(id(R),R):-default_host(H),default_port(P),
  R=server(bp,H,P).

serve_cin(X,R):-retract1(X),!,R=X.
serve_cin(_,no).

serve_rd(X,R):-clause(X,true),!,R=X.
serve_rd(_,no).

serve_out(no,no):-!.
serve_out(yes,no):-!.
serve_out(X,yes):-assert(X).

serve_collect(X,G,Xs):-nonvar(G),!,findall(X,G,Xs).
serve_collect(X,G,Xs):-serve_facts(X,G,Xs).

serve_facts(X,G,Xs):-findall(X,clause(G,true),Xs).

% Linda clients for Java and BinProlog Linda servers
% -- to be merged once both servers will provide similar functions

% Linda operations with Java server
java_out(X):-client(out(X),yes).
java_all(X,Xs):-client(all(X),Xs).

java_cin(X):-java_rd(X),java_in(X).
java_rd(X):-java_all(X,[X|_]).
java_in(X):-client(in(X),X).

% Linda operations with BinProlog server
bp_out(X):-client(out(X),yes).
bp_all(X,Xs):-client(all(X),Xs).

bp_rd(X):-client(rd(X),X).
bp_cin(X):-client(cin(X),X).
bp_in(X):-bp_cin(X)->true;wait_for_in(8,X). % to be redone with pthreads!

bp_all(X,G,Xs):-client(run(X,G),Xs).
bp_run(G):-client(run(G),G).            % runs G on server!
bp_run(X,G):-client(run(X,G),G).

wait_for_in(Max,X):-for(I,1,Max),sleep(I),bp_cin(X),!.


% simple benchmarks

obm(N):-for(I,1,N),linda_out(a(I)),fail.
obm(_).

ibm(N):-for(_,1,N),linda_in(a(_)),fail.
ibm(_).

bm(G):-
  statistics,
  ctime(T1),G,ctime(T2),T is T2-T1,
  statistics,
  write(time(T)),nl.

ibm:-bm(ibm(500)).

obm:-bm(obm(1000)).

