% Socket based Linda client/server - Java server comptible

% interface

  run_server:-default_port(Port),  run_server(Port).

  fork_server:-fork_server(_).

  fork_server(Pid):-default_port(Port),  fork_server(Port,Pid).

  stop_server:-stop_server(0).

  stop_server(Code):-bp_stop_server(Code).

  ask_server(Query,Answer):-
    default_host(Host),
    default_port(Port),
    ask_server(Host,Port,Query,Answer).

  server_info(S):-asserted(server_id(X)),!,S=X.
  server_info(S):-ask_server(id(R),R),assert(server_id(R)),S=R.

  set_host(Server):-
    retractall(server_id(_)),
    retractall(host(_)),assert(host(Server)).

  set_port(Port):-
    integer(Port),
    retractall(server_id(_)),
    retractall(port(_)),assert(port(Port)).

  % Linda operations: all based on ask_server/2


  out(X):-default_lang(L),
    ( L==bp->bp_out(X)
    ; L==java->java_out(X)
    ; errmes(linda_error,out(X))
    ).

  cin(X):-default_lang(L),
    ( L==bp->bp_cin(X)
    ; L==java->java_cin(X)
    ; errmes(linda_error,cin(X))
    ).

  in(X):-default_lang(L),
    ( L==bp->bp_in(X)
    ; L==java->java_in(X)
    ; errmes(linda_error,in(X))
    ).

  all(X,Xs):-default_lang(L),
    ( L==bp->bp_all(X,Xs)
    ; L==java->java_all(X,Xs)
    ; errmes(linda_error,all(X,Xs))
    ).

  rd(X):-default_lang(L),
    ( L==bp->bp_rd(X)
    ; L==java->java_rd(X)
    ; errmes(linda_error,rd(X))
    ).

% generic client/server predicates

default_host(X):-default(host(X),X=localhost).

default_port(X):-default(port(X),X=7777).

default_lang(Lang):-  server_info(server(Lang,_,_)).

fork_server(Port,Pid):-
    (free_port(Port)->true;errmes('port busy',Port)),
    unix_fork(Pid),
    ( Pid=:=0->
      quietmes(1,'forking with pid'(Pid)),
      run_server(Port), % child: a server
      quietmes(1,'forked server done, pid'(Pid)),
      halt
    ; sleep(1),default_host(Host),              % parent
      Id=server_id(server(bp,Host,Port)),
      assert(Id),
      out(Id)
    ).

free_port(Port):-new_server(Port,NewServer),close_socket(NewServer).

run_server(Port):-
  Fuel is 1<<28,
  run_server(serve_answer,Port,0,Fuel).

run_server(Interactor,Port,Timout,Fuel):-
  quietmes(1,'running server on port'(Port)),
  new_server(Port,NewServer),
  server_loop(Interactor,NewServer,Timout,Fuel),
  close_socket(NewServer),
  quietmes(1,'server done on port'(Port)).

server_loop(Interactor,Server,Timout,Fuel):-
  for(I,1,Fuel),
    quietmes(1,'serving client'(I)),
    interact(Interactor,Server,Timout),
  fail.
server_loop(_,_,_,_).

interact(Interactor,Server,Timout):-
  % blocks here until one of a max of 64 waiting clients connect
  new_service(Server,Timout,Service), 
  sock_read_term(Service,Input),
  quietmes(1,interactor_call(Interactor)),
  call(Interactor,Input,Output),
  quietmes(1,interacator_action(Input,Output)),
  sock_write_term(Service,Output),
  close_socket(Service),
  !,
  fail.
interact(Interactor,Server,Timout):-
  errmes(bad_interaction,server(Interactor,Server,Timout)).

ask_server(Host,Port,Question,Answer):-
  new_client(Host,Port,Connection),
  sock_write_term(Connection,Question),
  sock_read_term(Connection,Result),
  close_socket(Connection),
  !,
  Answer=Result.
ask_server(_,_,stop(_),yes):-!.
ask_server(Host,Port,Question,_):-
  errmes(bad_interaction,client(Host,Port,Question)).

sock_read_term(Socket,Term):-
  sock_readln(Socket,String),
  quietmes(1,'Read'(String)),
  read_term_from_chars(String,Term).

sock_write_term(Socket,Term):-
  swrite(Term,String),
  quietmes(1,'Written'(String)),
  sock_writeln(Socket,String).

% helpers for Linda server operations

serve_cin(X,R):-nonvar(X),retract1(X),!,R=X.
serve_cin(X,R):-var(X),clause(X,true),retract1(X),!,R=X.
serve_cin(_,no).

serve_rd(X,R):-clause(X,true),!,R=X.
serve_rd(_,no).

serve_out(no,no):-!.
serve_out(yes,no):-!.
serve_out(X,yes):-assert(X).

serve_collect(X,G,Xs):-nonvar(G),!,findall(X,G,Xs).
serve_collect(X,G,Xs):-serve_facts(X,G,Xs).

serve_facts(X,G,Xs):-findall(X,clause(G,true),Xs).

serve_clauses(DB,H,HBs):-
  findall((H:-B),db_clause(DB,H,B),HBs).

serve_file(File,HBs):-
  DB=mirror,
  consult(File,DB),
  serve_clauses(DB,_,HBs),
  db_clean(DB).

% BinProlog Linda server operations

serve_answer(cin(X),R):-serve_cin(X,R).
serve_answer(out(X),R):-serve_out(X,R).
serve_answer(rd(X),R):-serve_rd(X,R).
serve_answer(all(X),Xs):-serve_facts(X,X,Xs).
serve_answer(all(X,G),Xs):-serve_collect(X,G,Xs).
serve_answer(run(G),R):-G->R=G;R=no.
serve_answer(run(X,G),R):-G->R=X;R=no.
serve_answer(id(R),R):-default_host(H),default_port(P),
  R=server(bp,H,P).
serve_answer(stop(X),yes):-halt(X).
serve_answer(rsh(Cmd),Xs):-pcollect(Cmd,Xs).
serve_answer(consult(File),Cs):-serve_file(File,Cs).
serve_answer(clauses(DB,H),Cs):-serve_clauses(DB,H,Cs).

% Linda clients for BinProlog and Java Linda servers
% -- to be merged once both servers will provide similar functions

% Linda operations with BinProlog server
bp_out(X):-ask_server(out(X),yes).
bp_all(X,Xs):-ask_server(all(X),Xs).

bp_rd(X):-ask_server(rd(X),R),!,R\==no,X=R.
bp_cin(X):-ask_server(cin(X),R),!,R\==no,X=R.
bp_in(X):-bp_cin(X)->true;wait_for_in(8,X). % to be redone with pthreads!

bp_all(X,G,Xs):-ask_server(run(X,G),Xs).
bp_run(G):-ask_server(run(G),G).            % runs G on server!
bp_run(X,G):-ask_server(run(X,G),G).

bp_rsh(Cmd,Chars):-ask_server(rsh(Cmd),Chars).

bp_rsh(Cmd):-bp_rsh(Cmd,Xs),(member(X,Xs),put(X),fail;true).

bp_consult(File):-
   ask_server(consult(File),Cs),
   forall(member(C,Cs),assert(C)).

bp_clauses(DB,H,Cs):-
   ask_server(clauses(DB,H),Cs).

bp_stop_server(X):-ask_server(stop(X),yes).

wait_for_in(Max,X):-for(I,1,Max),sleep(I),bp_cin(X),!.

% Linda operations with Java server
java_out(X):-ask_server(out(X),yes).
java_all(X,Xs):-ask_server(all(X),Xs).

java_cin(X):-java_rd(X),java_in(X).
java_rd(X):-java_all(X,[X|_]).
java_in(X):-ask_server(in(X),X).


/*
% simple benchmarks

obm(N):-for(I,1,N),  out(a(I)),fail.
obm(_).

ibm(N):-for(_,1,N),  in(a(_)),fail.
ibm(_).

bm(G):-
  statistics,
  ctime(T1),G,ctime(T2),T is T2-T1,
  statistics,
  write(time(T)),nl.

ibm:-bm(ibm(500)).

obm:-bm(obm(1000)).
*/
