
% Program: Prototype based generic object system for BinProlog
% Author: Paul Tarau
% Version: Tue Jan 30 23:27:07 AST 1996
% To be included into a file defining concrete operations

/*
:-op(50,fx,(@)).
:-op(50,fx,(^)).
:-op(450,fx,(::)).
:-op(880,xfx,(extends)).
:-op(890,xfx,(with)).
:-op(890,xfy,(::)).
*/

% :-dynamic this/1.

% interface

(A extends P) with Parts:-!,
  clone(P,A),
  A :: Parts.
A with Parts:-
  default_proto(P),
  ( is_object(P)->true
    ; proto(P)
  ),
  clone(P,A),
  A :: Parts.

A extends P:-clone(P,A).

A :: Action :- call_with_this(A,Action).

proto(X):-parent_of(X,X).

clone(Parent,Clone):-is_object(Parent),make_clone(Parent,Clone).

%defines, sets and retrives the value of a variable
Var:=Val:-this(Object),add_var(Object,Var,Val).
Var<<Val:-this(Object),set_it(Object,Var,Val).
Var>>Val:-this(Object),get_it(Object,Var,Val).

% defines a verb which will be inherited by clones
:: Verb :- this(Object), add_verb(Object,Verb).

% defines dynamic object component clauses
(H:-B) :- this(Object),paste_verb(Object,H,B).

% calls a verb in its context
@Verb:-this(Object),call_verb(Object,Verb).

% calls a (usually overridden) verb in `super' 
^Verb:-
   this(Object),get_atomic(Object,':>',Parent),Object\==Parent,
   !,
   Parent :: Verb.
^Verb:-
   Verb.

% recognizers

is_object(Name):-get_atomic(Name,'<:',_).
is_proto(Name):-get_atomic(Name,'<:',Name).
%is_verb(Pred):-get_atomic('::',Pred,_).

% finders

get_proto(Name,Proto):-pval(Proto,'<:',Name,Proto).

get_verb(Object,Verb,NewVerb):-
   nval(Object,'<:',Verb,NewF),
   term_append(NewF,Verb,NewVerb).


% essential operations


make_clone(Parent,Clone):-
  make_name(Parent,Clone),
  parent_of(Clone,Parent),
  inherit_vars(Parent,Clone).

add_var(Object,Var,Val):-
   def_it(Object,Var,Val),!,
   add_it(Object,'>>',Var).
add_var(Object,Var,Val):-
   set_it(Object,Var,Val).

add_verb(Object,Verb):-
   new_verb(Object,Verb,Pred,LocalVerb),
   def_it(Object,Pred,LocalVerb), % links name to local name
   add_it('::',Pred,Object),        % puts object on pred's list
   add_it(Object,'::',Verb).        % puts verb on object's list

paste_verb(Object,H,B):-
   functor(H,F,N),
   add_verb(Object,F/N), 
   new_verb(Object,F,_,OF),
   term_append(OF,H,OH),
   decorate_body(B,NewB),
   make_method((OH:-NewB)).

call_verb(Object,Verb):-get_verb(Object,Verb,NewVerb),call_method(NewVerb).


% details

decorate_body((A,B),(DA,DB)):-!,
   decorate_body(A,DA),decorate_body(B,DB).
decorate_body((A;B),(DA;DB)):-!,
   decorate_body(A,DA),
   decorate_body(B,DB).
decorate_body((A->B),(DA->DB)):-!,
   decorate_body(A,DA),
   decorate_body(B,DB).
decorate_body(A,B):-is_builtin(A),!,B=A.
decorate_body(@A,B):-!,B= @A.
decorate_body(^A,B):-!,B= ^A.
decorate_body(A,@A).

parent_of(Name,Proto):-
     def_it(Name,'<:',Proto),
     add_it(Proto,':>',Name).

make_name(Name,NewName):-var(NewName),!,
   get_proto(Name,Proto),
   gensym(Proto,NewName).
make_name(_,_).

new_verb(Object,F/N,P,NewF):-!,namecat(Object,'::',F,NewF),functor(P,F,N).
new_verb(Object,F,F,NewF):-namecat(Object,'::',F,NewF).

inherit_vars(Parent,Clone):-
  search_it(Parent,'>>',Var),
    get_it(Parent,Var,Val),
    add_var(Clone,Var,Val),
  fail.
inherit_vars(_,_).

nval(O0,_,V0,V):-get_atomic(O0,V0,V),!.
nval(O0,Link,V0,V):-get_atomic(O0,Link,O1),O0\==O1,nval(O1,Link,V0,V).

pval(V,Link,Go,Go):-get_atomic(Go,Link,V),!.
pval(V,Link,Go0,Go2):-get_atomic(Go0,Link,Go1),pval(V,Link,Go1,Go2).


% show them

oblist0(Object,Object):-get_atomic(Object,'<:',Object).
oblist0(Object,extends(Object,P)):-get_atomic(Object,'<:',P),P\=Object.
oblist0(Object,Var:=Val):-search_it(Object,>>,Var),get_it(Object,Var,Val).
oblist0(Object,::Verb):-search_it(Object,::,Verb).

oblist(Object,X with Xs):-
  findall_conj(X,oblist0(Object,X),(X,Xs)).

oblist(Object):-
  oblist(Object,X),
  pp_clause((:-X)),nl,
  nl,
  fail.
oblist(_).

oblist:-
  default_proto(P),
  accessible_from(P,O),
  oblist(O),
  fail.
oblist.

accessible_from(P,P).
accessible_from(P,O):-
  search_it(P,':>',NewP),NewP\=P,
  accessible_from(NewP,O).

  
% generic tools to be defined in a file including this.
% for instance:

/*
default_proto(object).
make_method(C):-assert(C).
add_it(O,M,V):-addq(O,M,V).
search_it(O,M,V):-membq(O,M,V).
def_it(O,M,V):-bb_def(O,M,V).
set_it(O,M,V):-bb_set(O,M,V).
get_it(O,M,V):-bb_val(O,M,V).
get_atomic(O,M,V):-val(O,M,V).
kill_it(O,M):-bb_rm(O,M).
call_method(M):-M.
call_with_this(O,G) :- copy_term(G,NewG), this(O) =>> NewG.
*/
