/*LINTLIBRARY*/

/*
 *  Memory allocator based on a buddy system algorithm
 *
 *  28/02/93 Initial version: Koen De Bosschere
 */

#include <stdio.h>
#include "buddy.h"

extern void exit(int);

#define INVALID (buddyblock_t *)0x1


/* array containing the pointers to the lists */

void buddycreate(buddyblock_t **buddy, int size)
                     
/*
 * create a buddy system
 * size must be at least BUDDYSIZE bytes
 */
{
  unsigned int i;

  if (!buddy) {
    fprintf(stderr, "Nilpointer in buddycreate\n");
    exit(1);
  }
  if ((buddyblock_t **)*buddy != &(buddy[0]->next)) {
    fprintf(stderr, "Field next in struct buddyblock must have offset 0\n");
    exit(1);
  }
  if (size < BUDDYSIZE) {
    fprintf(stderr, "Data structure in buddycreate too small (<%d)\n",
       BUDDYSIZE);
    exit(1);
  }
  for (i=0; i<BUDDYMAX; i++) {
    buddy[i] = (buddyblock_t *)NULL;
  }
  buddy[BUDDYMAX] = INVALID;  /* special value to indicate the last list */
}

void buddyinit(buddyblock_t **buddy, buddyblock_t *address, unsigned int size)
                     
                      
                  
/*
 * Add a block of memory to the buddy system (address,size).
 *
 * The block is cut into pieces such that the addresses of
 * adjacent blocks that belong together differ in only
 * one bit. The position of this bit determnines size of
 * both blocks. This property allows for fast recombination.
 * The disadvantage is that the orginale chunk of memory
 * might be fragmented even before one malloc has been
 * executed. However, the biggest chunk will always be
 * at least 50% of the original chunk if that chunk's
 * size is a power of two.
 */
{
  unsigned int bi = BUDDYMAX-1;

  size &= 0xfffffffc;   /* chunk must consist of words */
  while (size >= 4) { 
    if ((1 << bi) <= size) {
      unsigned int mask = (1 << bi) - 1;
      if ((unsigned int)address & mask) { /* split */
        buddyblock_t *splitaddress = 
            (buddyblock_t *)((unsigned int)((char *)address+size-1) & ~mask);
        unsigned int sizeoffirstbuddyblock_t = (char *)splitaddress - (char *)address;
        buddyinit(buddy,address,sizeoffirstbuddyblock_t);
        buddyinit(buddy,splitaddress,size-sizeoffirstbuddyblock_t);
        return;
      } else { /* buddyblock_t is at properly aligned address */
        address->next = buddy[bi];
        buddy[bi] = address;
        address += 1 << bi-2;
        size -= 1 << bi;
      }
    } 
    bi--;
  }
}

unsigned int listlength(buddyblock_t *address)
                      
/*
 * subsidiairy function 
 */
{
  unsigned int n = 0;
  while (address) {
    n++;
    address = address->next;
  }
  return n;
}

unsigned int buddymax(buddyblock_t **buddy)
                     
/* 
 * return the size of the biggest buddy block available
 */
{
  unsigned int bi = BUDDYMAX-1;
  while (!buddy[bi] && bi) bi--;
  return bi ? (1 << bi) : 0;
}

unsigned int totalbuddymem(buddyblock_t **buddy)
                     
/*
 * return the total amount of memory available 
 */
{
  unsigned int total = 0, bi;
  for (bi=0; bi<BUDDYMAX; bi++) 
    total += (1 << bi) * listlength(buddy[bi]);
  return total;
}

int buddyshow(buddyblock_t **buddy)
{
  unsigned int bi;
  buddyblock_t *bp;

  if (!buddy) {
    fprintf(stderr, "Nilpointer in buddyshow\n");
    return 0;
  }
  printf("nn [   size]  length\n");
  printf("--------------------\n");
  for (bi=0; bi<BUDDYMAX; bi++) {
    bp = buddy[bi];
    if (bp) {
      printf("%2u [%7u]: %6u ",bi,  1<<bi, listlength(buddy[bi]));
      while (bp) {
        printf(" %08x ", bp); 
        bp = bp->next;
      }
      printf("\n");
    }
  }
  printf("--------------------\n");
  printf("Total memory: %u\n", totalbuddymem(buddy));
  return 1;
}

buddyblock_t *buddymalloc(buddyblock_t **buddy, unsigned int size)
                     
                  
/*
 * returns a pointer to a chunk of memory of at least size bytes
 * return NULL if there is not enough memory
 */
{
  register unsigned int buddysize = 4;
  register buddyblock_t *bp, **qq = (buddy+2);
  while (buddysize < size) {
    qq++; buddysize <<= 1;
  }
  if (*qq) {
    bp = *qq;
    *qq = bp->next;
    return bp;
  } else {
    register buddyblock_t **q;

    for (q=qq+1; !*q; buddysize <<= 1, q++) ;
    if (*q == INVALID) return (buddyblock_t *)NULL;
    bp = *q;
    *q = bp->next;
    for (q--; q>=qq; q--) {
      bp->next = (buddyblock_t *)NULL;
      *q = bp;
      bp += buddysize>>2;
      buddysize >>= 1;
    }
    return bp;
  }
}

void buddyfree(buddyblock_t **buddy, buddyblock_t *address, unsigned int size)
                     
                      
                  
/*
 * returns a block of memory at address of size bytes to
 * the system. If possible, this block will be fused
 * with other available blocks to create bigger blocks.
 */
{
  register buddyblock_t *p, *q, *tq, *qq = (buddyblock_t *)(buddy+2);
  register unsigned int buddysize = 4;

  while (buddysize < size) {
    qq++; buddysize <<= 1;
  }
tailcallentry:
  for (q = qq, p = q->next, tq = (buddyblock_t *)NULL; p>address; tq = q, q = p, p = p->next) ;
  if (((unsigned int)address ^ (unsigned int)p) == buddysize && p) {
    q->next = p->next;
    address = p;
    buddysize <<= 1; qq++;
    goto tailcallentry;  
  } else if (((unsigned int)q ^ (unsigned int)address) == buddysize && tq) {
    tq->next = q->next;
    buddysize <<= 1; qq++;
    goto tailcallentry; 
  } else {
    address->next = p;
    q->next = address;
  }
}

unsigned int realbuddysize(unsigned int size)
{
  register unsigned int buddysize = 4;

  while (buddysize < size) {
    buddysize <<= 1;
  }
  return buddysize;
}

