/*********************EXTRA LIBRARY TOOLS **************************/

% sort, adapted from public domain code written by R.A. O'Keefe
% use merge_sort(<,_,_) if you do not want duplications eliminated
% use merge_sort(>,_,_) for descending order

sort(L1,L2):-merge_sort(<,L1,DupL),remdup(DupL,L2).

msort(L1,L2):-merge_sort(<,L1,L2).

remdup([],[]).
remdup([X,Y|Xs],Ys):-compare(=,X,Y),!,remdup([X|Xs],Ys).
remdup([X|Xs],[X|Ys]):-remdup(Xs,Ys).
      
merge_sort(Rel, L,S ):-
	length(L,N),
	merge_sort1(N, Rel, L,S,[] ).

merge_sort1( 0,_,L,[],L ):-!.
merge_sort1( 1,_,[X|L],[X],L ):-!.
merge_sort1( N,Rel,L,S,R ):-	% N >= 2
	N1 is N >> 1,	N2 is N-N1,
	merge_sort1( N1,Rel,L,S1,R1),	
	merge_sort1( N2,Rel,R1,S2,R),
	merge_2( S2,Rel,S1,S ).

merge_2([],_,S,S ):-!.
merge_2([X|L1],Rel,[Y|L2],[X|L] ):-compare(Rel,X,Y),!,
	merge_2(L1,Rel,[Y|L2],L ).
merge_2(L1,Rel,[Y|L2],[Y|L] ):-
	merge_2(L2,Rel,L1,L ).

%   Keysorting.  Adapted by Mats Carlsson from R.O'Keefe's code, 
%		but uses recursion instead of an auxiliary stack.  
%		Takes care to check validity of arguments.
%   Could be speed up if there were an inline keycompare/3.

ksort(List, Sorted) :-
	keysort(List, -1, S, []), !,
	Sorted = S.
ksort(X, Y):-user_error('illegal_arguments',keysort(X,Y)).

keysort([Head|Tail], Lim, Sorted, Rest) :- !,
	nonvar(Head),
	Head = _-_,
	Qh = [Head|_],
	samkeyrun(Tail, Qh, Qh, Run, Rest0),
	keysort(Rest0, 1, Lim, Run, Sorted, Rest).
keysort(Rest, _, [], Rest).

keysort([Head|Tail], J, Lim, Run0, Sorted, Rest) :-
	J =\= Lim, !,
	nonvar(Head),
	Head = _-_,
	Qh = [Head|_],
	samkeyrun(Tail, Qh, Qh, Run1, Rest0),
	keysort(Rest0, 1, J, Run1, Run2, Rest1),
	keymerge(Run0, Run2, Run),
	K is J+J,
	keysort(Rest1, K, Lim, Run, Sorted, Rest).
keysort(Rest, _, _, Sorted, Sorted, Rest).

samkeyrun([Hd|Tail], QH, QT, Run, Rest) :-
	nonvar(Hd),
	Hd = H-_,
	QT = [Q-_|QT2], 
	Q @=< H, !,
	QT2 = [Hd|_],
	samkeyrun(Tail, QH, QT2, Run, Rest).
samkeyrun([Hd|Tail], QH, QT, Run, Rest) :-
	nonvar(Hd),
	Hd = H-_,
	QH = [Q-_|_],
	H @< Q, !,
	samkeyrun(Tail, [Hd|QH], QT, Run, Rest).
samkeyrun(Rest, Run, [_], Run, Rest).

% keymerge(+List, +List, -List).
keymerge([], L2, Out) :- !,
	Out = L2.
keymerge([H1|T1], L2, Out) :-	
	L2 = [K2-_|_],
	H1 = K1-_,
	K1 @=< K2, !,
	Out = [H1|Out1],
	keymerge(T1, L2, Out1).
keymerge(L1, [H2|L2], Out) :- !,
	Out = [H2|Out1],
	keymerge(L1, L2, Out1).
keymerge(List, _, List).

% -------------------- ---------------------

%   File   : SETOF.PL
%   Author : R.A.O'Keefe
%   Updated: 17 November 1983
%   Purpose: define setof/3, bagof/3, findall/3, and findall/4
%   Needs  : Not.Pl

% Adapted by Paul Tarau for BinProlog: uses a heap based findall/3.
% therefore some database hacking predicates have been removed.
% Updated: 19 July 1992

/*  This file defines two predicates which act like setof/3 and bagof/3.
    I have seen the code for these routines in Dec-10 and in C-Prolog,
    but I no longer recall it, and this code was independently derived
    in 1982 by me and me alone.

    Most of the complication comes from trying to cope with free variables
    in the Filter; these definitions actually enumerate all the solutions,
    then group together those with the same bindings for the free variables.
    There must be a better way of doing this.  I do not claim any virtue for
    this code other than the virtue of working.  In fact there is a subtle
    bug: if setof/bagof occurs as a data structure in the Generator it will
    be mistaken for a call, and free variables treated wrongly.  Given the
    current nature of Prolog, there is no way of telling a call from a data
    structure, and since nested calls are FAR more likely than use as a
    data structure, we just put up with the latter being wrong.  The same
    applies to negation.

    Would anyone incorporating this in their Prolog system please credit
    both me and David Warren;  he thought up the definitions, and my
    implementation may owe more to subconscious memory of his than I like
    to think.  At least this ought to put a stop to fraudulent claims to
    having bagof, by replacing them with genuine claims.

    Thanks to Dave Bowen for pointing out an amazingly obscure bug: if
    the Template was a variable and the Generator never bound it at all
    you got a very strange answer!  Now fixed, at a price.

	bagof/3,		%   Like bagof (Dec-10 manual p52)
	setof/3.		%   Like setof (Dec-10 manual p51)

:- mode
	bagof(+,+,?),
	concordant_subset(+,+,-),
	concordant_subset(+,+,-,-),
	concordant_subset(+,+,+,+,-),
	replace_key_variables(+,+,+),
	setof(+,+,?).

%   setof(Template, Generator, Set)
%   finds the Set of instances of the Template satisfying the Generator..
%   The set is in ascending order (see compare/3 for a definition of
%   this order) without duplicates, and is non-empty.  If there are
%   no solutions, setof fails.  setof may succeed more than one way,
%   binding free variables in the Generator to different values.  This
%   predicate is defined on p51 of the Dec-10 Prolog manual.
*/

setof(Template, Filter, Set) :-
	bagof(Template, Filter, Bag),
	sort(Bag, Set).


%   bagof(Template, Generator, Bag)
%   finds all the instances of the Template produced by the Generator,
%   and returns them in the Bag in they order in which they were found.
%   If the Generator contains free variables which are not bound in the
%   Template, it assumes that this is like any other Prolog question
%   and that you want bindings for those variables.  (You can tell it
%   not to bother by using existential quantifiers.)
%   bagof records three things under the key '.':
%	the end-of-bag marker	       -
%	terms with no free variables   -Term
%	terms with free variables   Key-Term
%   The key '.' was chosen on the grounds that most people are unlikely
%   to realise that you can use it at all, another good key might be ''.
%   The original data base is restored after this call, so that setof
%   and bagof can be nested.  If the Generator smashes the data base
%   you are asking for trouble and will probably get it.
%   The second clause is basically just findall, which of course works in
%   the common case when there are no free variables.

bagof(Template, Generator, Bag) :-
	free_variables(Generator, Template, [], Vars,1),
	Vars \== [],
	!,
	Key =.. [.|Vars],
	functor(Key, ., N),
	findall(Key-Template,Generator,Recorded),
	replace_instance(Recorded, Key, N, OmniumGatherum),
	keysort(OmniumGatherum, Gamut), !,
	concordant_subset(Gamut, Key, Answer),
	Bag = Answer.
bagof(Template, Generator, [B|Bag]) :-
	findall(Template,Generator,[B|Bag]).

_^Goal:-Goal.

replace_instance([], _, _, []) :- !.
replace_instance([NewKey-Term|Xs], Key, NVars, [NewKey-Term|OldBag]) :-
		replace_key_variables(NVars, Key, NewKey), !,
		replace_instance(Xs,Key, NVars, OldBag).

replace_key_variables(0, _, _) :- !.
replace_key_variables(N, OldKey, NewKey) :-
	arg(N, NewKey, Arg),
	nonvar(Arg), !,
	M is N-1,
	replace_key_variables(M, OldKey, NewKey).
replace_key_variables(N, OldKey, NewKey) :-
	arg(N, OldKey, OldVar),
	arg(N, NewKey, OldVar),
	M is N-1,
	replace_key_variables(M, OldKey, NewKey).

%   concordant_subset([Key-Val list], Key, [Val list]).
%   takes a list of Key-Val pairs which has been keysorted to bring
%   all the identical keys together, and enumerates each different
%   Key and the corresponding lists of values.

concordant_subset([Key-Val|Rest], Clavis, Answer) :-
	concordant_subset(Rest, Key, List, More),
	concordant_subset(More, Key, [Val|List], Clavis, Answer).


%   concordant_subset(Rest, Key, List, More)
%   strips off all the Key-Val pairs from the from of Rest,
%   putting the Val elements into List, and returning the
%   left-over pairs, if any, as More.

concordant_subset([Key-Val|Rest], Clavis, [Val|List], More) :-
	Key == Clavis,
	!,
	concordant_subset(Rest, Clavis, List, More).
concordant_subset(More, _, [], More).


%   concordant_subset/5 tries the current subset, and if that
%   doesn't work if backs up and tries the next subset.  The
%   first clause is there to save a choice point when this is
%   the last possible subset.

concordant_subset([],   Key, Subset, Key, Subset) :- !.
concordant_subset(_,    Key, Subset, Key, Subset).
concordant_subset(More, _,   _,   Clavis, Answer) :-
	concordant_subset(More, Clavis, Answer).


% default behaviour changed: quantification is intended to be lexical and
% not dynamic as this code supposes. Strange things tend to happen
% otherwise, when, for instance, terms containing ^ are passed out of 
% their scope.
% -- Paul Tarau Thu Sep 14 10:45:09 ADT 1995

% 0 disables use of explicit_binding, 1 enables them
% setof stuff still uses 1, that's closer to it's usual implementation
free_variables(A,B,C,D) :- free_variables(A,B,C,D,0). 

% ---extracted from: not.pl --------------------%

%   Author : R.A.O'Keefe
%   Updated: 17 November 1983
%   Purpose: "suspicious" negation 

%   In order to handle variables properly, we have to find all the 
%   universally quantified variables in the Generator.  All variables
%   as yet unbound are universally quantified, unless
%	a)  they occur in the template
%	b)  they are bound by X^P, setof, or bagof
%   free_variables(Generator, Template, OldList, NewList,CheckBindings=0,1)
%   finds this set, using OldList as an accumulator.

free_variables(Term, Bound, VarList, [Term|VarList],_) :-
	var(Term),
	term_is_free_of(Bound, Term),
	list_is_free_of(VarList, Term),
	!.
free_variables(Term, _, VarList, VarList,_) :-
	var(Term),
	!.
free_variables(Term, Bound, OldList, NewList,B) :- B=:=1,
	explicit_binding(Term, Bound, NewTerm, NewBound),
	!,
	free_variables(NewTerm, NewBound, OldList, NewList,B).
free_variables(Term, Bound, OldList, NewList,B) :-
	functor(Term, _, N),
	free_variables(N, Term, Bound, OldList, NewList,B).

free_variables(0,    _,     _, VarList, VarList,_) :- !.
free_variables(N, Term, Bound, OldList, NewList,B) :-
	arg(N, Term, Argument),
	free_variables(Argument, Bound, OldList, MidList,B),
	M is N-1, !,
	free_variables(M, Term, Bound, MidList, NewList,B).

%   explicit_binding checks for goals known to existentially quantify
%   one or more variables.  In particular \+ is quite common.

explicit_binding(\+ _,	               Bound, fail,	Bound    ).
explicit_binding(not(_),	       Bound, fail,	Bound	 ).
explicit_binding(Var^Goal,	       Bound, Goal,	Bound+Var).
explicit_binding(setof(Var,Goal,Set),  Bound, Goal-Set, Bound+Var).
explicit_binding(bagof(Var,Goal,Bag),  Bound, Goal-Bag, Bound+Var).

term_is_free_of(Term, Var) :-
	var(Term), !,
	Term \== Var.
term_is_free_of(Term, Var) :-
	functor(Term, _, N),
	term_is_free_of(N, Term, Var).

term_is_free_of(0, _, _) :- !.
term_is_free_of(N, Term, Var) :-
	arg(N, Term, Argument),
	term_is_free_of(Argument, Var),
	M is N-1, !,
	term_is_free_of(M, Term, Var).

list_is_free_of([], _).
list_is_free_of([Head|Tail], Var) :-
	Head \== Var,
	list_is_free_of(Tail, Var).

keysort(L,S):-ksort(L,S).


% for compatibility with various prologs

% SOUND negation -> replaced with (I think) an improved one

% REPLACED:
% not(X):-ground(X),!, \+ X.
% not(X):-user_error('should be ground',not(X)).

% I see no reason to prohibit free variables when the negation succeeds
% as no future bindings can change the logical meaning in this case.
% the case of failure is (of course) different

not(X):- \+ X,!.
not(X):-ground(X),!,fail.
not(X):-user_error('should be ground',not(X))->fail;abort.

vars_of(Term,Vars):-free_variables(Term,[],[],Vars,0).


% Linear and intuitionistic implication and assumption

% intuitionistic implication - stack (asserta) ordering
C=>>G :- 
  assume(push(intuitionistic(Scope)),C),  % reusable
  metacall(G),
  Scope='$closed'.

% intuitionistic implication - assertz ordering
C=>G :- 
  assume(intuitionistic(Scope),C),  % reusable
  metacall(G),
  Scope='$closed'.

% linear implication - stack (asserta) ordering
(C-::G) :- 
  assume(push(linear(Scope)),C), % usable_once
  metacall(G),
  Scope='$closed'. % weakening: unused is OK

% linear implication - assertz ordering
(C-:G) :- 
  assume(linear(Scope),C), % usable_once
  metacall(G),
  Scope='$closed'. % weakening: unused is OK

% linear implication - asserta ordering, quite close to "-o" in Lolli
(C=::G) :-
  assume(push(linear(Scope)),C), % usable_once
  metacall(G),
  Scope=='$closed'. % no weakening: should be consumed!

% linear implication - assertz ordering
(C=:G) :- 
  assume(linear(Scope),C), % usable_once
  metacall(G),
  Scope=='$closed'. % no weakening: should be consumed!


% semantics: Clause implies the current continuation, i.e.
%            it will be true until the end (success or failure)
%            of the current resolution branch.

% Depending on the type of implication get 2 versions of assume:
%   intuitionistic: ~~ 
%        assume_ll(intuitionistic(_),Clause,Cont)::-Clause=>Cont.
%   linear: ~~
%        assume_ll(linear(_),Clause,Cont)::- Clause-:Cont.

% defining

assumel(X):-assume(linear(_),X).
assumei(X):-assume(intuitionistic(_),X).

assumeal(X):-assume(push(linear(_)),X).
assumeai(X):-assume(push(intuitionistic(_)),X).

assume(_,C):-var(C),!,errmes(error_in_assume,bad_clause(C)).
assume(Hint,C):-assume1(Hint,C).

% stack ordering makes no sense for files
assume1(Hint,[F]):-!,assume_file(Hint,F).
assume1(Hint,C):-assume_ll(Hint,C).

assume_ll(Hint,C):-
  add_true_ll(C,H,NewC),
  lval('$assumed',H,PXsYs),
  assume_with_hint_ll(Hint,PXsYs,NewC).

assume_with_hint_ll(push(Hint),PXsYs,NewC):-!,
  assume_with_hint_ll0(PXsYs,cls(Hint,NewC)).
assume_with_hint_ll(Hint,PXsYs,NewC):-
  assume_with_hint_ll1(PXsYs,cls(Hint,NewC)).

assume_with_hint_ll0(PXsYs,NewC):-
  nonvar(PXsYs),PXsYs=pred(Xs-Ys),!,
  setarg(1,PXsYs,[NewC|Xs]-Ys).
assume_with_hint_ll0(pred([NewC|Cs]-Cs),NewC).

assume_with_hint_ll1(PXsYs,NewC):-
  nonvar(PXsYs),PXsYs=pred(Xs-[NewC|Ys]),!,
  setarg(1,PXsYs,Xs-Ys).
assume_with_hint_ll1(pred([NewC|Cs]-Cs),NewC).

add_true_ll((H:-B),H1,C):-!,H1=H,C=(H:-B).
add_true_ll(H, H,(H:-true)).


% using

assumed(H):-assumed_clause(H,B),call_body(B).

assumed_clause(H,B):-assumed_clause(H,B,_).

assumed_clause(H,B,Hint):-
  val('$assumed',H,pred(Cs-_)),nonvar(Cs),
  get_assumed_clause(H,Cs,B+Hint).


get_assumed_clause(G,[cls(Hint,C)|_],NewG+Hint):-
  copy_or_delete_ll(Hint,C,G,NewG).
get_assumed_clause(G,[_|Ys],NewGHint):-
  nonvar(Ys),
  get_assumed_clause(G,Ys,NewGHint).


/*
get_assumed_clause(G,Cs,NewGHint):-
  member_scan(cls(_, (G :- _ ) ),Cs,[cls(Hint,C)|Ys]),
  get_assumed_clause1(Hint,C,G,NewGHint,Ys).

get_assumed_clause1(Hint,C,G,NewG+Hint,_):-
  copy_or_delete_ll(Hint,C,G,NewG).
get_assumed_clause1(_,_,G,NewGHint,Ys):-
  get_assumed_clause(G,Ys,NewGHint).
*/

copy_or_delete_ll(linear(Scope),C,G,NewG):-
  var(Scope),Scope='$closed',C=(G:-NewG).
copy_or_delete_ll(intuitionistic(Scope),C,G,NewG):-
  var(Scope),
  copy_term(C,(G:-NewG)).

% linear and intuitionistic file operations
assume_file(Hint,F):-
  find_file(F,InFile),
  seeing(F0),see(InFile),
  assume_all(Hint),
  seen,see(F0).

assume_all(Hint):-
  gc_read_clause(C),
  assume_all0(C,Hint).

assume_all0(end_of_file,_):-!.
assume_all0(C,Hint):-
  assume_one(C,Hint),
  gc_read_clause(NewC),
  assume_all0(NewC,Hint).

assume_one(':-'([F]),Hint):-!,assume_file(Hint,F).
assume_one(':-'(G),_):-!,(consult_cmd(G)->true;true).
assume_one(C,Hint):-assume_ll(Hint,C).

% gets info without touching
% although in a predicate different types of clauses can be assumed
% this will only return the type of the first clause

assumed_info(H,Info):-
   Flag='$not_closed',
   member(Hint,[intuitionistic(Flag),linear(Flag)]),
   assumed_info0(H,Hint),
   functor(Hint,Info,_).

assumed_info0(H,Hint):-
    val('$assumed',H,PCs),nonvar(PCs),PCs=pred(Cs-_),
    member_scan(cls(Hint,(H:-_)),Cs,_).

% Linear implication (-:: in `asserta' order) based
% exception handling constructs (see LOPSTR'94 paper by Tarau & Dahl 
% for a backtrackable version of catch/throw)

topcall(G):-catch(metacall(G),E,handle_uncaught(E)).

handle_uncaught('$commit'):-!,topcall(true).
handle_uncaught(E):-errmes(uncaught_exception,E).

commit :-
  assumed(catchmarker('$commit',Do,Choice,_)),
  cut_to(Choice), Do.

/*
% ISO catch and throw by Bart Demoen

catch(Goal,Ball,Do)

	execute Goal
	on throw(Term): look for closest catch
			copy Term
			undo bindings until call to catch; remove choices
			unify with Ball
			  if fail: throw Ball again
			  if succeed: call Do, continuation of catch goal

*/


catch(Goal,Ball,Do,Cont) ::- catch0(Goal,Ball,Do,Cont,Cont) .
catch(_,_,_) :- fail .

catch0(Goal,Ball,Do,Cont) :- get_neck_cut(Choice) ,
		(catchmarker(Ball,Do,Choice,Cont) -:: Goal) .

throw(Term) :-	copy_term(Term,Copied) ,
		assumed(catchmarker(Ball1,Do1,Choice,Cont1)),!,
		Ball1 = Ball ,
		Do1 = Do,
		Cont1 = Cont ,
		untrail_to(Choice) ,
		(Ball = Copied -> Do,call_cont(Cont) ; throw(Term)) .


call_cont(NewCont,_) ::- true(NewCont).

swap_cont(NewCont,OldCont,OldCont) ::- true(NewCont).



% A special instance of Delphi Lemmas:
% is ON THE USER to ensure that ProbAndArgs = Prob / Arglist
% contains a Prob in 0..100 (memoing probability) and a list of
% argument position which will be ground when F/N is called

delphi(F/N):-!,delphi0(F/N-10).
delphi(FND):-delphi0(FND).

delphi0(F/N-D):-
  ( vget0(bbhi,int(0))->nogc,nobbgc % disables garbage collection
  ; !,errmes(high_blackboard,memoing_disabled)
  ),
  functor(P,F,N),
  ( integer(D)->ProbAndArgs=D
    ; D=Prob/_,integer(Prob)->ProbAndArgs=D
  ),!,
  bb_let(P,'$delphi',ProbAndArgs)
; errmes('bad memo declaration',F/N-D).

% moved to lib.pl
%is_delphi(P,DP):-val(P,'$delphi',DP).

make_delphi_call(G,100/Args,memo_call(G,Args)):-!.
make_delphi_call(G,Prob/Args,delphi_call(G,Prob,Args)):-!.
make_delphi_call(G,100,memo_call_1(G,X)):-!,arg(1,G,X).
make_delphi_call(G,Prob,delphi_call_1(G,Prob,X)):-arg(1,G,X).

memo(F/N):-delphi0(F/N-100).
memo(F/N-Args):-delphi0(F/N-100/Args).

delphi_call(P,_,Args):-delphi_key(Args,P,K),membq(K,P,P),!.
delphi_call(P,Delphi,Args):-
  random(R),Luck is R mod 100,Delphi>Luck,!,
  P,!,
  delphi_key(Args,P,K),addq(K,P,P).
delphi_call(P,_,_):-P,!.

delphi_call_1(P,_,X):-membq(X,P,P),!.
delphi_call_1(P,Delphi,X):-
  random(R),Luck is R mod 100,Delphi>Luck,!,
  P,!,
  addq(X,P,P).
delphi_call_1(P,_,_):-P,!.

term_hash(Term,IndexArgList,Key):-
   hkey(Term,K),
   delphi_key1(IndexArgList,Term,K,Key).

delphi_key(Args,P,K):-delphi_key1(Args,P,0,K).

delphi_key1([],_,R,R).
delphi_key1([I|Is],P,R1,R3):-
  arg(I,P,X),
  hkey(X,K),
  R2 is (R1<<4) + K,
  delphi_key1(Is,P,R2,R3). 

memo_call(P,Args):-delphi_key(Args,P,K),membq(K,P,P),!.
memo_call(P,Args):-P,!,delphi_key(Args,P,K),addq(K,P,P).

memo_call_1(P,X):-membq(X,P,P),!.
memo_call_1(P,X):-P,!,addq(X,P,P).


% ASSERTED AND ASSUMED CODE LISTERS

listing(catchmarker,_):-!. % hide this
listing(P,N):-functor(H,P,N),is_assumed(H),!,
	write('% assumed '),write(P/N),write(':'),nl,
	listing1(H).
listing(P,N):-functor(H,P,N),
  (
    is_compiled(H)->write('% dyn_compiled: ')
%  ; is_asserted(H)->write('% dynamic: ') % this is useless!
  ; is_asserted(H)->write('% ') 
  ; fail
  ),
  write(P/N),write(':'),nl,
  listing0(H),
  !.
listing(_,_).

listing1(H):-
  assumed_clause(H,B,Hint),functor(Hint,F,_),
  write('/*'),write(F),write('*/ '),
  list_clause((H:-B)),
  fail
; nl.

list_clause(C):-portray_clause(C).

listing(F/N):-!,
  listing(F,N),nl.
listing(P):-
  listing2(P).

listing2(P):-
  for(N,0,255),
  listing(P,N),
  fail.
listing2(_):-nl.

generate_run_time_predicate(P):-
   generate_run_time_predicate(F,N),
   functor(P,F,N).

generate_run_time_predicate(P,N):-db_pred(P,N).
generate_run_time_predicate(P,N):-
	bb_list(Xs),
        K1='$assumed'/0,K2=P/N,
	bb_element(K1+K2=V,Xs),nonvar(V).

listing:-
        generate_run_time_predicate(P,N),
	listing(P,N),
	fail.
listing.

listing0(H):-
  clause0(H,B),pp_clause((H:-B)),fail
; nl.


help(Name):-apropos(Name).

apropos(Name):-
  findall(FN,apropos(Name,FN),FNs),
  sort(FNs,Xs),
  member(X,Xs),
    write(X),nl,
  fail.
apropos(_).

apropos(Name,F/N-Prop):-
  name(Name,UnKnown),
  predicate_property(H,Prop),functor(H,F,N),
  name(F,Known),
  near_match(Known,UnKnown).

near_match(Known,Known):-!.
near_match(Known,UnKnown):-append3(_,UnKnown,_,Known),!.
near_match(Known,UnKnown):-append3(A,B,C,Known),append3(A,[_|B],C,UnKnown),!.
%near_match(Known,UnKnown):-near_match(UnKnown,Known).

append3(Xs,Ys,Zs,Us):-
  append(Xs,YsZs,Us),
  append(Ys,Zs,YsZs).

% -------------------------------------------------------------------------
%   File   : INTERP
%   Author : R.A.O'Keefe
%   Updated: 2 March 84
%   Purpose: Meta-circular interpreter for Prolog (tracing version)


% Tracing metainterpreter

trace(Goal) :- tr_body(Goal, 0).

tr_body(Body,_):-var(Body),!,user_error(unbound_goal,Body).
tr_body(Body, Depth) :-
	tr_body(Body, Depth, AfterCut, HadCut),
	(   HadCut = yes,
		!,
		tab(Depth), write('CUT'), nl,
		tr_body(AfterCut, Depth)
	;   HadCut = no
	).

tr_body((Conj1,Conj2), Depth, AfterCut, HadCut) :- !,
	tr_body(Conj1, Conj2, Depth, AfterCut, HadCut).
tr_body(!, _, true, yes) :- !.
tr_body((Disj1;Disj2), Depth, AfterCut, HadCut) :-!,
	tr_disj1(Disj1, Disj2, Depth, AfterCut, HadCut).
tr_body(true, _, true, no) :- !.
tr_body(Goal, Depth, true, no) :- tr_goal(Goal, Depth).

tr_disj1((If->Then),Else, Depth, AfterCut, HadCut):-!,
	tr_if1(If,Then, Else, Depth, AfterCut, HadCut).
tr_disj1(Disj1,_, Depth, AfterCut, HadCut):-
	tr_body(Disj1, Depth, AfterCut, HadCut).
tr_disj1(_,Disj2, Depth, AfterCut, HadCut) :-
	tr_body(Disj2, Depth, AfterCut, HadCut).

tr_if1(If,Then, Else, Depth, AfterCut, HadCut):-
  call_body(If)->tr_body(Then, Depth, AfterCut, HadCut)
;              tr_body(Else, Depth, AfterCut, HadCut).

tr_body(!, AfterCut, _, AfterCut, yes) :- !.
tr_body((A,B), Conj, Depth, AfterCut, HadCut) :- !,
	tr_body(A, (B,Conj), Depth, AfterCut, HadCut).
tr_body((Disj1;Disj2), Conj, Depth, AfterCut, HadCut) :- !,
        tr_disj2(Disj1,Disj2, Conj, Depth, AfterCut, HadCut).
tr_body(true, Body, Depth, AfterCut, HadCut) :- !,
	tr_body(Body, Depth, AfterCut, HadCut).
tr_body(Goal, Body, Depth, AfterCut, HadCut) :-
	tr_goal(Goal, Depth),
	tr_body(Body, Depth, AfterCut, HadCut).

tr_disj2((If->Then),Else, Conj, Depth, AfterCut, HadCut) :-!,
        tr_if2(If,Then, Else, Conj, Depth, AfterCut, HadCut).
tr_disj2(Disj1,_, Conj, Depth, AfterCut, HadCut) :-
	tr_body(Disj1, Conj, Depth, AfterCut, HadCut).
tr_disj2(_,Disj2, Conj, Depth, AfterCut, HadCut) :-
	tr_body(Disj2, Conj, Depth, AfterCut, HadCut).

tr_if2(If,Then, Else, Conj, Depth, AfterCut, HadCut):-
  call_body(If)->tr_body(Then, Conj, Depth, AfterCut, HadCut)
;              tr_body(Else, Conj, Depth, AfterCut, HadCut).

tr_goal(call(Goal), Depth) :- !,tr_body(Goal, Depth).
tr_goal(\+(Goal), Depth) :- tr_body(Goal, Depth), !, fail.
tr_goal(\+(_), _) :- !.
tr_goal(Goal, Depth) :-
	(  tab(Depth), write('Call: '), tprint(Goal), fail
        ;  is_interactive, write(' <ENTER=call, ;=trace, h=help>: '), fail
        ;  true
	),
        (  is_interactive->get_code(X),(X=:=10->true;get_code(_))
        ;  X=32,nl
        ),
        ( X=:=10,\+ spying(Goal)->!,
                   Goal
        ; [X]="l"->functor(Goal,F,N),nl,listing(F/N),!,
                   tr_call(Goal, Depth)
        ; member(X,"aq")->nl,abort
        ; [X]="p"->nl,(topstep('GOAL:> ')->nl;nl),!,
                   tr_call(Goal, Depth)
        ; member(X,"h?")->nl,trace_help,!,tr_call(Goal, Depth)
        ; [X]="t"->!,nl
        ; [X]="f"->!,nl,fail
        ; [X]="k"->!,nl,assumel('$saved_goal'(Goal,Depth)),
                   tr_call(Goal, Depth)
        ; [X]="s"->!,show_saved_goals,
                   tr_call(Goal, Depth)
        ; /* ; etc. */ Depth1 is 1+Depth,tr_call(Goal, Depth1)
	),
	( tab(Depth), write('Exit: '), tprint(Goal), nl, fail
	; true
	; tab(Depth), write('Redo: '), tprint(Goal), nl, fail
	)
; 	tab(Depth), write('Fail: '), tprint(Goal), nl, fail.


show_saved_goals:-
   assumed('$saved_goal'(Goal,Depth)),
   write([Depth]),write(': '),write(Goal),nl,
   fail
;  nl.

trace_help:-
  nl,
  member(C-Mes,
   [
     'ENTER'-'call without tracing',
     l-listing,
     'q,a'-abort,
     p-'toplevel Prolog query',
     t-'succeed, but do not call this goal',
     f-'fail, and do not call this goal',
     k-'keep goal for futher inspection',
     s-'show saved goals instances',
     h-help,
     (';')-'continue (default)'
   ]
  ),
  write(C),write(' ==> '),write(Mes),nl,
  fail
; nl.

tprint(G):-trim_term(5,G,NewG),print(NewG),fail.
tprint(_).

templ_do_goal(Goal,true,'@'(Goal)):-is_verb(Goal),!.
templ_do_goal(Goal,Body,assumed_clause(Goal,Body)):-is_assumed(Goal),!.
templ_do_goal(Goal,Body,clause(Goal,Body)):-is_asserted(Goal).

tr_call(bagof(X,Y,Z), Depth) :- !,	% include these 4 lines if you
	bagof(X, tr_body(Y,Depth), Z).	% really want them, but they do
tr_call(setof(X,Y,Z), Depth) :- !,	% slow things down a bit.
	setof(X, tr_body(Y,Depth), Z).
tr_call(findall(X,Y,Z), Depth) :- !,
	findall(X, tr_body(Y,Depth), Z).
tr_call(Goal,Depth):-
        templ_do_goal(Goal,Body,Clause),
        !,
	functor(Clause,Mes,_),
	tab(Depth),functor(Goal,F,N),
        write('!!! '),write(Mes),write(': '),write(F/N),nl,
	Clause,
	tr_body(Body, Depth, AfterCut, HadCut),
	(   HadCut = yes,
		!,
		tab(Depth), write('CUT'), nl,
		tr_body(AfterCut, Depth)
	;   HadCut = no
	).
tr_call(Goal, Depth) :-
	is_compiled(Goal), % <--- checks for compiled predicate
	!, tab(Depth),functor(Goal,F,N),write('!!! compiled'(F/N)),nl,
	Goal.
tr_call(G, Depth):-functor(G,call,_),!,
	do_apply(G,Goal),
	tr_call(Goal,Depth).
tr_call(Undef):-do_undef(Undef).

debug(F):- reconsult(F).

% :-[net].

/*
* Socket based operations with Java comptible Linda client/server - 
*/

% defaults: 
%  overridable with assumptions/assertions/compiled facts

  default_host(X):-default(host(X),detect_host(X)).

  set_host(Server):-set_default(host(Server)).

  default_port(X):-default(port(X),X=9001).
  set_port(Port):-
    integer(Port),
    set_default(port(Port)).

  default_server_port(P):-vget0(server_port,int(P)),P>1,!.
  default_server_port(P):-default(server_port(P),default_port(P)).

  set_server_port(Port):-
    integer(Port),
    vset(server_port,int(Port)).

  default_code(X):-default(code(X),X='$empty').

  set_code(RemoteFile):-
    set_default(code(RemoteFile)).

  default_client_action(X):-
    default(client_action(X),X=term_client_action).

  set_client_action(R):-
    set_default(client_action(R)).

  default_server_action(X):-
    default(server_action(X),X=term_server_action).

  set_server_action(W):-
    set_default(server_action(W)).

  default_server_interactor(X):-
    default(server_interactor(X),X=term_server_interactor).

  set_server_interactor(W):-
    set_default(server_interactor(W)).

  default_timeout(X):-
    default(timeout(X),X=0).

  set_timeout(T):-
    set_default(timeout(T)).

  get_id(X):-bb_val(id,no,X),!.
  get_id(0).

  set_id(X):-bb_let(id,no,X).

  default_login(X):-
    default(login(X),detect_user(X)).

  set_login(X):-
    set_default(login(X)).

  default_password(X):-
    default(password(X),X=P),
    random(N),
    symcat(pw,N,P),
    set_password(P).
    
  set_password(P):-
    set_default(password(P)).

  default_master_server(H,P):-
    default( master_server(H,P),
            (H='clement.info.umoncton.ca',P=9000)
          ).

  set_master_server(Host,Port):-
    set_default(master_server(Host,Port)).

  default_proxy_server(H,P):-
    default( proxy_server(H,P),
            % default_master_server(H,P)
            fail
          ).

  set_proxy_server(Host,Port):-
    set_default(proxy_server(Host,Port)).

% interface

  run_server:-run_server(_). % runs server on the first free port

  run_default_server:-default_server_port(Port),  run_server(Port).

  fork_server:-fork_server(_).

  fork_server(Pid):-default_server_port(Port),  fork_server(Port,Pid).

  stop_server:-stop_server(abracadabra).
 
% generic client/server predicates

fork_server(Port,Pid):-
    (free_port(Port)->true;errmes('port busy',Port)),
    unix_fork(Pid),
    ( Pid=:=0->
      quietmes(1,'forking with pid'(Pid)),
      quiet(Q),quiet(10),
                run_server(Port),               % child: a daemon
      quiet(Q),
      quietmes(1,'forked server done, pid'(Pid)),
      halt
    ; sleep(1)                                  % parent
    ).

free_port(Port):-new_server(Port,NewServer),close_socket(NewServer).

new_server_on_free_port(Port,NewServer):-
  nonvar(Port),!,
  new_server(Port,NewServer).
new_server_on_free_port(FreePort,NewServer):-
  MaxTry=64,
  ( new_server_on_free_port(MaxTry,FreePort,NewServer)->
    quietmes(2,new_server_on_free_port(FreePort))
  ; errmes('unable to find free port for server', tried(MaxTry,times))
  ).

new_server_on_free_port(MaxTry,FreePort,NewServer):-
  default_server_port(P),Max is MaxTry-1,
  quietmes(1,default_server_port(P)),
  ( FreePort=P
  ; for(_,1,Max),random(I),FreePort is P+1+(I mod 998)
  ),
  quietmes(1,trying_server_on_port(FreePort)),
  new_server(FreePort,NewServer),
  !.

run_server(Port):-
  Fuel is 1<<28,
  vset(err_id,int(77)),
  new_server_on_free_port(Port,NewServer),
  register_server(Port),default_password(Pwd),
  quietmes(2, running_server(port(Port), password(Pwd))),
    server(NewServer)=>>server_loop(NewServer,Fuel),
  close_socket(NewServer),
  quietmes(2, server_done(port(Port), password(Pwd))),
  vset(err_id,int(0)).

server_loop(Server,Fuel):-
  default_password(W),
  for(I,1,Fuel),
    quietmes(1,'serving socket, fuel'(Server,I)),
    interact(Server),   
    assumed(server_done(W)),
  !,
  quietmes(1,server_done(server=Server,fuel=I)).

new_service(Server,Service):-
  default_timeout(T),
  new_service(Server,T,Service).

interact(Server):-
  % blocks here until one of a max of 64 waiting clients connects
  default_server_action(Interactor),
  ( call(Interactor,Server)->do_postponed
  ; \+ errmes(failed_interaction(Server),Interactor)
  ),
  !.
interact(Server):-
  \+ errmes(bad_interaction,server(Server)).

do_postponed:-assumed(todo(G)),G,!,do_postponed.
do_postponed.

new_client(Connection):-
  default_host(Host),
  default_port(Port),
  new_client(Host,Port,Connection).

ask_server(Question,Answer):-
  quietmes(1,ask_server(Question)),
  assumed(to_all(ServerPattern)),!,
  all_servers(ServerPattern,Xs),
  ask_all_servers(Xs,Question),
  Answer=yes.
ask_server(Question,Answer):-
  ask_a_server(Question,Answer).

ask_all_servers(ListOfServers,Question):-
  quietmes(1,ask_all_servers(ListOfServers,Question)),
  member(server(_,H,P),ListOfServers),
  host(H)=>>port(P)=>>ask_a_server(Question,_),
  fail.
ask_all_servers(_,_).


ask_a_server(Question,Answer):-
  default_host(Host),
  default_port(Port),
  default_client_action(Interactor),
  ask_a_server(Interactor,Host,Port,Question,Answer).

ask_a_server(Host,Port,Question,Answer):-
  default_client_action(Interactor),
  ask_a_server(Interactor,Host,Port,Question,Answer).

ask_a_server(Interactor,Host,Port,Question,Answer):-
  default_proxy_server(ProxyH,ProxyP),!,
  (ProxyH==Host,ProxyP==Port->
    \+ errmes(bad_proxy_same_as_host(ProxyP,ProxyP),asking(Question)),
    R=no
  ; directly_ask_a_server(
      Interactor,ProxyH,ProxyP,proxy(Host,Port,Question),R
    )
  ),
  Answer=R.
ask_a_server(Interactor,Host,Port,Question,Answer):-
  directly_ask_a_server(Interactor,Host,Port,Question,Answer).

directly_ask_a_server(Interactor,Host,Port,Question,Answer):-
  call(Interactor,Host,Port,Question,Result),
  !,
  Answer=Result.
directly_ask_a_server(_,_,_,stop(_),yes):-!.
directly_ask_a_server(_,_,_,halt(_),yes):-!.
directly_ask_a_server(Interactor,H,P,Q,no):-
  \+ errmes(bad_interaction_with(Interactor,H,P),ask_a_server(Q)),
  ( non_local_host, not_master_server, assumed(to_all(_)) ->
    quietmes(1,asking_master_to_check(H,P)),
    refresh_servers(server_id(_,H,P))
  ; true
  ).

not_master_server:-
  \+ assumed(server_interactor(master_server_interactor)).

% helpers for Linda server operations

serve_cin(X,R):-nonvar(X),retract1(X),!,R=X.
serve_cin(X,R):-var(X),clause(X,true),retract1(X),!,R=X.
serve_cin(_,no).

serve_rd(X,R):-clause(X,true),!,R=X.
serve_rd(_,no).

serve_out(no,no):-!.
serve_out(yes,no):-!.
serve_out(X,yes):-assert(X).

serve_cout(X,R):-serve_rd(X,no)->serve_out(X,R)
               ; R=no.

%serve_collect(X,G,Xs):-nonvar(G),!,findall(X,G,Xs).
%serve_collect(X,G,Xs):-serve_facts(X,G,Xs).

serve_facts(X,G,Xs):-findall(X,clause(G,true),Xs).

serve_halt(W,yes):-
  default_password(W0),W==W0,
  assumed(socket(S)),
  sock_write(S,"yes"),!,
  halt.
serve_halt(_,no).

this_id(R):-
  default_host(H),default_server_port(P),
  this_id(H,P,R).

this_id(H,P,server_id(InfoList,H,P)):-
   get_id(No),
   default_login(U),default_password(W),
   % default_host(HN),unix_pid(Pid),
   %,default_server_interactor(I),default_server_action(A),
   InfoList=[server(No),login(U),password(W),
   %host(HN),pid(Pid),
   %,interactor(I),action(A),
   language(bp)
   ].

% pretty secure, Linda, proxy and chat-only server


term_server_interactor(cin(X),R):-serve_cin(X,R).
term_server_interactor(out(X),R):-serve_out(X,R).
term_server_interactor(cout(X),R):-serve_cout(X,R).
term_server_interactor(rd(X),R):-serve_rd(X,R).
term_server_interactor(all(X),Xs):-serve_facts(X,X,Xs).
term_server_interactor(all(X,G),Xs):-serve_facts(X,G,Xs).
term_server_interactor(id(R),R):-this_id(R).
term_server_interactor(ping(T),T):-ctime(T).
term_server_interactor(stop(W),yes):-assumel(server_done(W)).
term_server_interactor(halt(X),R):-serve_halt(X,R).
term_server_interactor(add_servant(X),R):-serve_add_servant(X,R).
term_server_interactor(mes(From,Cs),Answer):-
  show_mes(From,Cs,Answer),
  forward_to_servants(mes(From,Cs)).
term_server_interactor(proxy(H,P,Q),R):-
  ask_a_server(H,P,Q,A)->R=A;R=no.
term_server_interactor(in(X),R):-serve_cin(X,R). % $$thish should block!!!

proxy(Host,Port,Question,Answer):-
  ask_a_server(proxy(Host,Port,Question),Answer).

show_mes(From,Cs,yes):-quiet(Q),Q<5,!,
  write(From),write('> '),write_chars(Cs),nl.
show_mes(_,_,no).

forward_to_servants(Query):-
   clause(servant_id(Id),true),
   assert(query(Id,Query)),
   fail.
forward_to_servants(_).

term_server_action(Server):-
  new_service(Server,Socket),
  socket(Socket)=>>term_server_step(Socket),
  close_socket(Socket).

term_server_step(Socket):-
  server_try(Socket,sock_read(Socket,QuestionChars)),
  question2answer(QuestionChars,AnswerChars),
  server_try(Socket,sock_write(Socket,AnswerChars)).

question2answer(QuestionChars,AnswerChars):-
  term_chars(Question,QuestionChars),
  quietmes(1,question(Question)),
  default_server_interactor(Filter),
  call(Filter,Question,Answer),!,
  term_chars(Answer,AnswerChars),
  quietmes(1,answer(Answer)),
  !.
question2answer(QuestionChars,"no"):-
  (term_chars(QuestionChars,Question)->true;Question=bad_question),
  \+ errmes(failed_or_unauthorized_server_operation,Question).

server_try(_,Goal):-Goal,!.
server_try(Socket,Goal):-
  close_socket(Socket),
  default_server_action(Action),
  errmes(server_socket_error_in(Action),Goal).

% BinProlog Linda client operations
term_client_action(Host,Port,Question,Answer):-
  new_client(Host,Port,Socket),
  socket(Socket)=>>term_client_step(Socket,Question,Answer),!,
  close_socket(Socket).

term_client_step(Socket,Question,Answer):-
  term_chars(Question,QuestionChars),
  client_try(Socket,sock_write(Socket,QuestionChars)),
  client_try(Socket,sock_read(Socket,AnswerChars)),
  term_chars(Answer,AnswerChars),
  quietmes(1,term_client_step(Socket,Question,Answer)),
  !.
term_client_step(_,_,no).

client_try(_,Goal):-Goal,!.
client_try(Socket,Goal):-
  close_socket(Socket),
  default_client_action(Action),
  errmes(client_socket_error_in(Action),Goal).



% Linda clients for BinProlog and Java Linda servers
% -- to be merged once both servers will provide similar functions

% Linda operations common to Java and BinProlog servers

out(X):-ask_server(out(X),yes).
cout(X):-ask_server(cout(X),_).
all(X,Xs):-ask_server(all(X),Xs).
rd(X):-ask_server(rd(X),R),!,R\==no,X=R.
cin(X):-ask_server(cin(X),R),!,R\==no,X=R.
in(X):-cin(X)->true;wait_for_in(8,X). % to be redone with pthreads!
id(X):-ask_server(id(X),R),R\==no,X=R.

% BinProlog specific

all(X,G,Xs):-ask_server(all(X,G),Xs).

wait_for_in(Max,X):-for(I,1,Max),sleep(I),cin(X),!.

ping:-ping(_).

ping(Time):-
   quiet(Q), quiet(10),
   (ask_a_server(ping(_),R),R\==no->Ok=true;Ok=fail),
   quiet(Q),
   Ok,
   Time=R.

% currently only on BinProlog servers

remote_run(G):-ask_server(run(G),G).            % runs G on server!
remote_run(X,G):-ask_server(run(X,G),X).

remote_mes(From,Mes,Answer):-
  listify(Mes,Chars),
  ask_server(mes(From,Chars),Answer).

say(Mes):-
  default_login(From),
  remote_mes(From,Mes,Answer),
  Answer==yes,
  !.
say(Mes):-
 chat_err(Mes).

chat_err(Mes):-
  listify(Mes,Cs),
  term_chars(T,Cs),
  errmes(lost_message,T).

yell(Mes):-to_all(_)=>>say(Mes).

stop_server(X):-ask_server(stop(X),yes).

halt_server:-halt_server(abracadabra).

halt_server(X):-ask_server(halt(X),yes).

rsh(Cmd,Chars):-remote_run(pcollect(Cmd,Chars)).

rsh(Cmd):-rsh(Cmd,Xs),(member(X,Xs),put(X),fail;true).

rexec(Cmd):-rexec(Cmd,Xs),(member(X,Xs),put(X),fail;true).

rexec(Cmd0,Cs):-
   Temp='cmd.tmp',
   make_cmd([Cmd0,'>',Temp],Cmd),
   remote_run(system(Cmd)),
   remote_run(Cs,file2chars(Temp,Cs)).

non_local_host:-default_host(Host),Host\==localhost.

detect_user(U):-unix_getenv('USER',U),!.
detect_user(guest).

detect_host(R):-ip_addr(H),H\=='127.0.0.1',!,R=H.
detect_host(localhost).

fetch_and_call(H):-fetch_pred_for(H),asserted(H).

fetch_pred_for(H):- 
   \+ \+ fetch_pred_from_remote_code(H).

fetch_pred_from_remote_code(H):-is_dynamic(H),!.
fetch_pred_from_remote_code(H):-
  quietmes(1,fetch_pred_from_remote_code(H)),
  default_code(File),
  functor(H,F,N),functor(H0,F,N),
  fetch_file(File,File,H0,user),
  is_asserted(H0).

fetch_file(File):-fetch_file(File,File,_,user).

% loads to LocalDB all clauses matching H from default remote server 
% which will mirror in RemoteDB on the server the RemoteFile
% however if RemoteFile is a unknown it will try to pick them from
% RemoteDB where RemoteFile is expected to be preloaded
fetch_file(RemoteFile,RemoteDB,H,LocalDB):-
   quietmes(1,entering-fetch_file(RemoteFile,RemoteDB,H,LocalDB)),
   ( var(RemoteFile)->true
   ; remote_run(RemoteDB,db_mirror(RemoteFile,RemoteDB))
   ),
   % fetch_remote_operators,
   start_remote_engine(db_clause(RemoteDB,H,B),(H:-B),E),
   repeat,
     ( ask_remote_engine(E,Clause)->
       db_assert(LocalDB,Clause),fail
     ; true
     ),
   !,
   stop_remote_engine(E),
   quietmes(1,finished(file=RemoteFile,engine=E)).

fetch_remote_operators:-   
   remote_run(Ops,findall(op(X,Y,Z),current_op(X,Y,Z),Ops)),
   forall(member(Op,Ops),Op).

start_remote_engine(G,X,E):-
   heap_size(Hx),stack_size(Sx),trail_size(Tx),
   remote_run(E,create_engine(Hx,Sx,Tx,E)),
   quietmes(1,created_remote_engine(E)),
   remote_run(load_engine(E,G,X)),
   quietmes(1,loaded_remote_engine(E,G,X)).

ask_remote_engine(E,X):-
   remote_run(ask_engine(E,X)).

stop_remote_engine(E):-
   remote_run(destroy_engine(E)),
   quietmes(1,stopped_remote_engine(E)).

db_mirror(File,DB):- 
   ( asserted(mirroring(File,DB))->true
   ; db_clean(DB),
     assert(mirroring(File,DB)),
     consult(File,DB)
   ).

% remote file get/put operations

rget(File):-rget(File,File).

rget(RemoteFile,LocalFile):-rget(256,RemoteFile,LocalFile).

rget(BufSize,RemoteFile,LocalFile):-
   quietmes(1,entering_rget(RemoteFile,LocalFile)),
   remote_run(RS,fopen(RemoteFile,'rb',RS)),
   fopen(LocalFile,'wb',LS),
   repeat,
     remote_run(NotRead-Cs,fget_chars(RS,BufSize,NotRead,Cs)),
     %write(got(NotRead,Cs)),nl,
     fput_chars(LS,Cs),
     NotRead>0,
   !,
   remote_run(fclose(RS)),
   fclose(LS),
   quietmes(1,finished_rget(remote=RemoteFile,local=LocalFile,rs=RS)).

rput(File):-rput(File,File).

rput(LocalFile,RemoteFile):-rput(256,LocalFile,RemoteFile).

rput(BufSize,LocalFile,RemoteFile):-
   quietmes(1,entering_rput(LocalFile,RemoteFile)),
   remote_run(RS,fopen(RemoteFile,'wb',RS)),
   fopen(LocalFile,'rb',LS),
   repeat,
     fget_chars(LS,BufSize,NotRead,Cs),
     %write(sent(NotRead,Cs)),nl,
     remote_run(fput_chars(RS,Cs)),
     NotRead>0,
   !,
   remote_run(fclose(RS)),
   fclose(LS),
   quietmes(1,finished_rput(local=LocalFile,remote=RemoteFile,rs=RS)).

fput_chars(File,Cs):-nonvar(Cs),!,fput_chars0(File,Cs).
fput_chars(File,Cs):-errmes(bad_list,fput_chars(File,Cs)).

fput_chars0(File,Cs):-
  member(C,Cs),
  fputc(File,C),
  fail.
fput_chars0(_,_).

fget_chars(File,BufSize,PlacesNotFilled,Chars):-
   fget_chars0(BufSize,PlacesNotFilled,File,Chars).

fget_chars0(N,N2,F,[C|Cs]):-
  N>0,N1 is N-1,fgetc(F,C),C>=0,!,
  fget_chars0(N1,N2,F,Cs).
fget_chars0(N,N,_,[]).

% direct C-based fast file transfer

fget(File):-fget(File,File).

fget(RemoteFile,LocalFile):-
   quietmes(1,entering_fget(RemoteFile,LocalFile)),
   remote_run(RF,fopen(RemoteFile,'rb',RF)),
   fopen(LocalFile,'wb',LF),
   term_chars(to_sock(RF),Cmd),
   new_client(Socket),
     sock_write(Socket,Cmd),
     socket(Socket)=>>from_sock(LF,_),
     fclose(LF),
     sock_read(Socket,_),
   close_socket(Socket),
   remote_run(fclose(RF)),
   quietmes(1,finished_fget(remote=RemoteFile,local=LocalFile)).


fput(File):-fput(File,File).

fput(LocalFile,RemoteFile):-
   quietmes(1,entering_fput(LocalFile,RemoteFile)),
   remote_run(RF,fopen(RemoteFile,'wb',RF)),
   fopen(LocalFile,'rb',LF),
   term_chars(from_sock(RF),Cmd),
   new_client(Socket),
     sock_write(Socket,Cmd),
     socket(Socket)=>>to_sock(LF,_),
     fclose(LF),
     sock_read(Socket,_),
   close_socket(Socket),
   remote_run(fclose(RF)),
   quietmes(1,finished_fget(local=LocalFile,remote=RemoteFile)).

to_sock(F,yes):-
   assumed(socket(S)),
   file2sock(F,S),!,
   quietmes(1,to_sock(file=F,socket=S)).
to_sock(F,no):-
   quietmes(1,failing(to_sock(F))).

from_sock(F,yes):-
   assumed(socket(S)),
   sock2file(S,F),!,
   quietmes(1,to_sock(file=F,socket=S)).
from_sock(F,no):-
   quietmes(1,failing(from_sock(F))).

cache_file(F,Suf,CachedF):-
  CacheDir='cache',
  (exists_file(CacheDir)->true
  ; namecat(mkdir,' ',CacheDir,Cmd),
    system(Cmd,X),
    ( X==0->true
    ; errmes(unable_to_do,Cmd)
    )
  ),
  namecat(F,'.',Suf,File),
  ( remote_run(Size,file_size(File,Size))->
    namecat(CacheDir,'/',File,CachedF),
    ( file_size(File,Size)->true % this is not 100% reliable but is simple
    ; fget(File,CachedF)
    )
  ; errmes(remote_file_not_found,File)
  ).

% remote code operations

rconsult(F):- %$$
  cache_file(F,'pl',CF),
  reconsult(CF).

rcompile(F):-
  cache_file(F,'pl',CF),
  compile(CF).

rload(F):-
  cache_file(F,'wam',CF),
  load(CF).

% master server: hub keeping track of various 
% BinProlog servers alive on the net

run_local_master_server:-
  default_master_server(_,P),
  default_host(H),
  master_server(H,P)=>>server_port(P)=>>run_master_server.

run_master_server:-run_master_server(run_default_server).

fork_master_server:-run_master_server(fork_server).

run_master_server(ServerGoal):-
  default_master_server(H,P),
  default_host(ThisH),detect_host(LH),
  ( (ThisH==H;ThisH==LH)->
      host(H)=>>
       server_port(P)=>>
        server_interactor(master_server_interactor)=>>
    ServerGoal,
    !
  ; errmes(unable_to_run_master_server(H,P), on(host=ThisH,port=P))
  ).

master_server_interactor(refresh(X),yes):-serve_facts(X,X,Xs),
  assumel(todo(cleanup_dead_servers(Xs))).
master_server_interactor(halt(N),R):-
  kill_all_servers(N,_),
  term_server_interactor(halt(N),R).
master_server_interactor(add(X),R):-register_on_master(X,R).
master_server_interactor(all(X),Xs):-serve_facts(X,X,Xs).
master_server_interactor(all(X,G),Xs):-serve_facts(X,G,Xs).
master_server_interactor(ping(T),R):-
  term_server_interactor(ping(T),R).

register_on_master(Id,R):-
  Id=server_id([server(_)|Xs],H,P),
  gensym_no(client_no,N),
  NewId=server_id([server(N)|Xs],H,P),
  retractall(server_id(_,H,P)),
  serve_out(NewId,_),
  !,
  R=N.
register_on_master(_,no).

kill_all_servers(Pword,X):-
  serve_facts(X,X,Xs),
  member(server_id(_,H,P),Xs),
  ask_a_server(H,P,halt(Pword),_),
  fail.
kill_all_servers(_,_).

cleanup_dead_servers(Xs):-
   member(X,Xs),
   remove_dead(X),
   fail.
cleanup_dead_servers(_).

remove_dead(X):-X=server_id(_,H,P),
   host(H)=>>port(P)=>>ping(_),!.     % just check if alive
remove_dead(X):-
   retract1(X).                   % remove if not

register_server:-
   default_server_port(Port),
   register_server(Port).

register_server(Port):-register_server(Port,Answer),Answer\==no,!.
register_server(Port):- 
   default_host(Host),
   default_master_server(H,P),
   \+ errmes(
       unable_to_register_server_on_master_server(H,P),
       server(Host,Port)
   ).

register_server(Port,Answer):-
   default_host(Host),
   register_server(Host,Port,Answer).

register_server(localhost,_,yes):-!,
   default_master_server(H,P),
   make_cmd(["use ?-host('my.own.host.name')=>run_server.",[10],
   "to register on master server at ", H, ":", P],Mes),
   quietmes(2,Mes).
register_server(Host,Port,R):-default_master_server(Host,Port),!,R=yes.
register_server(Host,Port,Answer):-!,
   this_id(Host,Port,Id), % i,i,o
   quietmes(1,registering(Host,Port)),
   %quiet(Q), quiet(9),
   ( ask_master_server(add(Id),R) -> Answer=R
   ; Answer=no
   ),
   %quiet(Q),
   ( Answer=no->true
   ; set_id(Answer) 
   )
   .
  
% master server interface:

ask_master_server(Query,Answer):-
   default_master_server(H,P),
   default_client_action(A),
   directly_ask_a_server(A,H,P,Query,Answer).

all_servers(H,Xs):-
   ask_master_server(all(server(N,H,P),server_id([server(N)|_],H,P)),Xs).

all_servers(Xs):-all_servers(_,Xs).

all_servers:-all_servers(Xs),forall(member(X,Xs),quietmes(5,X)).

refresh_servers:-refresh_servers(_).


refresh_servers(X):-ask_master_server(refresh(X),yes).


% unrestricted server: use only on trusted Intranet

run_unrestricted_server:-
  server_interactor(unrestricted_server_interactor)=>>
  run_server.

unrestricted_server_interactor(to_sock(F),Ok):-!,to_sock(F,Ok).
unrestricted_server_interactor(from_sock(F),Ok):-!,from_sock(F,Ok).
unrestricted_server_interactor(run(G),R):-G->!,R=G;R=no.
unrestricted_server_interactor(run(X,G),R):-!,G->R=X;R=no.
unrestricted_server_interactor(Q,A):-term_server_interactor(Q,A).

% executes queries obeying to sever at default host, port

register_servant(R):-
  get_id(Id),
  ask_server(add_servant(Id),A),
  A\==no,
  !,
  R=A.
register_servant(_):-
  errmes(unable_to_register,servant).
  

run_servant:-
  vset(err_id,int(78)), % makes this aboort on ^C
  default_server_interactor(Interactor),
  ( register_servant(No)->
    quietmes(5,'SERVANTS ALIVE ON THE NET'),all_servants,
    servant_loop(No,Interactor),
    in(servant_id(No))
  ; true
  ),
  vset(err_id,int(0)).


servant_loop(No,Interactor):-
  repeat,
    in(query(No,Query)),
    ( call(Interactor,Query,Reply)->Answer=Reply
    ; Answer=no
    ),
    ( Query=mes(_,_)->true
    ; out(answer(No,Answer))
    ),
    quietmes(1,servant(Query,Answer)),
  functor(Query,stop,_), 
  !.

serve_add_servant(N,Id):-N=:=0,
   get_id(ServerId),
   gensym_no(servant_no,ServantNo),
   Id=servant(ServerId,ServantNo),
   serve_out(servant_id(Id),_),
   !.
serve_add_servant(Id,Id).

ask_servant(ID,Query,Answer):-
   out(query(ID,Query)),
   in(answer(ID,R)),
   Answer=R.

all_servants:-all_servants(Xs),forall(member(X,Xs),quietmes(5,X)).

all_servants(Xs):-all(X,servant_id(X),R),R=Xs.

% chat client - broadcasts chat input to registered servers

chat:- ip_addr(H),host(H)=>>chat(_).

chat(Target):- 
  refresh_servers, 
  quietmes(5,'TARGET SERVERS ALIVE ON THE NET:'),
  (quiet(Q),Q<5->all_servers;true),
  quietmes(5,'^D on Unix or ^Z on PCs to end'),
  repeat,    
    ( write('>'),read_chars(Cs)->to_all(Target)=>>say(Cs),fail
    ; true
    ),
  !.

% talks to a given server and its servants

talk:-
  quietmes(5,'^D on Unix or ^Z on PCs to end'),
  repeat,    
    ( write('>'),read_chars(Cs)->say(Cs),fail
    ; true
    ),
  !.

% requires unrestricted server on target

rshell:-
  quietmes(5,'^D on Unix or ^Z on PCs to end'),
  default_host(H),
  repeat,    
    ( write(H),write('>'),
       read_chars(Cs),term_chars(Cmd,Cs)->rexec(Cmd),fail
    ; true
    ),
  !.
