% High-level compiling and consulting and restart/halt operations

[File]:-include(File).

include(File):-
   vget0(load_method,int(I)),
   load_method(I,Method),
   call(Method,File).

load_method(1,mcompile).
load_method(2,scompile).
load_method(3,oconsult).
load_method(4,dconsult).
load_method(5,sconsult).

'~'(F) :- reconsult(F).

compile(File):-apply_load_method(compile,File).

reconsult(File):-apply_load_method(reconsult,File).

apply_load_method(Type,File):-
  vget0(Type,M/0),
  load_method(I,M),
  !,
  vset(load_method,int(I)),
  call(M,File).

% compile and loads to memory
mcompile(File):-
  mcompile_file(File),
  abort.

fcompile(F):-stat_time(fcompile_file(F)).

load(F):-
  survive_cleanup(F,NewF),
  stat_time(load_file(NewF)),
  abort.

scompile(F):- \+ is_prolog(binprolog),mcompile(F).
scompile(F):-
  survive_cleanup(F,NewF),
  init_cmd,
  stat_time(scompile_file(NewF)),
  abort.

pc:-push_code(_).

end_of_file:-halt.

abort :- abort0.

restart :- restart0.

% Operting System (mostly Unix) interface

shell(C):-system(C).

system(Cmd):-system(Cmd,_).

system([],0):-!.
system([X|Xs],Ret):-!,name(Command,[X|Xs]),system0(Command,Ret).
system(Command,Ret):-atomic(Command),system0(Command,Ret).

cd:-unix_cd.

cd(X):-unix_cd(X).

pwd:-system(pwd).

pwd(Cs):-pcollect(pwd,Cs).

unix_cd:-
   ( unix_getenv('HOME',HomeDir)->true
   ; HomeDir=(/)
   ),
   unix_cd(HomeDir).

unix_argv(Xs):-
  unix_argc(Max),
  Last is Max-1, 
  % with unix_argv(0,X) you can also get the program's name
  findall(X,(for(I,1,Last),unix_argv(I,X)),Xs).


unix(argc(X)):-unix_argc(X).
unix(argv(Xs)):-unix_argv(Xs).
unix(argv(N,X)):-unix_argv(N,X).
unix(getenv(X,Y)):-
  unix_getenv(X,Y)->true
; errmes('no such environment variable',X).
unix(access(X,Y)):-unix_access(X,Y).
unix(cd(X)):-unix_cd(X).
unix(cd):-unix_cd.
unix(kill(X,Y)):-unix_kill(X,Y).
unix(popen(X,Y,Z)):-popen(X,Y,Z).
unix(pclose(X)):-pclose(X).
unix(pcollect(X,Y)):-pcollect(X,Y). % collects a pipe to a list
unix(fork(X)):-unix_fork(X). % forks
unix(pid(X)):-unix_pid(X). % forks
unix(cat(X)):-unix_cat(X). % prints a file

% pushes code into the BinProlog kernel
% the code is protected and it will not be cleaned-up by :-[...].

ls:-system('ls -tF').
dir:-system(dir).

unix_cat(FName):-
   fopen(FName,'rb',F),
     fsize(F,N),for(I,1,N),fgetc(F,C),put_code(C),I=:=N,!,
   fclose(F).
   
file2chars(FName,Cs):-
   fopen(FName,'rb',F),
   fgetc(F,C),fcollect(C,F,Cs),
   fclose(F).

fcollect(-1,_,[]):-!.
fcollect(X,F,[X|Xs]):-
   fgetc(F,NewX),
   fcollect(NewX,F,Xs).

edit(_,user):-!,include(user).
edit(Editor,File):-
        find_file0(File,F),
        make_cmd([Editor," ",F],R),
        system(R),
        include(F).

get_editor(_,Ed):-(unix_getenv('EDITOR',Ed);unix_getenv('VISUAL',Ed)),!.
get_editor(Default,Default).

current_user_file(F):-vget0(current_user_file,F/0).

set_current_user_file(File):-vset(current_user_file,File/0).

my_edit(DefaultEditor):-
   current_user_file(File),
   edit(DefaultEditor,File).

my_defedit(DefaultEditor):-
   current_user_file(File),
   defedit(DefaultEditor,File).

defedit(DefaultEditor,File):-
   get_editor(DefaultEditor,Ed),
   edit(Ed,File).

ed:-my_defedit(emacs).

edit:-my_defedit(edit).

textedit:-my_edit(textedit).
emacs:-my_edit(emacs).
notepad:-my_edit(notepad).
pico:-my_edit(pico).
vi:-my_edit(vi).

co:-
   current_user_file(File),
   include(File).
        
stat:-statistics.

% some code for multiple engines
clean_up_engines:-list_engines(Xs),member(X,Xs),
   ( destroy_engine(X)->true
      ,debugmes(destroyed_engine(X))
   ;  debugmes(failed_to_destroy_engine(X))
   ),
   fail.
clean_up_engines.

show_engine:-
  show_area(global_stack),
  show_area(local_stack),
  show_area(trail).
  
show_area(Area):-
  statistics(Area,[H1,H2]),
  H is H1+H2,
  debugmes(max(Area)=H),
  fail.
show_area(_).

% LOW-LEVEL ACCESS TO DATA `shared' with C (run-time system)

% relies on HDEFS,HDFI in ru.c

vread(CVar,Val):-val(bp_state,CVar,Val).

heap_size(X):-default(heap(X),vread(heap_size,X)).
stack_size(X):-default(stack(X),vread(stack_size,X)).
trail_size(X):-default(trail(X),vread(trail_size,X)).

hostname(X):-vread(hostname,X).
ip_addr(X):-vread(ip_addr,X).

% relies on VSHARE in ru.c
% using them on compound objects needs knowing what's going on in C

vget(CVar,Val):-val(bp_state,CVar,A),array_get(A,0,Val).
vget0(CVar,Val):-val(bp_state,CVar,A),array_get0(A,0,Val).

vset(CVar,Val):-val(bp_state,CVar,A),array_set(A,0,Val).

gvset(CVar,Val):-gval(Val,GroundVal),vset(CVar,GroundVal).

gvget(CVar,GroundVal):-vget(CVar,Val),atomize(Val,GroundVal).

atomize(var(Adr),Val):-!,array_get(Adr,0,Val).
atomize(Val,Val).

set_bp_error(Id,Mes,Arg1,Arg2):-
  vset(err_id,int(Id)),
  gvset(err_mes,Mes),
  gvset(err_arg1,Arg1),
  gvset(err_arg2,Arg2),
  vset(err_wam,var(0)),
  !.
set_bp_error(Id,Mes,_Arg1,_Arg2):-
  errmes(error_in_error_handler,set_bp_error(Id,Mes)).

clear_bp_error:-
  set_bp_error(0,'$empty','$empty','$empty').

get_bp_error(Id,Mes,Arg1,Arg2):-
  vget0(err_id,int(Id)),
  vget0(err_mes,Mes/0),
  vget0(err_arg1,Arg1),
  vget0(err_arg2,Arg2),!.
get_bp_error(Id,Mes,_Arg1,_Arg2):-
  errmes(error_in_error_handler,get_bp_error(Id,Mes)).

% GC options, enable, desable etc.

gc:-vset(gc,int(1)).

nogc:-vset(gc,int(0)).

gc_status(X):-vget0(gc,int(X)).

dynbbgc:-vset(bbgc,int(2)).

bbgc:-vset(bbgc,int(1)).

nobbgc:-vset(bbgc,int(0)).

bbgc_status(X):-vget0(bbgc,int(X)).

% control on system messages `quietness'

quiet(X):-integer(X),!,vset(quiet,int(X)).
quiet(X):-vget0(quiet,int(X)).

is_bbvar(X):-var(X),gval(X,var(AX)),
   vget0(bboard,var(B)),vget0(bboard_margin,var(M)),
   AX>=B,AX<M.

is_heapvar(X):-var(X),gval(X,var(AX)),
   vget0(heap,var(B)),vget0(heap_margin,var(M)),
   AX>=B,AX<M.

create_engine(Engine):-
  heap_size(H),
  stack_size(S),
  trail_size(T),
  create_engine(H,S,T,Engine).

open_engine(Goal,X,Engine):-
  create_engine(Engine),
  load_engine(Engine,Goal,X).

call_engine(E,G,X):-
  load_engine(E,G,X),
  ask_engine(E,A),
  copy_term(A,X).
  

% -----------------------------------------------------------------
% LIBRARY of basic predicates

is_engine(E):-list_engines(Es),member(E,Es).

% expands to strange things in Sicstus: written in binary Prolog here

':'(M,P,Cont)::-module_call(M,P,Cont).

module_call(_,P) :- is_public(P),!,P.
module_call(M,P) :- module_predicate(M,P,MP), MP.

module_predicate(M,Pred,MPred):-
  functor(Pred,F,_),
  module_name(M,F,MF),
  term_append(MF,Pred,MPred).

module_name(M,P,MP):-
  nonvar(M),nonvar(P),!,
  namecat(M,':',P,MP).
module_name(M,P,_):-
  user_error(should_be_nonvar,[M,P]).

module(M):-begin_module(M).

begin_module(M):-
  bb_let(current,module_name,M),
  bb_let(module_name,M,M).

end_module:-begin_module(user).

end_module(M):-current_module(M),!,end_module.
end_module(M):-errmes(bad,end_module(M)).

module(M,FNs):-
  module(M),
  member(FN,FNs),
    public0(FN),
  fail.
module(_,_).

current_module(M):-bb_val(current, module_name , M0),!,M=M0.
current_module(user).

is_module(M):-nonvar(M),!,bb_val(module_name,M,_).
is_module(M):-bb_list(Xs),bb_element(module_name/0+M/0=_,Xs).

modules(Ms):-findall(M,is_module(M),Ms).

hide_atom(G,H):-hide_atom0(G,H),!.
hide_atom(G,G).

hide_atom0(_,_):-current_module(user),!,fail.
hide_atom0(Pred,_):-is_public(Pred),!,fail.
hide_atom0(Pred,MPred):-current_module(M),module_predicate(M,Pred,MPred).

public0(F/N):-
   functor(P,F,N),
   current_module(M),
   bb_def((public),P,M),
   !.
public0(Bad):-
   errmes('bad or repeated declaration',public(Bad)).

public((A,B)):-!,public(A),public(B).
public(A):-public0(A).

is_public(Pred):-is_public(Pred,_).

is_public(Pred,Module):-bb_val((public),Pred,M),!,Module=M.
is_public(Pred,_):-is_builtin(Pred).

init_gensym(Root):-bb_let(gensym,Root,0).

gensym(Root,Symbol):-gensym_no(Root,N),symcat(Root,N,Symbol).

gensym_no(Root,N1):-
        val(gensym,Root,N),!,
        N1 is N+1,
        set(gensym,Root,N1).
gensym_no(Root,N):-N=1,
        def(gensym,Root,N).

current_op(Pri,Assoc,Name):-nonvar(Name),!,
  get_op(Name,Assoc,Pri).
current_op(Pri,Assoc,Name):-
  for(I,0,10000),
  (val(opmark,I,Name)->get_op(Name,Assoc,Pri);!,fail).

op(Pri,Assoc,Name):-op0(Name,Assoc,Pri).

get_op(Name,Assoc,Pri):- % nonvar(Name),
	op_type(Assoc,Cls),
	get_op0(Name,Cls,Assoc,Pri).

get_op0(Name,Cls,Assoc,Pri):- % nonvar(Name), nonvar(Cls),
	val(Cls,Name,Assoc),
	val(Name,Cls,Pri),
	Pri>0.

op_type(xfy,infixop).
op_type(xfx,infixop).
op_type(yfx,infixop).

op_type(fx,prefixop).
op_type(fy,prefixop).

op_type(xf,postfixop).
op_type(yf,postfixop).

set_prop(X,Prop):-
  append_conj(X,true,FNs),
  member_conj(F/N,FNs),
    functor(P,F,N),
    bb_let(P,Prop,yes),
  fail.
set_prop(_,_).

is_multifile(P):-val(P,'$multifile',_).

multifile(X):-set_prop(X,'$multifile').
  
is_discontiguous(P):-is_multifile(P).

discontiguous(X):-multifile(X).

bp_info(F/N,I):-nonvar(F),nonvar(N),!,
  functor(H,F,N),N1 is N+1,functor(H1,F,N1),
  gen_bp_info(H,H1,I),
  !.
bp_info(F/N,I):-
  gen_bp_info(H,_,I),
  functor(H,F,N).

gen_bp_info(H,H1,I):-
  ( bu0(H1,_,_,I),functor(H1,F,N1),N is N1-1,functor(H,F,N)
  ; bu1(H,I)
  ).

bp_info(FN,Text,Exs):-
  bp_info(FN,I),
  (I=''->fail
  ;I=Text-Ex->make_exs(FN,Ex,Exs)
  ;Text=I,Exs=[]
  ).
  
make_exs(FN,Ex,Exs):-
  to_list(Ex,Xs),
  map(extract_ex(FN),Xs,Exs).

extract_ex(F/_,T0,Vs^T):-
  term_append(F,T0,T), 
  vars_of(T,Vs).

has_info(FN):-bp_info(FN,Text,_),Text\==''.

info:-info(_),fail;true.

info(FN):-nonvar(FN),FN=_/_,!,info0(FN).
info(F):-
   for(N,0,254),
     has_info(F/N),write(F/N),write(': '),info0(F/N),
   fail
;  true.

info0(FN):-
  show_info(name,FN,Text+Exs),
  show_info(text,Text),
  show_info(examples,Exs).

show_info(name,FN,Text+Exs):-
  ( var(FN)->Show=yes
  ; Show=no
  ),
  bp_info(FN,Text,Exs),
  ( Show=yes->write('PREDICATE: '),write(FN),nl
  ; Show=no
  ).
show_info(text,Txt):-
  write('INFO: '),write(Txt),nl,nl.
show_info(examples,Exs):-
  ( Exs=[]->true
  ; write('EXAMPLE(S):'),nl,nl,
      member(Vs^G,Exs),namevars0(Vs^G,Ns^G0),
      write('?-'),write(G0),write('.'),nl,
      ( G,
         show_bindings(Ns,Vs),nl,
        fail
      ; (Vs=[]->write(yes);write(no)),nl,nl
      ),
    fail
  ; true
  ).

show_bindings([],[]).
show_bindings([V|Vs],[A|As]):-
  write(V=A),write((;)),nl,
  show_bindings(Vs,As).

is_builtin_nonvar(T):-bu1(T,_),!.
is_builtin_nonvar(T):-term_append(T,cont(_),TC),bu0(TC,_,_,_).

is_builtin_var(T):-bu1(T,_).
is_builtin_var(T):-bu0(TC,_,_,_),strip_cont(TC,T,_).

is_builtin(T):-nonvar(T),!,is_builtin_nonvar(T).
is_builtin(T):-is_builtin_var(T).

interactive(X):-var(X),!,(is_interactive->X=yes;X=no).
interactive(X):-nonvar(X),interactive0(X).

is_interactive:- \+ current_op(1199,fx,('?-')).

interactive0(yes):-op(1200,fx,('?-')).
interactive0(no):-op(1199,fx,('?-')).

expand_term(C,E):-portable_expand_term(C,E).
std_expand_term(C,D):-portable_expand_term(C,D).

strip_cont(TC,T,C):- strip_cont0(TC,C-T).
% NASTY BUG IN BUILTIN strip_cont0... => SOLVED !!!

metatrue(TC)::-strip_cont(TC,T,C,metacall(T,C)).

metacall(G):-call_body(G).

bincall(BinCont,Cont,Cont) ::- BinCont.

call(FXs,Y1):-term_append(FXs,args(Y1),G),G.
call(FXs,Y1,Y2):-term_append(FXs,args(Y1,Y2),G),G.
call(FXs,Y1,Y2,Y3):-term_append(FXs,args(Y1,Y2,Y3),G),G.
call(FXs,Y1,Y2,Y3,Y4):-term_append(FXs,args(Y1,Y2,Y3,Y4),G),G.
call(FXs,Y1,Y2,Y3,Y4,Y5):-term_append(FXs,args(Y1,Y2,Y3,Y4,Y5),G),G.
call(FXs,Y1,Y2,Y3,Y4,Y5,Y6):-term_append(FXs,args(Y1,Y2,Y3,Y4,Y5,Y6),G),G.


override(P,NewP):-override(0,P,NewP). % backtrackable if 0, hard otherwise

once(G):-G,!.

map(F,Xs,Ys):-map0(Xs,F,Ys).

map0([],_,[]).
map0([X|Xs],F,[Y|Ys]):-
   call(F,X,Y),
   map0(Xs,F,Ys).
 
foldl(F,Z,Xs,R):-foldl0(Xs,F,Z,R).
  
foldl0([],_,R,R).
foldl0([X|Xs],F,R1,R3):-call(F,R1,X,R2),foldl0(Xs,F,R2,R3).

foldr(F,Z,Xs,R):-foldr0(Xs,F,Z,R).
  
foldr0([],_,Z,Z).
foldr0([X|Xs],F,Z,R2):-foldr0(Xs,F,Z,R1),call(F,X,R1,R2).

sum(Xs,R):-foldl(+,0,Xs,R).

prod(Xs,R):-foldl(*,1,Xs,R).

termcat(T,C,TC):-term_append(T,cont(C),TC).
 
% true.
% call(X):-X.

otherwise.

false:-fail.

call_ifdef(G,_):-is_compiled(G),!,G.
call_ifdef(_,Else):-Else.

callable(G):-is_compiled(G),!.
callable(G):-is_dynamic(G),!.
callable(G):-nonvar(G),functor(G,call,_).

default(G,_):-callable(G),G,!.
default(_,AltG):-AltG,!.

set_default(H):-
  functor(H,F,N),functor(H0,F,N),
    retractall(H0),
  assert(H).

% (A:-B):-assert((A:-B)).

X=X.

A->B :- A,!,B.

A->B ; C :- !,if(A,B,C).
X ; _ :-X.
_ ; Y :-Y.

or(A,_):-A.
or(_,B):-B.

if(A,B,_):-A,!,B.
if(_,_,C):-C.

(X,Y):-X,Y.

\+(X):-X,!,fail.
\+(_).

\=(X,X):-!,fail.
\=(_,_).

halt:-halt(0).

repeat.
repeat:-repeat.

forall(Goal):-Goal,fail.
forall(_).

forall(Goal,OtherGoal):- \+ ((Goal, \+ OtherGoal)).

findall_workhorse(X,G,Cons,AnswerSize,_):-
	lift_heap(Cons,AnswerSize), % cons,answer_size: 0=automatic
	dummycall(G,X) .
findall_workhorse(_,_,_,_,Xs):-
	findall_load_heap(Xs).	

dummycall(G,X) :- G , findall_store_heap(X).
dummycall(_,_) :- fail .

findall(X,G,Xs,End):-findall_workhorse(X,G,'.',0,[End|Xs]).

%findall(X,G,Xs):-findall_workhorse(X,G,'.',0,[[]|Xs]).

findall(X,G,Xs):-findall(X,G,Xs,[]).

findall_conj(X,G,Xs):-findall_conj(X,G,Xs,true).

findall_conj(X,G,Xs,End):-findall_workhorse(X,G,',',0,(End,Xs)).

findall_disj(X,G,Xs):-findall_disj(X,G,Xs,fail).

findall_disj(X,G,Xs,End):-findall_workhorse(X,G,';',0,(End;Xs)).

% findall(G,Gs):-findall(G,G,Gs).

% The following can together be used to give the illusion of having
% a parallel engine which produces answers lazily, as needed.
% Intended to be used as take and drop work in haskell,
% on possibly infinite streams of solutions,

% they use WAM-level means as change_arg/3 in has_fuel/3
% the reader is challenged to express them in classical Prolog :-)
% To do more, a separate Prolog engine or first order manipulation
% of OR continuations is needed...

% answers G as far as constraint C holds
while(C,G):-G,(C->true;!,fail).

skip_until(C,G):-G,(C->fail;!).

skip_when(C,G):-G,(C->fail;true).

% gives only the N-th answer of G
nth_answer(N,G):-N1 is N-1,Max=s(N1),skip_until(has_fuel(Max),G).

% generates at most the first N answers of G
take_at_most(N,G):-Max=s(N),while(has_fuel(Max),G).

% drops at least the first N answers of G
drop_at_least(N,G):-Max=s(N),skip_when(has_fuel(Max),G).

% re-entrant on-place counter
has_fuel(Max):-arg(1,Max,N),N>0,N1 is N-1,change_arg(1,Max,N1).

% answer_stream to list converters
find_while(C,X,G,Xs):-findall(X,while(C,G),Xs).

find_at_most(N,X,G,Xs):-findall(X,take_at_most(N,G),Xs).

all_but_at_least(N,X,G,Xs):-findall(X,drop_at_least(N,G),Xs).

det_call(G):-find_at_most(2,G,G,Gs),!,
  ( Gs=[]->fail
  ; Gs=[G]->true
  ; % member(G,Gs),
    errmes('expected to be deterministic',G)
  ).

% if V is not common to X and G then V cannot be bound by all/3
% => such variables should be shared
% this is _different_ from what bagof does, and I think, better !
% try: ?-all_answers(X,member(s(X),[A,B,B,C]),Xs).


all_answers(X,G,Xs):-
  are_shared(X,G,Vs),
  findall(X-Vs,G,Ys),
  share_vars(Ys,Vs,Xs).

are_shared(X,G,vars(Is,Os)):-
  free_variables(X,G,[],Is,0), % free variables in X not occurring in G
  free_variables(G,X,[],Os,0). % free variables in G not occurring in X

share_vars([],_,[]).
share_vars([T-Vs|TVs],Vs,[T|Ts]):-share_vars(TVs,Vs,Ts).

gc_read(R):-findall(X,read(X),[R]).

gc_read_clause(R):-findall(X,read_clause(X),[R]).

gc_call(G):-findall(G,G,Gs),member(G,Gs).

answer_of(G,X):-findall(X,G,Xs),sort(Xs,Sorted),member(X,Sorted).

% Mercury style: Closure+LastArg->ListOfLastArgsCollectedInSolutions

solutions(Closure,Xs):-
  term_append(Closure,lastarg(X),Goal),
  findall(X,Goal,Xs).

member_conj(C,(C,_)).
member_conj(C,(_,Cs)):- member_conj(C,Cs).

% souce level member/2 + for/3

member(C,[C|_]).
member(C,[_|Cs]):- member(C,Cs).

memberchk(X,Xs):-member_scan(X,Xs,[Y|_]),!,Y=X.

/*
for(Min,Min,Max):-Min=<Max.
for(I,Min,Max):-Min<Max,Min1 is Min+1,for(I,Min1,Max).

member(X,Xs) :- member_entry(X,Xs).
member(X,[Y|Ys]) :- member3(X,Ys,Y).

memberchk(C,[C|_]):-!.
memberchk(C,[_|Cs]):- memberchk(C,Cs).
*/

full_mem(X,[Y|Ys]) :- member3(X,Ys,Y).

member3(X,_,X) .
member3(X,[Y|Ys],_) :- member3(X,Ys,Y).

for(I,Min,Max):-for_entry(I,Min,Max).
for(I,Min,Max):-for1(I,Min,Max).

for1(Min,Min,Max):-Min=<Max.
for1(I,Min,Max):-Min<Max,Min1 is Min+1,for1(I,Min1,Max).


argn(I,T,X):-nonvar(I),!,arg(I,T,X).
argn(I,T,X):-functor(T,_,N),for(I,1,N),arg(I,T,X).

numbervars(VAR_V, N0, N) :- ##numbervar_name(VAR_V,N0),!, N is N0+1.
numbervars(X, N0, N) :- atomic(X), !, N0=N.
numbervars([X|Xs], N0, N) :- !,
        numbervars(X, N0, N1),
        numbervars(Xs, N1, N).
numbervars(X, N0, N) :-
        functor(X, _, A),
        numbervars(0, A, X, N0, N).

numbervars(A, A, _, N0, N) :- !, N0=N.
numbervars(A0, A, X, N0, N) :-
        A1 is A0+1,
        arg(A1, X, X1),
        numbervars(X1, N0, N1),
        numbervars(A1, A, X, N1, N).

ground(T):-numbervars(T,0,N),N>0,!,fail.
ground(_).

copy_term(T,CT):-copy_term(0,T,CT).
save_term(T,CT):-copy_term(1,T,CT).

clone_term(SharedVars,T,CT):-copy_term(SharedVars^T,SharedVars^CT).

flush:-seen_told(2).

see_or_fail(File):-see_tell(0,File).

exists_file(File):-see_or_fail(File),seen.

/* % OLD definitions
sread(I,O):-string_op(0,I,O).
swrite(I,O):-string_op(1,I,O).

*/

swrite0(I,O):-string_op(1,I,O).


% symbol <--> term conversion

sread(I,O):-sread(I,O,_).

sread(I,O,Vs):-name(I,Xs),read_term_from_chars(Xs,O,Vs).

swrite(I,O):-term_chars(I,Xs),name(O,Xs).

swrite(I,Vs,O):-write_term_to_chars(I,Vs,Xs),name(O,Xs).


/* TERM TO LIST OF CHARS CONVERSION */

/* OLD:
term_chars(T,Cs):-nonvar(T),!,swrite(T,S),name(S,Cs).
term_chars(T,Cs):-name(S,Cs),sread(S,T).
*/

% string of chars <--> term conversion

term_chars(T,Cs):-nonvar(T),!,t2s(T,Cs).
term_chars(T,Cs):-nonvar(Cs),!,s2t(Cs,T-_).
term_chars(_,"_"). % when both args are vars!

atom_chars(T,A):-term_chars(T,A). % compatibility with others

read_term_from_chars(Cs,T):-read_term_from_chars(Cs,T,_).

read_term_from_chars(Cs,T,Vs):-s2t(Cs,T-Vs).

write_term_to_chars(T,Cs):-write_term_to_chars(T,[],Cs).

write_term_to_chars(T,Vs,Xs):-
  findall(Xs,write_term_to_chars1(T,Vs,Xs),[Xs]).

write_term_to_chars1(T,Vs,Xs):-
   list2conj(Vs,G),G,
   term_chars(T,Xs).

list2conj([],true).
list2conj([X|Xs],(X,Ys)):-list2conj(Xs,Ys).

% term to string

t2s(T,S):-findall(S,do_t2s(T,S),[S]).

do_t2s(T,S):-
  override(nl,dcg_nl),
  override(put_code(_),dcg_put_code(_)),
  override(fast_write(_),dcg_cwrite(_)),
  numbervars(T,0,_),
  dcg_def(S),  % opens current dcg stream
    generic_write(T,writeq),
  dcg_val([]). % closes current dcg stream

% the only way to properly emulate write is with side effects
% as write itself fails (to save heap space after doing its work)
% however, generic_write/2 is clean 
% of side effects so grammar emulation is ok

% string to term

s2t(S,TVs):-nonvar(S)->findall(TVs,do_s2t(S,TVs),[TVs])
; errmes(bad_argument(S),'list of chars expected').

do_s2t(S,T-Vs):-
  End=[46,10], % "." + end-of-line
  override(get_code(_),dcg_get_code(_)),
  append(S,End,NewS),
  dcg_def(NewS), % opens current dcg stream
   % read(T),Vs=[],
    read_or_fail(T,Vs),
  dcg_val(LeftOver),   % closes current dcg stream
  %quietmes(left_over(LeftOver)),
  maybe_full_stop(LeftOver,End).

maybe_full_stop([Dot,Eol],[Dot,Eol]):-!.
maybe_full_stop([Dot,32],[Dot,_]):-!.
maybe_full_stop([],_):-!.

% tools using AGs (IO overrading with assumption grammars)

dcg_get_code(X):- '#'(X).

dcg_nl:- '#'(10).
dcg_put_code(X) :- '#'(X).

fix_float(X,S):-float(X),!,swrite0(X,S).
fix_float(X,X).

dcg_cwrite(X):- fix_float(X,S),name(S,Xs), dcg_write_name(Xs).

dcg_write_name([]).
dcg_write_name([X|Xs]):- '#'(X), dcg_write_name(Xs).

% string to tokens

read_tokens_from_chars(S,Ts,Vs):-
  nonvar(S)->findall(X,do_s2tokens(S,X),[Ts-Vs])
; errmes(read_tokens_from_chars,'arg 1 should be nonvar').

do_s2tokens(S,Ts-Vs):-
  End=[46,10], % "." + end-of-line
  override(get_code(_),dcg_get_code(_)),
  append(S,End,NewS),
  dcg_def(NewS), % opens current dcg stream
    read_tokens(Ts,Vs),
  dcg_val([]).   % closes current dcg stream

% FILE operations

file_search_path(["","myprogs/",PROGS,SRC]):-
  getenv_path("../src",SRC),
  getenv_path("../progs",PROGS).

file_extension_list([".pl",".pro",""]).

file_library([LIBRARY],[".pl"]):-getenv_path("../library",LIBRARY).


getenv_path(Default,Computed):-
  Slash="/",
  ( unix_getenv('BP_PATH',Where)->
    name(Where,Xs),det_append(Xs,Slash,RootPath)
  ; RootPath=""
  ),
  det_append(Default,Slash,Local),
  det_append(RootPath,Local,Computed).
  

make_file_name(Prefs,File,Sufs,NewFile):-
        member(Pref,Prefs),
        member(Suf,Sufs),
        name(File,L),
        det_append(L,Suf,R),
        det_append(Pref,R,Fname),
        name(NewFile,Fname).

statistics(Area,[Used,Free]):-stat_dict(Area,No),stat0(Used,Free,No).

statistics:-
	statistics(Name,Data),
	fast_write(Name),fast_write(=),fast_write(Data),nl,
	fail.
statistics.	


atom(X):-integer(X),!,fail.
atom(X):-atomic(X).

% float(X):-nonvar(X),X='$float'(_,_,_).

float(T):-nonvar(T),functor(T,'$float',3),arg(1,T,X),X>=0.

number(X):-integer(X),!.
number(X):-float(X).

compound(X):-nonvar(X), \+(atomic(X)).

=..(T,[X|Xs]):-var(T),!,list2term([X|Xs],T).
=..(T,L):-term2list(T,('.'),[],L).

term2list(T,L):-term2list(T,('.'),[],L).

term2dlist(T,Head,Queue):-term2list(T,('.'),Queue,Head).

append([],Ys,Ys).
append([A|Xs],Ys,[A|Zs]):-
        append(Xs,Ys,Zs).

/*
append([],Ys,Ys).
append([A|Xs],Ys,[A|Zs]):-
  det_append0(Xs,NewXs,EndXs-Zs),
  append(EndXs,Ys,NewXs).
*/

% det_append0 works with an arbitrary constructor
% it guesses it from the first constructor in Xs

det_append(Xs,Ys,Zs):-det_append0(Xs,Ys,_-Zs).

det_append(Xs,Ys,Zs,End):-det_append0(Xs,Ys,End-Zs).

appendN(Xss,Xs):-appendN(Xss,Xs,[]).

appendN([],End,End).
appendN([Xs|Xss],S1,S2):-det_append0(Xs,End,_-S1),appendN(Xss,End,S2).

append_body(C1,C2,C3):-append_body(C1,C2,C3,_).

append_body(C1,C2,C3,True):-
  functor(C1,F,2),
  det_append0(C1,T,End-C3),
  ( End==True->T=C2 
    ;          T=..[F,End,C2]
  ).

append_conj(C1,C2,C3):-append_body((X,C1),C2,(X,C3),true).

append_disj(C1,C2,C3):-append_body((X;C1),C2,(X;C3),fail).

% this allows using list-functions in is: Xs is [1,2]++[3]++[4,5]

'.'(X,Xs,[X|Xs]).

++(Xs,Ys,Zs) :- det_append0(Xs,Ys,[]-Zs). 

length(L,N):-var(N),!,get_length(L,0,N).
length(L,N):-make_length(L,0,N).

get_length([],I,I).
get_length([_|L],I0,I):-I1 is I0+1,get_length(L,I1,I).

make_length([],I,I):-!.
make_length([_|L],I0,I):-I0<I,I1 is I0+1,make_length(L,I1,I).

not_subsumes0(A,B):-copy_term(B,CB),numbervars(CB,0,_),CB=A,!,fail.
not_subsumes0(_,_).

subsumes_chk(A,B):-not_subsumes0(A,B),!,fail.
subsumes_chk(_,_).

variant_of(A,B):-subsumes_chk(A,B),subsumes_chk(B,A).

tab(Expr) :- expr(Expr,N),integer(N),!,tab0(N).
tab(Bad) :- user_error('in tab/1: should evaluate to an integer', Bad).

tab0(N):-for(_,1,N),put_code(32),fail.
tab0(_).

get(R):-repeat,get_code(X),(X>32;X<0),!,R=X.

% !! use something like Y=3+4, X is 1+expr(Y) instead of
% !! Y=3+4, X is 1+Y which will not work in compiled code

X is E:-expr(E,R),!,X=R.
X is E:-errmes(error_in_is,X is E).

expr(E,R):-atomic(E),!,R=E.
expr(E,R):-float(E),!,R=E.
expr(E,_Error):-var(E),!,user_error(variable_in_is,E).
expr(E,R):-functor(E,Op,2),!,
	arg(1,E,E1),arg(2,E,E2),
	expr(E1,X1),
	expr(E2,X2),
	G=..[Op,X1,X2,R],
	G,!.
expr(E,R):-functor(E,Op,1),
	arg(1,E,E1),
	expr(E1,X1),
	G=..[Op,X1,R],
	G,!.


compare(R,X,Y):-compare0(X,Y,R).

+(X,N):- +(0,X,N).
-(X,N):- -(0,X,N).
\(X,N):- \(0,X,N).

A==B :- compare0(A,B,=).
A\==B :- compare0(A,B,R),'$noteq'(R).

A @< B :- compare0(A,B,<).
A @> B :- compare0(A,B,>).
A @=< B :- compare0(A,B,R),'$lesseq'(R).
A @>= B :- compare0(A,B,R),'$gteq'(R).

A < B :-expr(A,X), expr(B,Y), less(X,Y).

A > B :-expr(A,X), expr(B,Y), greater(X,Y).

A =< B :-expr(A,X), expr(B,Y), less_eq(X,Y).

A >= B :-expr(A,X), expr(B,Y), greater_eq(X,Y).

A =:= B :-expr(A,X), expr(B,Y), arith_eq(X,Y).

A =\= B :-expr(A,X), expr(B,Y), arith_dif(X,Y).

'$lesseq'(<).
'$lesseq'(=).

'$gteq'(>).
'$gteq'(=).

'$noteq'(<).
'$noteq'(>).

% BinProlog USER LEVEL naming primitives
% DO NOT rely on def/3, set/3, val/3,rm/2. Thay are unsafe
% and their implementation may change in the future...

/********* BASIC operations: safe with respect to bb_gc ***********/

% WARNING: may be moved to C, do not use their componnents directly

bb_def(A,B,X):-saved(X,S),def(A,B,S).
bb_set(A,B,X):-saved(X,S),set(A,B,S).
bb_val(A,B,CX):-val(A,B,X),copy_term(X,CX).
bb_rm(A,B):-rm(A,B).

/******************DERIVED OPERATIONS ***********************/

% safe bboard operations
bb_let(A,B,X):-val(A,B,_),!,bb_set(A,B,X).
bb_let(A,B,X):-bb_def(A,B,X).

bb_def(A,X):-bb_def(A,A,X).
bb_set(A,X):-bb_set(A,A,X).
bb_val(A,X):-bb_val(A,A,X).
bb_rm(A):-bb_rm(A,A).
bb_let(A,X):-bb_let(A,A,X).

bb_default_val(A,B,_,X):-bb_val(A,B,X),!.
bb_default_val(_,_,D,D).

saved(X,S):-save_term(X,NewX),!,unify_to(NewX,S).
saved(X,S):-bb_gc,save_term(X,NewX),!,unify_to(NewX,S).
saved(X,S):-bb_fail(saved(X,S)).

% unsupported & unsafe

let(X,Y,V):-def(X,Y,V),!.
let(X,Y,V):-set(X,Y,V).

def(A,X):-def(A,A,X).
set(A,X):-set(A,A,X).
val(A,X):-val(A,A,X).
let(A,X):-let(A,A,X).
rm(A):-rm(A,A).

% INVISIBLE DCG connect operation: normally macro-expanded
% renamed as Hidden Accumulator Grammars

'#'(Word):-dcg_connect(Word).
'#<'(Xs):-dcg_def(Xs).
'#>'(Xs):-dcg_val(Xs).

% example: dcg_phrase((#a,#b,#c),X).
dcg_phrase(Axiom,Phrase):-dcg_phrase(1,Axiom,Phrase).

dcg_phrase(DcgStream,Axiom,Phrase):-
  dcg_telling(X),dcg_tell(DcgStream),
    dcg_def(Phrase),
      Axiom,
    dcg_val([]),
  dcg_tell(X).

dcg_apply(Method):-
  term_append(Method,args(I,O),Goal),
  dcg_val(I),
  Goal,
  dcg_def(O).


% GLOBAL LOGICAL VARIABLES

':=:'(K1#K2,X):-!,lval(K1,K2,X).
':=:'(K,X):-lval(K,K,X).

/* bboard garbage collector */

area_used(Area,Used):-
  statistics(Area,[Used,_Left]).

heap_used(Used):-area_used(global_stack,Used).

bb_used(Used):-area_used(bboard,Used).


bb_gc:-bbgc_status(0),!,clear_bp_error,errmes(blackboard_gc, disabled).
bb_gc:-bb_gc0.

bb_gc0:-	
  debugmes(starting(bb_gc)),
  bb_used(BB1),
  heap_used(H1),
  bb_list0(3,1,B),              % length(B,_L1),
  heap_used(H2),
  copy_term(B,NewB),            % length(NewB,_L2),
  heap_used(H3),
  H is (H3-H2)-(H2-H1)//2,
  bb_reset(H),
  bb_put_back(NewB),
  bb_used(BB2),BB is BB1-BB2,
  debugmes(bb_gc(before=BB1,estimated=H,after=BB2,savings=BB)),
     %     bb_list0(3,1,AfterB),length(AfterB,_L3),bb_valid(_L1,_L2,_L3),
  bb_check, 
  fail.
bb_gc0.

%bb_valid(L,L,L):-!,debugmes(bb_gc_valid_records_ok(L)).
%bb_valid(L1,L2,L3):-bb_fail(lost_records_during_gc(L1,L2,L3)).

bb_check:-
   statistics(bboard,[Used,Left]),
   ( Left > Used>>4 -> true
   ; bb_fail(not_enough_garbage_colected(bb_used(Used),bb_left(Left)))
   ).

bb_put_back(Xs):-bb_put_back1(Xs),fail.
bb_put_back(_).

bb_put_back1([P,N,F,K|VXs]):-
        unify_to(VXs,[V|_]),
        functor(O,P,N),
        functor(M,F,K),
%        debugmes(putting_back(P/N+F/K)),
	bb_put_1(O,M,V).
bb_put_back1([_P,_N,_F,_K,_V|Xs]):-
	bb_put_back1(Xs).

% fix to ugly bug from 5.00 to 5.25
% no need to put back assumed stuff, it is happy on the heap!!!
bb_put_1('$assumed',_,_):-!,fail.
bb_put_1(O,M,V):-save_term(V,NewV),set(O,M,NewV),!,fail.
bb_put_1(O,M,V):-bb_fail(bb_put_1(O,M,V)).

bb_fail(X):-
	statistics(bboard,[_,Z]),
	user_error('blackboard overflow, left only'(Z),culprit(X)). %fail
bb_fail(_):-
	quietmes('*** discarding inconsistent blackboard content'),
        restart,
        abort.


% An easy patch to read.pl for basic float input - Paul Tarau, Oct. 1993

try_float(_Ds,[atom('.'),integer(box(_Fraq,FDs))|Ts],I,[atom(Float)|Ts1]):-!,
	try_float_exp(Ts,Ts1,Exp),name(NFraq,FDs),
	input_float(I,NFraq,Exp,Float).
try_float(Ds,Tokens,I,[integer(box(I,Ds))|Tokens]).

try_float_exp([atom(e)|Ts],Ts1,Exp):-try_float_exp1(Ts,Ts1,Exp),!.
try_float_exp([atom(EN)|Ts],Ts,Exp):- [E]="e",
	name(EN,[E|NL]),name(Exp,NL),integer(Exp),!.
try_float_exp(Ts,Ts,0).

try_float_exp1([atom('-'),integer(box(Exp,_))|Ts],Ts,Exp1):-!,
	Exp1 is 0-Exp.
try_float_exp1([atom('+'),integer(box(Exp,_))|Ts],Ts,Exp).

float_minus(F,NegF):-NegF is 1/2-F-1/2.

'**'(A,X,R):-float_fun2(pow,A,X,R). % same as pow
pow(A,X,R):-float_fun2(pow,A,X,R).
log(A,X,R):-float_fun2(log,A,X,R).
atan2(Y,X,R):-float_fun2(atan2,Y,X,R).

%sqrt(X,R):-A=1,B=2,E is A/B,float_fun2(pow,X,E,R).
%hypot(X,Y,R):- '*'(X,X,X2), '*'(Y,Y,Y2),'+'(X2,Y2,S),sqrt(S,R).

sqrt(X,R):-float_fun('Q',X,R).
hypot(X,Y,R):-float_fun2(hypot,X,Y,R).

% F(X+Y) --> R used for Y=0
exp(X,R):-float_fun(exp,X,R).
log(X,R):-float_fun(log,X,R).
sin(X,R):-float_fun(sin,X,R). 
cos(X,R):-float_fun(cos,X,R).
tan(X,R):-float_fun(tan,X,R).

% F(X)+Y --> R used for Y=0
atan(X,R):-float_fun('T',X,R).
asin(X,R):-float_fun('S',X,R).
acos(X,R):-float_fun('C',X,R).

integer(X,R):-float_fun(integer,X,R).
float(X,R):-R is (1/2)+X-(1/2).

% some ISO arithmetic stuff

sign(X,S):- X>0,!,S=1.
sign(0,0):- !.
sign(_,-1).

abs(X,R):- X>=0,!,R=X.
abs(X,R):- R is -X.

floor(X,R):-integer(X,R).
ceiling(X,R):-integer(X,I),R is I+1.
truncate(X,R):-integer(X,R).
round(X,R):-R is integer(1/2+X).

% for portability...
ctime(T):-statistics(runtime,[T,_]).

is_predicate(H):-is_builtin(H),!.
is_predicate(H):-is_compiled(H),!.
is_predicate(H):-is_dynamic(H).

predicate_property(H,Prop):-nonvar(H),!,predicate_property0(H,Prop).
predicate_property(H,Prop):-
  generate_a_predicate(H),
  predicate_property0(H,Prop).

predicate_property0(H,Prop):-is_assumed(H),!,
  (Prop=(assumed);Prop=(dynamic);Prop=(interpreted)).
predicate_property0(H,Prop):-is_asserted(H),!,
  (Prop=(asserted);Prop=(dynamic);Prop=(interpreted)).
predicate_property0(H,(built_in)):-is_builtin(H),!.
predicate_property0(H,Prop):-is_multifile(H),!,
  (Prop=(multifile);Prop=(discontiguous)).
predicate_property0(H,(compiled)):-is_compiled(H).

current_predicate(F,H):-nonvar(H),!,is_predicate(H),functor(H,F,_).
current_predicate(F,H):-generate_a_predicate(H),functor(H,F,_).

current_predicate(F/N):-nonvar(F),nonvar(N),!,
  functor(H,F,N),is_predicate(H),!.
current_predicate(F/N):-
  generate_a_predicate(H),
  functor(H,F,N).

predicate_key0('$assumed'/0,PN,V,PN):-nonvar(V).
predicate_key0(PN,'$first'/0,_V,PN).
predicate_key0(predmark/0,P/N,predmark,P/N1):-N>0,N1 is N-1. % for true/0

generate_a_predicate(P):-
  ( is_builtin_var(P)
  ; generate_compiled(P), \+ is_builtin_nonvar(P)
  ; generate_run_time_predicate(P)
  ).

generate_compiled(P):-
	bb_list0(0,0,Xs),
 	bb_element(K1+K2=V,Xs),
        predicate_key0(K1,K2,V,F/N),
        functor(P,F,N).

% DYNAMIC CODE

% Blackboard related utilities

% bboard visualisation

bb_element(P/N+F/K=V,[P,N,F,K,V|_]).
bb_element(D,[_,_,_,_,_|L]):-bb_element(D,L).

bb_list(L):-bb_list0(3,0,L). % lists for arity >= 0

bb0:-bb(plain).

bb:-bb(unsorted).

bb(How):-
	statistics(bboard,X),write(bboard-X),nl,
	bb_list(L),
	( How\==sorted->findall(E,bb_element(E,L),Es),sort(Es,Bs)
	; Bs=L
        ),
	member(B,Bs),bb_orig(B,B0),
	( How==plain->write(B0),nl;pp_term(B0)),
	fail
; nl.

bb_orig(K1/N1+K2/N2=_,K1/N1+K2/N2=V):-
  functor(T1,K1,N1),
  functor(T2,K2,N2),
  (  val(T1,T2,V)
  ;  lval(T1,T2,V)
  ),!.

% C-based low-level stack/queue bboard operations
% implementation: please consider it as a black box!

popq1(O,M,X):-val(O,M,S),popq0(S,X).

addq1(O,M,X):-val(O,M,S),!,addq0(S,X).
addq1(O,M,X):-addq0(S,X),def(O,M,S).

pushq1(O,M,X):-val(O,M,S),!,pushq0(S,X).
pushq1(O,M,X):-pushq0(S,X),def(O,M,S).

addq(O,M,X):-addq1(O,M,X),!.
addq(O,M,X):-get_bp_error(_,Mes,_,_),
  (
    Mes=='BB_OVERFLOW',bb_gc,addq1(O,M,X),clear_bp_error,!
    ; errmes(Mes,addq(O,M,X))
  ).

pushq(O,M,X):-pushq1(O,M,X),!.
pushq(O,M,X):-get_bp_error(_,Mes,_,_),
  (
    Mes=='BB_OVERFLOW',bb_gc,pushq1(O,M,X),clear_bp_error,!
    ; errmes(Mes,pushq(O,M,X))
  ).

cpopq(O,M,X):-popq1(O,M,X),!.
cpopq(O,M,X):-
  get_bp_error(_,Mes,_,_),Mes\=='$empty',
  errmes(Mes,cpopq(O,M,X)).

memberq(X,Xs):-member_scan(X,Xs,[Y|Zs]),mq1(X,Y,Zs).

mq1(X,Y,_):-unify_to(Y,X).
mq1(X,_,Zs):-memberq(X,Zs).

cmember_filt(X0,Xs,[Y|Xs2]):-
   cmember_scan(X0,Xs,[Y|Xs1]),!,
   cmember_filt(X0,Xs1,Xs2).
cmember_filt(_,_,[]).

% vulnerable to bb_gc occurring while code is remaining code is executing
% cmemberq(C,I):-cq0(C,I).

% slightly less efficient for large predicates
cmemberq(C,I):-cmember_filt(C,I,Cs),member(C,Cs).

cq0(X,Xs):-cmember_scan(X,Xs,[Y|Zs]),cq1(X,Y,Zs).

cq1(X,X,_).
cq1(X,_,Xs):-cq0(X,Xs).

cdelq_all(C,Xs):-dq0(C,Xs).

dq0(C,Xs):-
  cdel_scan(C,Xs,[NewC|Ys]),
  dq1(C,NewC,Ys).

dq1(C,C,_).
dq1(C,_,Ys):-dq0(C,Ys).

% removes the first item matching X and takes a copy
% nondestructive on X
cdelq(Key,Name,X,Deleted):-
  tval(Key,Name,var(XsYs)),array_get(XsYs,1,var(Xs)),
  cdel_scan(X,Xs,[Deleted|_]).

% end of (low-level) primitives

% higher level persistent queue/stack manipulation tools

% checks if there
membq(Key,Name,X):- 
  val(Key,Name,XsYs),
  unify_to(XsYs,Xs-_),
  memberq(X,Xs).

% nondestructively checks/copies matching X's
cmembq(Key,Name,X):- 
  tval(Key,Name,var(XsYs)),array_get(XsYs,1,var(Xs)),
  cmemberq(X,Xs).

% adds a (assumed ground) object unless it is there
memoq(Key,Name,X):-
  cmembq(Key,Name,X),!.
memoq(Key,Name,X):-
  addq(Key,Name,X).


/* dynamic/1, assert/1 & retract/1 predicates

This is an approximation of other Prologs assert & retract predicates.
For efficiency and programming style reasons we strongly suggest
not to use them too much.

If you want maximal efficiency use bb_def/3 bb_set/3 bb_val/3 
and lval/3. They give you acces to a very fast
hashing table <key,key>--> value, the same that BinProlog
uses internally for indexing by predicate and first argument.

Lval/3 (i,i,?) introduces named global logical variables.
A friendlier syntax (with no extra cost) is

Key1#Key2:=:Var  ...for lval(Key1,Key2,Var)

and Key:=:Var    ...for lval(Key,Key,Var)

Beware that assert & retract are not optimized for large databases
or frequent use of an asserted predicate.

To have dynamic predicates in a compiled file you must declare 
them as such with dynamic/1. All predicated in a consulted file
are dynamic. Dynamic code DOES NOT overrides compiled code 
anymore in 5.0 although it still turn overriden by assumed code.
*/

dynamic(Ps):-dynamic0(Ps).

dynamic0(Ps):-define_dynamic(Ps),fail.
dynamic0(_).
 
define_dynamic((P1,P2)):-!,define_dynamic(P1),define_dynamic(P2).
define_dynamic(P/N):-
	functor(T,P,N),
	check_dynamic(T,P/N),
	define_dynamic1(T).

check_dynamic(T,PN):-is_compiled(T),!,
  debugmes('WARNING: compiled predicate'(PN)),nl.
check_dynamic(_,_).

define_dynamic1(H):-asserta(H),retract1(H),!.

asserta(C):-
  current_db(DB),
  db_asserta(DB,C).

assertz(C):-
  current_db(DB),
  db_assertz(DB,C).

db_asserta(DB,C0):-
  db_keep_clause(DB,C0,H,C),
  pushq(DB,H,C).

db_assertz(DB,C0):-
  db_keep_clause(DB,C0,H,C),
  addq(DB,H,C).

assert(C):-assertz(C).

db_assert(DB,C):-db_assertz(DB,C).

db_asserted(DB,H):-nonvar(H),db_clause(DB,H,B),call_body(B).

asserted(H):-nonvar(H),clause(H,B),call_body(B).

clause(H,B):- current_db(DB),db_clause(DB,H,B).

db_clause(DB,H,B):-nonvar(H),!,cmembq(DB,H,(H:-B)).
db_clause(DB,H,B):-db_head(DB,H),cmembq(DB,H,(H:-B)).

clause0(H,B):- current_db(DB),membq(DB,H,(H:-B)).

% call/update ratio that triggers dyn_compile

db_ratio(X):-integer(X),!,vset(db_ratio,int(X)).
db_ratio(X):-vget0(db_ratio,int(X)).

check_for_clash(DB,DB):-!,
  errmes('known bug: unable to assert in db'(DB),DB/0).
check_for_clash(_,_).

db_keep_clause(DB,C0,H,C):-
  add_true(C0,C),
  unify_to(C,(H:-_)),
  check_for_clash(DB,H),
  disable_static(H,DB),
  db_ratio(R),
  ( val(H,DB,K)->K1 is K +R,set(H,DB,K1)
  ; functor(H,F,N),
    addq('$clauses',DB,F/N),
    def(H,DB, R)
  ).

% DYNAMIC COMPILATION HEURISTIC
dynamize(DB,H):-
    val(H,DB,K),db_ratio(R),
    K1 is K +R,set(H,DB,K1),
    disable_static(DB,H).       % H in DB is ready for updates!

% will inc counter for <DB,H> by 1 until =<0 (when dyn_compiles)

dyn_compile(H):-current_db(DB),dyn_compile(DB,H).

/*
% moved to C
dyn_compile(DB,H):-
   val(H,DB,K),K>0,!,
   K1 is K - 1,       % to eventually become =< 0
   set(H,DB,K1),
   fail.
*/
dyn_compile(DB,H):-
   debugmes(dyn_compiling(H)),
   functor(H,F,N),make_static(F/N),
   rm(H,DB).

% db enumerators
db_pred(DB,F,N):-cmembq('$clauses',DB,F/N).

db_pred(F,N):- current_db(DB),cmembq('$clauses',DB,F/N).

db_head(DB,H):-db_pred(DB,F,N),functor(H,F,N).

% retracts at most 1 matching clause
db_retract1(DB,H):-C=(H:-_),dynamize(DB,H),cdelq(DB,H,C,C).

retract1(H):-current_db(DB),db_retract1(DB,H).

% retracts and backtracks - with immediate update semantics -
% might cause problems with multiple engines !!!

db_retract(DB,H):-
  tval(DB,H,var(XsYs)),array_get(XsYs,1,var(Xs)),
  dynamize(DB,H),
  cdelq_all((H:-_),Xs).

/*
% retracts and backtracks - with immediate update semantics -
% restarts from the beginning of a predicate each time
% on backtracking - should be safe with multiple engines
% although concurrent backtracking is STRONGLY discuraged
% anyway because of its too complex operational semantics!

db_retract(DB,H):- \+ db_clause(DB,H,_),!,fail.
db_retract(DB,H):- db_retract1(DB,H).
db_retract(DB,H):- db_retract(DB,H).
*/

retract(H):-
  current_db(DB),
  db_retract(DB,H).

% retracts all matching clauses

db_retractall(DB,H):-db_retract(DB,H),fail.
db_retractall(_,_).

retractall(H):-current_db(DB),db_retractall(DB,H).

db_clean:-
  current_db(DB),
  db_clean(DB).

db_clean(DB):-
  db_pred(DB,F,N),
    db_abolish(DB,F,N),
  fail.
db_clean(_).


db_move(From,To):-
  db_head(From,H),
    db_keep_clause(To,(H:-true),_,_),
    val(From,H,Saved),
    let(To,H,Saved),
    (val(H,From,Ctr)->let(H,To,Ctr);true),
    functor(H,F,N),
    db_abolish(From,F/N),
  fail.
db_move(_,_).

db_abolish(DB,F,N):-db_abolish(DB,F/N).

db_abolish(DB,F/N):-
  functor(H,F,N),
  ( val(DB,H,_)->
    rm(DB,H),    
    cdelq('$clauses',DB,F/N,_),
    rm(H,DB)
  ; true
  ).

abolish(F,N):-abolish(F/N).

abolish(FN):-
  current_db(DB),
  db_abolish(DB,FN).

% on-place static<-->dynamic conversion

make_static(F/N):-
  functor(H,F,N),
  is_asserted(H),
  dyn2stat(H),
  !.
make_static(FN):-
  errmes(unable_to,make_static(FN)).

make_dynamic(F/N):-
  functor(H,F,N),
  is_compiled(H),
  stat2dyn(H),
  !.
make_dynamic(FN):-
  errmes(unable_to,make_dynamic(FN)).

make_all_static:-dyn2stat(_).

make_all_dynamic:-stat2dyn(_).

dyn2stat(H):-
  current_db(DB),
  dyn2stat(DB,H).

dyn2stat(DB,H):-
  disable_static(DB,H),
  vget0(code_top,Top),
  vget0(code_oldtop,OldTop),
  vset(code_oldtop,Top),
    make_static0(DB,H), 
  terminate_file0(mem,1),
  vset(code_oldtop,OldTop).

stat2dyn(H):-
  current_db(DB),
  stat2dyn(DB,H).

stat2dyn(DB,H):-
  db_head(DB,H),
    disable_static(DB,H),
  fail.
stat2dyn(_,_).

make_static0(DB,H):-
  db_clause(DB,H,B),
    quickcomp((H:-B)),
  fail.
make_static0(_,_).

disable_static(DB,H):-current_db(DB),!,disable_static(H).
disable_static(_,_).

disable_static(H):-is_compiled(H),!,override(2,H,fail).
disable_static(_).

% Spying on compiled/dynamic/assumed code.
% Note that for compiled code, declarations ":-spy(F/N)." or
% :-spy(all). or :-trace. should be
% _BEFORE_ the clauses in the file they are intended to refer.

spy_goal(G):-trace(G).

trace:-spy(all).
notrace:-nospy(all).

spying_strictly(H):-val(spying,H,yes).

spying(H):-val(spying,H,yes),!.
spying(H):-val(spying,all,all), \+ is_builtin(H).

spy(all):-!,bb_let(spying,all,all),
   seeing(F),
   ttyprint(spying_all_predicates_in_file(F)).
spy(F/N):-functor(H,F,N),\+ is_builtin(H),!,bb_let(spying,H,yes).
spy(X):-user_error(unable_to_spy_on,X).

nospy(all):-!,bb_rm(spying,all).
nospy(F/N):-functor(H,F,N),bb_rm(spying,H).


% consulting `interpreted code'

% reconsult/1 variant:
% clean up data areas, consults, makes all static
sconsult(File):-
  clean_up_data_areas(File,F),
  mainconsult(consult0(F),make_all_static).

% reconsult/1 variant:
% cleans up data areas, consults, allowing dynamic recompilation
dconsult(File):-
  clean_up_data_areas(File,F),
  mainconsult(consult0(F),true).

% reconsult/1 variant:
% consults and overwrites old clauses
oconsult(File):-
	find_file(File,F),
        current_db(Current),
        symcat(Current,F,New),
        set_db(New),
        mainconsult(consult0(F),true),
        db_move(New,Current),
        set_db(Current).

% consults with possible dupplication of clauses,
% allows later dynamic recompilation
consult(File):-
  current_db(DB),
  consult(File,DB).

% allows consulting into an inactive database
% this allows interference with compiled code
consult(File,DB):-
	find_file(File,F),
        mainconsult(consult0(F,DB),true).

clean_up_data_areas(File,F):-
  find_file(File,F0),
  survive_cleanup(F0,F),
  set_current_user_file(F).


mainconsult(ConsultStep,CompileStep):-
   statistics(code,[C1,_]),
   ctime(T1),
   ConsultStep,
   ctime(T2),
   CompileStep,
   ctime(T3),
   statistics(code,[C2,_]),
   CT is T2-T1,
   ST is T3-T2,
   C is C2-C1,
   quietmes(time(consulting=CT,quick_compiling=ST,static_space=C)).

consult0(F):-
   current_db(DB),
   consult0(F,DB).

consult0(F,DB):-
	quietmes(consulting(F)),
	see(F),
        consult1(DB),
	seen,
	quietmes(consulted(F)).

consult1(DB):-
	repeat,
		gc_read_clause(C),consult_embedded(C,DB),C=end_of_file,!.
   
consult_included(F,DB):-seeing(G),find_file(F,F0),consult0(F0,DB),see(G).

consult_embedded(':-'(Cmd),DB):-!,consult_cmd(Cmd,DB).
consult_embedded(C,DB):-assert_it(C,DB).

assert_it(end_of_file,_):-!.
assert_it(C,DB):-db_assertz(DB,C).

consult_cmd(Cmd):-current_db(DB),consult_cmd(Cmd,DB).

consult_cmd([F],DB):-!,consult_included(F,DB).
consult_cmd(consult(F),DB):-!,consult_included(F,DB).
consult_cmd(oconsult(F),DB):-!,consult_included(F,DB).
consult_cmd(sconsult(F),DB):-!,consult_included(F,DB).
consult_cmd(dconsult(F),DB):-!,consult_included(F,DB).
consult_cmd(reconsult(F),DB):-!,consult_included(F,DB).
consult_cmd(compile(F),DB):-!,consult_included(F,DB).
consult_cmd(scompile(F),DB):-!,consult_included(F,DB).
consult_cmd(mcompile(F),DB):-!,consult_included(F,DB).
consult_cmd(C,_):-metacall(C).


/* Mutant version of R.A. O'Keefe's meta-circular interpreter */
/* first slightly modified by Paul Tarau, */
/* later modified beyond recognition and moved to lib.pl :-) */

call_body(Body):-var(Body),!,user_error(unbound_goal,Body).
call_body(Body) :- % compile time execution
   ## expand_call_body(Body,Code),Code. 

is_delphi(P,DP):-val(P,'$delphi',DP).

is_dynamic(H):-is_asserted(H).
is_dynamic(H):-is_assumed(H).
is_dynamic(H):-is_verb(H).

db_is_asserted(DB,H):-val(DB,H,_).

%is_asserted(H):- get_asserted(H,_).
is_asserted(H):- current_db(DB),val(DB,H,_).
is_assumed(H):-assumed_info(H,_).
is_verb(Pred):-val('::',Pred,_).

is_assumed(H,Info):-assumed_info(H,Info).

was_assumed(H):-val('$assumed',H,X),nonvar(X).

do_goal(Goal) :-
   is_compiled(Goal),!,
   Goal.
do_goal(Goal):-
%  spying_strictly(Goal),!,
   val(spying,Goal,yes),!,
   spy_goal(Goal).
do_goal(Goal):-
%  is_verb(Goal),!,
   val('::',Goal,_),!,
  '@'(Goal).
do_goal(Goal):-
%  was_assumed(Goal),!,assumed_clause(Goal,Body),
   val('$assumed',Goal,pred(Cs-_)),nonvar(Cs),!,
   get_assumed_clause(Goal,Cs,Body+_),
   %  silly optimization ^^^^^^^^^^^
   ## expand_call_body(Body,Code),Code.
do_goal(Goal):-
   get_asserted(Goal,I),!,
   ( I=:=0,dyn_compile(Goal)->Goal
   ; cmemberq((Goal:-Body),I),
     ## expand_call_body(Body,Code),Code
   ).
do_goal(G):-
   functor(G,call,_),!,
   do_apply(G,Goal),
   do_goal(Goal).
do_goal(Goal):-is_dynamic(Goal),!,fail.
do_goal(Goal):-
   /* to enable this you should do something like:
   *   host(..remote_host..)=>code(..remote_file..)=>remote_predicate(...).
   */
   call_ifdef(non_local_host,fail),
   call_ifdef(default_code(Code),fail),Code\=='$empty',
   fetch_pred_for(Goal),!,
   asserted(Goal).
do_goal(Undef):-
   do_undef(Undef).

do_undef(Undef):-
  functor(Undef,F,N),
  errmes(undefined_predicate=F/N,culprit=Undef).

/*
do_undef(Undef) @@ Cont:-
  functor(Undef,F,N),
  functor(Cont,G,_),Goal=..[G,'...'],
  errmes(undefined_predicate=F/N,continuation=Goal).
*/

do_apply(G,Goal):-
	G=..[_,Closure|Args],
	Closure=..L1,det_append(L1,Args,L2),
	Goal=..L2.

do_body((Goal,Body), AfterCut, HadCut):-!,
	do_conj(Goal,Body, AfterCut, HadCut).
do_body((Goal;Body), AfterCut, HadCut):-!,
	do_disj(Goal,Body, AfterCut, HadCut).
do_body(!, true, yes):-!.
do_body(Goal, true, no) :- do_goal(Goal).

do_conj(!,AfterCut, AfterCut, yes) :- !.
do_conj(Goal,Body, AfterCut, HadCut) :- 
	call_body(Goal),
	do_body(Body, AfterCut, HadCut).

do_disj((If->Then),Else, AfterCut, HadCut) :-!,
   do_if_then_else(If,Then,Else, AfterCut, HadCut).
do_disj(Disj1,Disj2, AfterCut, HadCut) :-
   do_disj0(Disj1,Disj2, AfterCut, HadCut).

do_if_then_else(If,Then,_, AfterCut, HadCut) :- call_body(If),!,
	do_body(Then, AfterCut, HadCut).
do_if_then_else(_,_,Else, AfterCut, HadCut) :-
	do_body(Else, AfterCut, HadCut).

do_disj0(Disj1,_, AfterCut, HadCut) :-
	do_body(Disj1, AfterCut, HadCut).
do_disj0(_,Disj2, AfterCut, HadCut) :-
	do_body(Disj2, AfterCut, HadCut).



