#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# agt_insert
#
# Tries to assert a belief to Belief-Base.
#
# Arguments:
#
# 	belief	- belief to assert
#
# Results:
#
#  Returns 0 on successfull insertion, 1 if an error occurrs.
#  Shows a remark in logfile.
#
proc agt_insert {belief} {
global agt_VarTimeOffset

 # check if 'belief' contains strings that are not allowed:
#
 set error 0

 switch -glob -- $belief {
     *:-* {set failString :-; set error 1}
     *set_time* {set failString set_time; set error 1}
     *time* {set failString time; set error 1}
     *invalid* {set failString invalid; set error 1}
     *valid* {set failString valid; set error 1}

 }

 if {$error} {
   agt_puts "'$failString' not allowed in '$belief'\n\nBelief-Base\
             not changed.\n"
   return 1 
 }


 agt_puts "Insert Belief:"

 # remove leading and succeding spaces from belief
 set belief [string trimleft $belief]


 if {[cindex $belief 0] == "("} {
   # in this case the belief has a time specification, i.e. it is of the form
   #  (<belief>), <from-expr>, <till-expr>  or
   #  (<belief>), <from-expr>

   # check ()-brackets are balanced in 'belief'
   # 
   set brackets 0

   for {set index 0} {$index <= [clength $belief]} {incr index} {
     set char [cindex $belief $index]

     switch -glob -- $char {
         (  {incr brackets  1}
         )  {incr brackets -1}
     } 
   }

  if {$brackets} {
     agt_puts "unbalanced brackets in '$belief'\n\nBelief-Base\
               not changed.\n" 
     return 1 
   }

   # get the index of the close bracket
   set index1 [string last ")" $belief]

   # get the characters of the time specification
   set timeString [string trimleft [crange $belief $index1+1 end]]

   # get just the belief, without brackets
   set belief [crange $belief 1 $index1-1]

   # check if the ',' exists after the belief
   if {[cindex $timeString 0] != ","} {
     agt_puts "',' expected after '$belief'\n\nBelief-Base\
              not changed.\n" 
     return 1 
   }

   # check if time specification includes a <till-expr>
   # and set 'timeFrom', 'timeTill' depending on it:

   if {[set index2 [string last "," $timeString]] ==  0} {
      # no <till-expr>, set 'timeTill' to 0, i.e. belief is valid
      # after 'timeFrom'
     set timeFrom [crange $timeString 1 end]
     set timeTill 0
   } else { 
     # <till-expr> exists so get 'timeFrom' and 'timeTill' from 'timeString'
     set timeFrom [crange $timeString 1 $index2-1]
     set timeTill [crange $timeString $index2+1 end]
   }

 } else {
   # check ()-brackets are balanced in 'belief'
   # 
   set brackets 0

   for {set index 0} {$index <= [clength $belief]} {incr index} {
     set char [cindex $belief $index]

     switch -regexp -- $char {
         [(]  {incr brackets  1}
         [)]  {incr brackets -1}
         [,]|[;] {if {$brackets == 0} {
                agt_puts "character '$char' in illegal position in\
                          '$belief'\n\nBelief-Base\
                          not changed.\n" 
                return 1
	      }
	     }
     } 
   }

  if {$brackets} {
     agt_puts "unbalanced brackets in '$belief'\n\nBelief-Base\
               not changed.\n" 
     return 1 
   }

   # no time specification, so the belief is always valid
   set timeFrom now
   set timeTill 0
 }


 # use Tcl-command 'convertclock' to get the UNIX-time from
 # time specification; subtract a value 'agt_VarTimeOffset' so that BinProlog
 # can get the integer value
 if {[catch {set timeFrom [expr [convertclock $timeFrom]-$agt_VarTimeOffset]; \
             set timeTill [expr [convertclock $timeTill]-$agt_VarTimeOffset] } \
            errorMesg]} {
    agt_puts "$errorMesg\n\nBelief-Base\
             not changed.\n"
    return 1 
 }


 # remove a possibly existing fact with same name
 #
 if {[agt_listQuery retract(($belief:-A))] == -1} {
   agt_puts "Belief-Base not changed.\n"
   return 1
 }

 if {[agt_listQuery retract(validity($belief,B,C))] == -1} {
   agt_puts "Belief-Base not changed.\n"
   return 1
 }

 # build up the belief to insert to Belief-Base
 set belief_1 "$belief:-validity(($belief), FROM, TILL), \
                        (valid(FROM, TILL); \
                         (invalid(FROM, TILL), \
                         retract((($belief):-A)),\
                         retract(validity(($belief), FROM, TILL)),fail))"

 if {[catch {agt_quietQuery "assert(($belief_1))"} errorMesg]} {
   agt_puts "Prolog - $errorMesg\n\nBelief-Base\
             not changed.\n"
   return 1
 }

 # built up the validity-belief for the just inserted belief
 set belief_2 "validity(($belief), $timeFrom, $timeTill)"

 if {[catch {agt_quietQuery "assert(($belief_2))"} errorMesg]} {
   agt_puts "Prolog - $errorMesg\n\nBelief-Base\
             not changed.\n"
 } else {
   agt_puts "Inserted: '$belief'\n"

   # set the actual time in Prolog
   agt_timeSet
   agt_displayBelBase

   return 0
 } 

}



#
# agt_delete
#
# Tries to retract a belief from Belief-Base.
#
# Arguments:
#
# 	belief	- belief to retract
#
# Results:
#
#  Returns 0 on successfull retraction, 1 if an error occurrs.
#  Shows a remark in control display.
#
proc agt_delete {belief} {

 agt_puts "Delete Belief:"

 # check if 'belief' contains strings that are not allowed:
#
 set error 0

 switch -glob -- $belief {
     *:-* {set failString :-; set error 1}
     *set_time* {set failString set_time; set error 1}
     *time* {set failString time; set error 1}
     *invalid* {set failString invalid; set error 1}
     *valid* {set failString valid; set error 1}

 }

 if {$error} {
   agt_puts "'$failString' not allowed in '$belief'\n\nBelief-Base\
             not changed.\n"
   return 1 
 }

 # build up belief to delete
 set belief_1 "($belief:-A)"

 # delete the belief
 if {[set result [agt_listQuery retract($belief_1)]] == -1} {
   agt_puts "Belief-Base not changed.\n"
   return 1
 } elseif {$result == 0} {
   agt_puts "Prolog - '$belief' not exists in Belief-Base\n\nBelief-Base\
             not changed.\n"

   return 1
 } else {
   # remove the validity-belief for the just removed belief
   set belief_2 "validity($belief, FROM, TILL)"

   if {[agt_listQuery retract($belief_2)] == -1} {
     agt_puts "Belief-Base not changed.\n"
     return 1
   }

   agt_puts "Deleted: '$belief'\n"

   # set the actual time in Prolog
   agt_timeSet
   agt_displayBelBase
   
   return 0
 } 

}



#
# agt_saveBelBase
#
#  Saves content of Belief-Base to a file with extension '.pl'
#  in directory 'ideas/save'.
#
# Arguments:
#
#	fileName - name (path) of file to save to
#
# Results:
#
#	0 - everything ok
#	1 - cannot create fileName
#	2 - write error
#
#  Shows an errorMesg in logfile.
#
proc agt_saveBelBase {fileName} {
 
 # if extension of fileName is not '.pl' set it to '.pl'
 #
 if {![cequal [file extension $fileName] .pl]} {
   append fileName .pl
 }

 # open file
 if {[catch {set tmpFid [open "save/$fileName" w]} errorMesg]} {
   agt_puts "Failed to save Belief-Base:\n$errorMesg\n"
   return 1
 }

 # read actual Belief-Base from text widget
 set beliefBase [.f3.text get 0.0 end]

 # write actual beliefs to this file
 if {[catch {puts $tmpFid $beliefBase; flush $tmpFid} errorMesg]} {
   agt_puts "Failed to save Belief-Base:\n$errorMesg\n"
   return 2
 }

 catch {close $tmpFid}

 agt_puts "Saved Belief-Base to file '$fileName'.\n"

 return 0
}



#
# agt_loadBelBase
#
#  Loads a Belief-Base from a file in directory 'ideas/save'.
#
# Arguments:
#
#	fileName - name (path) of file to load from
#
# Results:
#
#	0 - everything ok
#	1 - prolog returned error
#
#  Shows an message in logfile.
#
proc agt_loadBelBase {fileName} {

 # if extension of fileName is .pl remove it
 #
 if {[cequal [file extension $fileName] .pl]} {
   set fileName [csubstr $fileName 0 [expr [clength $fileName]-3]]
 }

 # reconsult the file
 if {[catch {agt_quietQuery "reconsult('save/$fileName')"} \
        errorMesg]} {
   agt_puts "Failed to load Belief-Base:\n$errorMesg\n"
   return 1
 }

 agt_displayBelBase

 agt_puts "Loaded Belief-Base from file '$fileName.pl'.\n"

 return 0

}


