#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# agt_parseDescFile
#
#  Reads the agents description file and calls
#  procs to parse ACTION-C-SOURCE, BELIEFS, 
#  BELIEF-RULES, MESSAGE-RULES, ACTIONS 
#  and to initialize their stuff.
#
# Arguments:
#
# 	fileName	-	name of file to parse
#
# Results:
#
#  Returns an error with error message when a parsing
#  error occurs.
#
proc agt_parseDescFile {fileName} {

 # try to open fileName in directory 'desc' for reading
 if {[catch {set fid [open desc/$fileName]} errorMesg]} {
   return -code error $errorMesg
 }

 # read file and split it in lines
 set file [split [read $fid] \n]

 close $fid

 #
 # This is the initial situation for parsing:
 # Variable 'file' is a list containing the content of agents type file,
 # each list entry is one line of type file.
 # Line by line is popped from this list and the actual line to parse
 # is hold in variable 'line'. So 'line' and 'file' allways contain the 
 # unparsed rest of the type file.
 #

 # pop the first line from file and remove leading and
 # succeeding white spaces
 set line [string trim [lvarpop file]]

 # As described in the ASL-Syntax the type file has 5 parts which have
 # to occure in a fixed order. The variable 'itemCt' is used to count this
 # parts and shows which parts are legal next, belonging to the following list:
 # 
 #  <= 0  ACTION-C-SOURCE
 #  <= 1  BELIEFS 
 #  <= 2  BELIEF-RULES
 #  <= 3  MESSAGE-RULES
 #  <= 4  ACTIONS
 #
 set itemCt 0
 
 # start parsing 'file' line by line
 #
 while {1} {
     switch -regexp $line {

      ^$|^# {
             # This is an empty line or a remark, so get the next line
             # frome 'file' if it is not empty
             if {[llength $file] == 0} {
               return -code ok
             }
             set line [string trim [lvarpop file]]
            }

      ^load {
             # This is a 'load' command to load a description part from
             # a file specified in the rest of this line
             agt_loadDescPart file [string trimleft [csubstr $line 4 end]]

             # get the next line from file
             set line [string trim [lvarpop file]]}

      ^ACTION-C-SOURCE {
                # Check the 'itemCt' if a 'ACTION-C-Source' declaration
                # is legal at this position (see above list)
                if {$itemCt > 0} {
                  return -code error \
                    "ASL-Syntax-Error:\nno ACTION-C-SOURCE declaration\
                     expected at '$line'"
                }
                # remove 'ACTION-C-SOURCE' and following spaces from line
                set line [string trimleft [crange $line 15 end]]

                # call 'agt_parseActSrc' to read the source specification
                if {[catch {agt_parseActSrc file line} errorMesg]} {
                  return -code error \
                    "ASL-Syntax-Error - ACTION-C-SOURCE:\n$errorMesg"
                }

                # set 'itemCt' to 1, i. e. ASL-part to read next must have
                # an index greater or equal 1 in above list
                set itemCt 1
               }


      ^BELIEFS {if {$itemCt > 1} {
                  return -code error \
                    "ASL-Syntax-Error:\nno BELIEF declaration\
                     expected at '$line'"
                }
                # remove 'BELIEFS' and following spaces from line
                set line [string trimleft [crange $line 7 end]]

                # call 'agt_parseInits' with index 1 to read the initial
                # beliefs and insert them in BeliefBase.
                if {[catch {agt_parseInits file line 1} errorMesg]} {
                  return -code error \
                    "ASL-Syntax-Error - BELIEFS:\n$errorMesg"
                }
                set itemCt 2
               }

      ^BELIEF-RULES {
                if {$itemCt > 2} {
                  return -code error \
                    "ASL-Syntax-Error:\nno BELIEF-RULES declaration\
                     expected at '$line'"
                }
                # remove 'BELIEF-RULES' and following spaces from line
                set line [string trimleft [crange $line 12 end]]

                # call 'agt_parseInits' with index 2 to read the
                # belief-rules and insert them in BeliefBase.
                if {[catch {agt_parseInits file line 2} errorMesg]} {
                  return -code error \
                    "ASL-Syntax-Error - BELIEF-RULES:\n$errorMesg"
                }
                set itemCt 3
               }

      ^MESSAGE-RULES {
                if {$itemCt > 3} {
                  return -code error \
                    "ASL-Syntax-Error: no MESSAGE-RULES declaration\
                     expected at '$line'"
                }
                # remove 'MESSAGE-RULES' and following spaces from line
                set line [string trimleft [crange $line 13 end]]

                # call 'agt_parseInits' with index 3 to read the 
                # message-rules.
                if {[catch {agt_parseInits file line 3} errorMesg]} {
                  return -code error \
                    "ASL-Syntax-Error - MESSAGE-RULES:\n$errorMesg"
                }
                set itemCt 4
               }

      ^ACTIONS {
                if {$itemCt > 4} {
                  return -code error \
                    "ASL-Syntax-Error:\nno ACTIONS declaration\
                     expected at '$line'"
                }
                # remove 'ACTIONS' and following spaces from line
                set line [string trimleft [crange $line 7 end]
]
                # call 'agt_parseInits' with index 4 to read the 
                # actions and insert them in TCL-interpreter.
                if {[catch {agt_parseInits file line 4} errorMesg]} {
                  return -code error \
                    "ASL-Syntax-Error - ACTIONS:\n$errorMesg"
                }
                set itemCt 5
               }

      default {return -code error \
                "ASL-Syntax-Error:\nunexpected symbol at '$line'"}
    }
 }

}


