#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# agt_parseInits
#
#  Used to scan initial BELIEFS, BELIEF-RULES, MESSAGE-RULES
#  and ACTIONS from agents type File and initialize the belief
#  data base.
#  Calls procs agt_parseOneBelief, agt_parseOneBeliefRule,
#  agt_parseOneMesgRule, agt_parseOneAction (resp.) to do so.
#
# Arguments:
#
# 	typeFile (referenz)	- list of lines of agt type file not
#				  scanned till now
#	actLine	(referenz)	- actual line to scan
#	procNo			- flag that shows whats to scan:
#				   1 BELIEFS
#				   2 BELIEF-RULES
#				   3 MESSAGE-RULES
#				   4 ACTIONS
#
# Results:
#
#  typeFile is scanned for 1, 2, 3 or 4 and read lines are popped.
#  actLine contains the unscanned rest of the actual line to scan. 
#  Returns an error with an errorMesg if a scan error occurs
#

proc agt_parseInits {typeFile actLine procNo} {
upvar $typeFile file \
      $actLine line


# The next character in this line must be an open bracket
#
 if {[cindex $line 0] != "\{"} {
   return -code error "'\{' expected at '$line'"
 }

 # counter for open brackets 
 set brackets 1

 # first character in $line is now a \{-bracket
 # remove the bracket and following spaces
 set line [string trimleft [csubstr $line 1 end]]

 # run through file until brackets are balanced
 #
 while {$brackets != 0} {
 
   switch -regexp -- $line {

     ^$|^# {
            # This is an empty line or a remark, so get the next line
            # frome 'file' if it is not empty
            if {[llength $file] ==0} {
              return -code error \
                "'\}' expected"
            }
            set line [string trim [lvarpop file]]
           }

     ^load {
            # This is a 'load' command to load a description part from
            # a file specified in the rest of this line
            agt_loadDescPart file [string trim [csubstr $line 4 end]]

            # get the next line from file
            set line [string trim [lvarpop file]]}

     ^[(] {
           # This is the open bracket for a BELIEF, BELIEF-RULE, 
           # MESSAGE-RULE or ACTION.
           # Remove the bracket and following spaces from 'line' and
           # call a parsing subroutine depending on 'procNo'.
           set line [string trimleft [csubstr $line 1 end]]

           switch $procNo {
            1 {set result [catch {agt_parseOneBelief file line} errorMesg]}
            2 {set result [catch {agt_parseOneBeliefRule file line} errorMesg]}
            3 {set result [catch {agt_parseOneMesgRule file line} errorMesg]}
            4 {set result [catch {agt_parseOneAction file line} errorMesg]}
           }

           # if the result of the subroutine is an error return
           # the error message of the subroutine
           if {$result} {
             return -code error $errorMesg
           }
          }

     ^\} {incr brackets -1}

     default {return -code error \
                "'\}' expected at '$line'"}
   }
 }


 # first character in $line is now a \}-bracket
 # remove the bracket and following spaces
 set line [string trimleft [csubstr $line 1 end]]

 return
}


