#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# agt_parseOneBelief
#
#  Parses the type file for one correct belief and inserts it in BeliefBase.
#
# Arguments:
#
#	typeFile (referenz)	- list of lines of agt type file not
#				  scanned till now
#	actLine	(referenz)	- actual line to scan next;
#				  starts with the begin of a belief
#
# Results:
#
#  Removes the belief from 'typeFile',
#  set 'actline' so that its first character is the first one 
#  after the end of belief and adds the belief to the belief-base.
#  Returns an error with error message if syntax error is 
#  detected. 
# 

proc agt_parseOneBelief {typeFile actLine} {
upvar $typeFile file \
      $actLine line
global agents agt_VarTimeOffset

 # the first character of line is here the first one _AFTER_
 # the (-bracket and line has no succeeding white spaces

 # the scanning itself is done in a subroutine that puts the read
 # characters of the action in variable 'belief' or
 # returns with error if it fails
 #
 if {[catch {agt_parseSubPart file line belief} errorMesg]} {
   return -code error $errorMesg
 }

 if {[cindex $belief 0] == "("} {
   # in this case the belief has a time specification, i.e. it is of the form
   #  (<belief>), <from-expr>, <till-expr>  or
   #  (<belief>), <from-expr>

   # get the index of the close bracket
   set index1 [string last ")" $belief]

   # get the characters of the time specification
   set timeString [string trimleft [crange $belief $index1+1 end]]

   # get just the belief, without brackets
   set belief [crange $belief 1 $index1-1]

   # check if the ',' exists after the belief
   if {[cindex $timeString 0] != ","} {
     return -code error "',' expected after '$belief'"
   }

   # check if time specification includes a <till-expr>
   # and set 'timeFrom', 'timeTill' depending on it:

   if {[set index2 [string last "," $timeString]] == 0} {
      # no <till-expr>, set 'timeTill' to 0, i.e. belief is valid
      # after 'timeFrom'
     set timeFrom [crange $timeString 1 end]
     set timeTill 0
   } else { 
     # <till-expr> exists so get 'timeFrom' and 'timeTill' from 'timeString'
     set timeFrom [crange $timeString 1 $index2-1]
     set timeTill [crange $timeString $index2+1 end]
   }

 } else {
   # no time specification, so the belief is always valid

   # check ()-brackets if there are no "," and ";" outside
   # balanced brackets
   #
   set brackets 0

   for {set index 0} {$index <= [clength $belief]} {incr index} {
     set char [cindex $belief $index]
  
     switch -regexp -- $char {
         [(]  {incr brackets  1}
         [)]  {incr brackets -1}
         [,]|[;] {if {$brackets == 0} {
                    set errorMesg "character '$char' in illegal position in\
                              '$belief'\n" 
                    return -code error $errorMesg
                  }
                 }
     } 
   }

   set timeFrom now
   set timeTill 0
 }

 if {[string match *:-* $belief]} {
   return -code error "no ':-' expected in '$belief'"
 }
 
 # use Tcl-command 'convertclock' to get the UNIX-time from
 # time specification; subtract a value 'agt_VarTimeOffset' so that BinProlog
 # can get the integer value
 if {[catch {set timeFrom [expr [convertclock $timeFrom]-$agt_VarTimeOffset]; \
             set timeTill [expr [convertclock $timeTill]-$agt_VarTimeOffset] } \
            errorMesg]} {
   return -code error $errorMesg
 }


 # substitute all occurenses of $SELF in the scanned
 # belief with agents name
 regsub -all {\$SELF} $belief [keylget agents(SELF) NAME] belief 

 # remove a possibly existing fact with same name
 #
 if {[agt_listQuery retract(($belief:-A))] != 0} {
   agt_listQuery retract(validity($belief,B,C))
 }

 # build up the belief to insert to BeliefBase
 set belief_1 "$belief:-validity($belief, FROM, TILL), \
                        (valid(FROM, TILL); \
                         (invalid(FROM, TILL), \
                         retract(($belief:-A)),\
                         retract(validity($belief, FROM, TILL)),fail))"

 if {[catch {agt_quietQuery assert(($belief_1))} errorMesg]} {
   return -code error "Prolog - $errorMesg"
 }

 # built up the validity-belief for the just inserted belief
 set belief_2 "validity($belief, $timeFrom, $timeTill)"

 if {[catch {agt_quietQuery assert(($belief_2))} errorMesg]} {
   return -code error "Prolog - $errorMesg"
 }

 agt_puts "Read Belief: >$belief<\n"
}


