#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# agt_pathViewStuff.tcl
#
#  This file concludes the procedures used for displaying an agents
#  communication paths with bitmaps in a widget.
#
#


# 
# agt_displAgtBitmap
#
#  Displays a weak drawn agent-symbol in the pathView canvas widget
#  and creates a line, symboling the communication path, to agent with tag SELF
#  but doesn't show it yet.
#  
# Arguments:
# 
#	agtName	-	name of the agent bitmap belongs to
#
# Results:
#
#  None.
#
proc agt_displAgtBitmap {agtName} {

 # check for position - belongs to agent name
 #
 switch $agtName {
   SELF	{
         set xPos 175
	 set yPos 90
	}
   UAM	{set xPos 50
	 set yPos 90
	}
   default {
            set xPos 350
	    # search free place in widget
	    for {set i 20} \
	        {[.f2.c find overlap ${xPos} ${i} [expr $xPos+0.1] [expr $i+1]]!=""} \
	        {set i [expr $i+50]} {}
            set xPos ${xPos}
            set yPos ${i}
	    }
 }

 # create bitmap and add tag with agents name
 .f2.c create bitmap ${xPos} ${yPos} -bitmap questhead \
	-foreground slategray3 \
	-anchor center \
	-tag $agtName.bitmap
 .f2.c addtag $agtName withtag $agtName.bitmap 

 # create text for bitmap and add tag with agents name
 .f2.c create text ${xPos} [expr ${yPos}+20] -text $agtName \
	-font -Adobe-Helvetica-Medium-R-Normal--*-120-* \
	-anchor center \
	-tag $agtName.text
 .f2.c addtag $agtName withtag $agtName.text

 set agtCoord [.f2.c coord $agtName]
 set selfCoord [.f2.c coord SELF]

 # create line from new bitmap to bitmap with tag SELF and add tag with agents name
 set edge [.f2.c create line [lindex $agtCoord 0] [lindex $agtCoord 1] \
		[lindex $selfCoord 0] [lindex $selfCoord 1] \
		-tag $agtName.edge \
		-fill ""
	  ]
 .f2.c addtag $agtName withtag $agtName.edge
 .f2.c lower $edge

 # do the binding for the bitmap
 .f2.c bind $agtName <Button-1> "agt_showAgtStatus $agtName %x %y"

 .f2.c bind $agtName <ButtonRelease-1> {
	.f2.c delete status; destroy .f2.c.status
 }

 .f2.c bind $agtName <B1-Motion> {
	.f2.c coords status %x %y
 }

 if {$agtName != "SELF"} {
   .f2.c bind $agtName <B2-Motion> "agt_positBitmap $agtName %x %y"
 }
  

 update idletasks
}


#
# agt_positBitmap
#
#  Positionizes an agents bitmap new in the canvas.
#  Bind to an agents bitmap on holding button 1 pressed.
#
# Arguments:
#
#	agtName	-	name of bitmap to positionize
#	xPos	-	new x coordinate
#	yPos	-	new y coordinate
#
# Results:
#
#  None.
#
proc agt_positBitmap {agtName xPos yPos} {
	.f2.c coords $agtName.bitmap $xPos $yPos
	.f2.c coords $agtName.text $xPos [expr $yPos + 20]
	.f2.c coords $agtName.edge $xPos $yPos [lindex [.f2.c coord SELF.bitmap] 0] \
					[lindex [.f2.c coord SELF.bitmap] 1]
}


#
# agt_drawWeak
#
#  Displays the line from a bitmap to SELF in a weak colour;
#  used when doing handshake and connection is still not established
#  completely.
#
# Arguments:
#
#	agtName	-	name of the bitmap
#
# Results:
#
#  None.
#
proc agt_drawWeak {agtName} {

 .f2.c itemconfigure $agtName.edge -fill slategray3
 .f2.c itemconfigure $agtName.bitmap -foreground slategray3

 update idletasks
}


#
# agt_drawSolid
#
#  Displays the line from a bitmap to SELF in a solid colour;
#  used when handshake is done and the connection is established.
#
# Arguments:
#
#	agtName	-	name of the bitmap
#
# Results:
#
#  None.
#
proc agt_drawSolid {agtName} {

 .f2.c itemconfigure $agtName.edge -fill black
 .f2.c itemconfigure $agtName.bitmap -foreground black

 update idletasks
}


#
# agt_delAgtBitmap
#
#  Deletes an agents bitmap from canvas.
#
# Arguments:
#
#	agtName	-	name of the bitmap
#
# Results:
#
#  None.
#
proc agt_delAgtBitmap {agtName} {
 
 .f2.c delete $agtName

}


# 
# agt_showAgtStatus
#
#  Creates a message widget in the canvas and displays information about
#  a selected agent.
#  Bind to an agents bitmap on holding button 2 pressed.
#
# Arguments:
#
#	agtName	-	name of the bitmap
#	xPos	-	x coordinate for message widget
#	yPos	-	y coordinate for message widget
#
# Results:
#
#  None.
#
proc agt_showAgtStatus {agtName xPos yPos} {
 global agents

 catch {.f2.c delete status}
 catch {destroy .f2.c.status}

 set text ""
 foreach key [keylkeys agents($agtName)] {
   set text "${text}$key\t: [keylget agents($agtName) $key]\n" 
 }

 message .f2.c.status  -text $text \
	-relief ridge \
	-justify left \
	-font -Adobe-Courier-Bold-R-Normal--*-120-* \
	-width 200 
 
 .f2.c create window $xPos $yPos -window .f2.c.status -tag status -anchor n

}


