#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# agt_prologServ.tcl
#
# Tcl/Tk side of a BinProlog-Tcl/Tk interface for IDEAS
# tested on tcl7.3 with tk3.6 with tclX
#
# by Olaf Scheew, 1995
#
#

#
# agt_startProlog
#
#  Starts a new Prolog process and registers its filedesc.
#
# Arguments:
#
#  None.
#
# Results:
#
#  0, if process is OK, sets in global variable agents(SELF)
#      key PROLOG the filedesc
#  1, if an error occurred
#
proc agt_startProlog {} {
global agents ideas_PrologPath

 # initialize the globale value for the Prolog filedescriptor
 # with an empty string, so that there is 'something' to read
 # even if the Prolog process can't be started
 #
 keylset agents(SELF) PROLOG ""

 agt_puts "Starting Prolog process ..." 0

 if {![file executable ${ideas_PrologPath}/bp]} {
   agt_puts "failed!\n\nError when trying to start\
             Prolog:\n${ideas_PrologPath}/bp: " 0
   if {![file exists ${ideas_PrologPath}/bp]} {
     agt_puts "Command not found !\nCheck entry in file\
               'init.tcl'.\n"
     return 1
   }
   agt_puts "Permission denied \n"
   return 1
 }

 if {![file exists agt_prologServ.bp]} {
   agt_puts "failed!\n\nError when trying to start\
             Prolog:\nFile 'agt_prologServ.bp' not found!\nSee the\
             README file for compiling 'agt_prologServ.pl' using\
             BinProlog.\n"
   return 1
 }

 set prologFd [open "|${ideas_PrologPath}/bp -q5 agt_prologServ.bp" r+]

 keylset agents(SELF) PROLOG $prologFd

 agt_listenProlog 

 agt_puts "established on $prologFd .\n"

 return 0
}


#
# agt_listenProlog
#
#  Reads the output of a Prolog process.
#
# Arguments:
#
#	outputFlag -	0 quietQuery 
#			1 loudQuery
#			2 listQuery
#
# Returns:
#
#  Depends from query type/outputFlag:
#   0 no output except error: errorcode is set and error returned
#   1 full output, i.e. varaiable values are returned in Prolog manner
#   2 If a query is started by procedure 'agt_listQuery', Prolog
#     will answer it in a special form:
#     The symbol '|' will be sent in front of each possible
#     solution of the query.
#     If this form is detected, 'agt_listenProlog' will build up 
#     and return the response to the query in form of a list.
#     Each list element is a Tcl-KeyedList which keys are the variables
#     of the query and its Prolog values are the values of the keys.
#     If there is no such list is returned by Prolog, 1 or 0 is returned
#     for Prolog yes and no.
#     Errors return allways no.
#
proc agt_listenProlog {{outputFlag 0}} {
global agents

  keylget agents(SELF) PROLOG f
 
  # listFlag set when a "|"-symbol is read in the
  # stream of the Prolog answer.
  # In this case a result list is build up. If not, the
  # Prolog answer is just put in a simple string.
  set listFlag 0  

  set respText ""
  set partList ""

  while {1} {
    # gets the next line of Prolog answer
    gets $f line

    # checks for end of prologs answer
    #
    if {$line == "EOPA"} break

    # check for Prolog error (they start with "*" or ">")
    # when occurred read Prolog answer till end
    # and return with error and set errorcode
    # 
    if {[regexp {^(>|\*)} $line]} {
      fcntl $f NONBLOCK 1
      while {![catch {gets $f errorMesg}]} {
        if {$errorMesg != "EOPA"} { 
          set line "$line\n$errorMesg"
        }
      }
      fcntl $f NONBLOCK 0

      # catch the error 'undefined_predicate'
      # it is stupid when asking for a non existing fact
      # to get an error
      # 
      if {[regexp {^\*\*\* undefined_predicate} $line]} {
	if {$outputFlag == 2} {
          return 0
        }
        return -code ok
      }

      return -code error $line
    }

    # checks for new part list
    # 
    if {$line=="|"} {
      # detected a "|"-symbol, so set the listFlag
      set listFlag 1

      # if this is not the beginning of the result list
      # append the part read so far to the final result
      # variable 'respList'
      if {$partList!=""} {
        lappend respList $partList
      }
      set partList ""
      continue
    } 
    
    # builts up a part of the result list with the line
    # just read, if the "|"-symbol was read, or just a
    # normal String, if not.
    if {$listFlag} {
      # remove the '=' from answer list,
      # e.g. the result could be "| X=1 Y=2 | X=2 Y=4"
      #
      lappend partList [lreplace $line 1 1]
    } else {
      set respText "${respText}$line\n"
    }   
  }

  # now the end of the Prolog answer (EOPA) has been read
  # so append the last part to the final response list
  lappend respList $partList

  # belongig to the output flag the result is returned
  #
  if {$outputFlag == 1} {
    return $respText
  } elseif {$outputFlag == 2} {

    if {!$listFlag} {
      return 0
    } elseif {$respList == "{}"} {
      return 1
    } else {     
     return $respList
    }

  } else {
    return
  }
}


# agt_writeProlog
#
#  Writes a string as an query to the Prolog process and waits for a result.
#
#  Arguments:
#
#    query - the string to send
#    flag  - a flag specifying the form of the expected Prolog result
#            (s. agt_listenProlog)
#
#  Results:
#
#  Return the result of the Prolog query as read by agt_listenProlog or an 
#  error with an error message if one occurrs.
#
proc agt_writeProlog {query flag} {
global agents

 set error 0

 # check if quotes and double-quotes are balanced because tests had 
 # shown that an unbalanced quote makes Prolog hang.
 set quote 0
 set dblquote 0

 for {set index 0} {$index <= [clength $query]} {incr index} {
   set char [cindex $query $index]

   switch -glob -- $char {
    \'  {incr quote    [expr ($quote)    ? -1 : 1]}
    \"  {incr dblquote [expr ($dblquote) ? -1 : 1]}
   } 
 }

if {$quote} {
   set errorMesg "unbalanced quotes in '$query'"
   set error 1
 }

if {$dblquote} {
   set errorMesg "unbalanced double-quotes in '$query'"
   set error 1
 }

 if {$error} {
   return -code error $errorMesg
 }  

  keylget agents(SELF) PROLOG f

  if {[catch {puts $f $query. ; flush $f}]} {

    # Prolog process crashed, try to rebuild it
    #
    if {![agt_catchPrologCrash $f]} {
      agt_puts "Trying the query again...\n"
      return [agt_writeProlog $query $flag]
    } else {
      return -code error "Cannot rebuilt Prolog!"
    }

  } else {
    return [agt_listenProlog $flag]
  }
}


# agt_quietQuery
#
# NOT TO USE BY AGENT PROGRAMMER !
#
# sends a goal to be quietly evaluated
# to the Prolog process,
# i.e. no result is returned, only errors are shown
#
# Arguments:
#
#  query - the query to Prolog
#
# Results:
#
#  Returns an error with an error message, if a Prolog error
#  occurred.
#
proc agt_quietQuery {query} {

 return [agt_writeProlog "call_prolog(($query))" 0]
}


# agt_loudQuery
#
# NOT TO USE BY AGENT PROGRAMMER !
#
# sends a query to the Prolog process and wants back the
# full Prolog answer as a simple string.
# => 'loud'
# agt_loudQuery is used to get a Prolog listing only.
#
# Arguments:
#
#  query - the query to Prolog
#
# Results:
#
#  String including the Prolog answer to the query, or
#  an error with an error message, if a Prolog error
#  occurred.
#
proc agt_loudQuery {query} {

 return [agt_writeProlog "query_prolog(($query))" 1]
}

# agt_listQuery
#
# NOT TO USE BY AGENT PROGRAMMER !
#
#  sends a query to the Prolog process and
#  wants the result back in a special form, 
#  which is parsed by 'agt_listenProlog' and formed
#  to a Tcl list,
#  i.e. returns 1 or 0 for yes/no or a list of Tcl-KeyedLists
#  with variable values as describe in 'agt_listenProlog'.
#
# Arguments:
#
#  query - the query to Prolog
#
# Results:
#
#  Returns -1 and writes a message to logfile, if an error occurs.
#  Returns 0, if answer to query is TRUE.
#  Returns 1, if answer to query is FALSE.
#  Returns a list of variable values, if query includes Prolog 
#  variables Prolog found values for.
#
proc agt_listQuery {query} {

 if {[catch {set result [agt_writeProlog "query_prolog_list(($query))" 2]} \
	     errorMesg]} {
   agt_puts "Prolog-Error when trying to eval '$query':\n$errorMesg\n"
   return -1
 } else {
   return $result
 }
}



#
# agt_query
#
# The 'standard' query to use by agent programmer:
#
#  uses 'agt_listQuery' to send a query to the Prolog process
#  and to get the result back in a special form, 
#  which is parsed by 'agt_listenProlog' and formed
#  to a list of Tcl-KeyedLists,
#  i.e. returns 1 or 0 for yes/no or a Tcl list
#  with variable values.
#  -1 is returned if an error occurs
#
# Arguments:
#
#  query - the query to Prolog
#
# Results:
#
#  Returns -1 and writes a message to logfile, if an error occurs.
#  Returns 0, if answer to query is TRUE.
#  Returns 1, if answer to query is FALSE.
#  Returns a list of variable values, if query includes Prolog 
#  variables Prolog found values for.
#
proc agt_query {query} {

 # set the actual time in Prolog
 agt_timeSet
 agt_displayBelBase

 # do not allow changing of belief-base by simple queries
 #
 switch -regexp -- $query {
   assert[(]|asserta[(]|assertz[(]|retract[(]|consult[(]|compile[(]|abolish[(] {
     agt_puts "Prolog-Error when trying to eval '$query':\nNot allowed to\
               assert, retract, consult ... in a query !\n"
     return -1}
 }

 return [agt_listQuery $query]
 
}

