#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# agt_dispatchMesgs
# 
#  Sends well formatted messages of global agt_SendMesgList 
#  to other agents.
#  If format is not correct, an error is displayed,
#  the message is not sent and an error-message is sent to self.
#
# Arguments:
#
#	sendMesgList	-	messages to send
#
# Results:
#
#  Format of possible error-messages:
#  {:TYPE error} {:CODE 1} {:COMMENT "message format error"}
#  {:TYPE error} {:CODE 3} {:COMMENT "no connection error"}
#  {:TYPE error} {:CODE 4} {:COMMENT "write error"}
#

proc agt_dispatchMesgs {} {
 global agents agt_SendMesgList SELF
 global agt_VarComFilter

 keylget agents(SELF) NAME SELF

 keylset error_mesg :TYPE error \
                    :SENDER $SELF \
                    :RECEIVER $SELF 

 set error 0

 set cp_SendMesgList $agt_SendMesgList

 set agt_SendMesgList ""

 foreach mesg $cp_SendMesgList {
 
   keylset error_mesg :REPLY-TO [keylget mesg :REPLY-WITH]

   # fill in return sender of message
   #
   if {[keylget mesg :SENDER {}] == 0} {
     keylset mesg :SENDER [keylget agents(SELF) NAME]
   }

   # replace SELF in list of receiver with agents own
   # name as saved in $SELF
   #
   keylget mesg :RECEIVER receiverList

   regsub SELF $receiverList $SELF receiverList

   keylset mesg :RECEIVER $receiverList

   # check format of message
   #
   if {[catch {set text [agt_formatMesg $mesg]} errorMesg]} {
     agt_puts "FORMAT-ERROR when trying to send:\n$mesg\nFailure:\
               '$errorMesg'\n"
     keylset error_mesg :CODE 1 \
                        :COMMENT "message format error"
     agt_send $error_mesg
     set error 1
     break
   }

   if {[catch {keylget mesg :REPLY-TO}]} {
     keylset mesg :REPLY-TO none
   }
    
   foreach agtName $receiverList {

     # if receiver is agents own name, replace it by SELF
     # to find the correct fid in next step
     #
     if {[cequal $agtName $SELF]} {
       set agtName SELF
     }

     # check if agents is connected to the receiver
     # and get the fid for outgoing messages
     #
     if {([catch {keylget agents($agtName)}]) || \
         (![keylget agents($agtName) OUTFID agtFid])} {
       agt_puts "NO-CONNECTION-ERROR when trying to send\
	       to '$agtName':\n${mesg}Message not sent!\n"
       keylset error_mesg :CODE 3 \
                          :COMMENT "no connection error"
       agt_send $error_mesg
       set error 1
       break
     }

     if {[catch {puts $agtFid $mesg; flush $agtFid }]} {
      # cannot write to pipe because its broken
      # so remove the information about the pipe
      #
      agt_puts "WRITE ERROR when trying to send\
                to '$agtName':\n${mesg}Message not sent!\n"

      agt_conBroken $agtFid $agtName

      if {![cequal $agtName $SELF]} {
        keylset error_mesg :CODE 4 \
                           :COMMENT "write error"

        agt_send $error_mesg
        set error 1

      }
      break
     }

     agt_puts "Message sent to '$agtName':\n$text" 

     if {[lsearch $agt_VarComFilter $agtName] != -1} {
       agt_logComm "Message sent to '$agtName':\n$text"
     }
   }
 }

 # if an error occured send the error-messages to self
 #
 if {$error} {
   agt_dispatchMesgs
 }

} 


#
# agt_send
#
# Appends a message to the global list of messages to be
# sent at the end of the cycle.
#
# Arguments:
#
# 	mesg	-	mesg to append to agt_SendMesgList
#
# Results:
#
#  Global list agt_SendMesgList includes mesg,
#  returns number of the mesg (in :REPLY-WITH) or 
#  -1 , if one of the keys :TYPE or :RECEIVER is missing in mesg 
#       or format error in keyedList
#
proc agt_send {mesg} {
global agt_SendMesgList agt_MesgCounter


 if {![keylget mesg :RECEIVER {}]} {
   agt_puts "agt_send - FORMAT ERROR:\n:RECEIVER specification missing in\
             '$mesg'\nMessage not appended to send.\n"
   return -1
 }

 if {![keylget mesg :TYPE {}]} {
   agt_puts "agt_send - FORMAT ERROR:\n:TYPE specification missing in\
             '$mesg'\nMessage not appended to send.\n"
   return -1
 } 

 if {[catch {keylset mesg :REPLY-WITH $agt_MesgCounter} errorMesg]} {
   agt_puts "agt_send - FORMAT ERROR:\n$errorMesg in\
             '$mesg'\nMessage not appended to send.\n"
   return -1
 } 
   

 lappend agt_SendMesgList $mesg

 incr agt_MesgCounter

 agt_puts "Append to send: $mesg\n"

 return [expr $agt_MesgCounter-1]
}


