#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# uam_getActSrcFile
#
#  Gets the Action-C-Source file specification out of the ADL type file
#  of an agent.
#  Called when starting a new agent from 'uam_startAgent'.
#
#  Arguments:
#
#	agtType - name of type file (in directory 'desc')
#
#  Results:
#
#    Name of the Action-C-Source file for the new agent or an error
#    with error message if an ADL-error occurrs in type file.
#
proc uam_getActSrcFile {agtType} {

 # try to open 'agtType' for reading
 set fid [open desc/$agtType]

 # read file and split it in lines
 set file [split [read $fid] \n]

 close $fid

 set error 0

 set actSrcFile ""

 # start parsing the type file for the Agent-C-Specification:

 # pop the first line from file and remove leading and
 # succeeding white spaces
 set line [string trim [lvarpop file]]

 while {1} {
     switch -regexp $line {
      ^$|^# {if {[llength $file] == 0} {
               return -code ok
             }
             set line [string trim [lvarpop file]]
            }

      ^load {set fileName [string trim [csubstr $line 4 end]]

             if {[catch {set fid [open desc/$fileName]} errorMesg]} {
               uam_logText "Error when loading '$fileName': $errorMesg"
               return
             }

             uam_logText "Loaded: $fileName\n"

             set data [split [read $fid] \n]
             close $fid

             set file [concat $data $file]
             set line [string trim [lvarpop file]]
            }

      ^ACTION-C-SOURCE {
             set line [string trim [crange $line 15 end]]
	     # initialize a counter for \{\}--brackets
             #
             set brackets 0
             set chCt -1

	     while {$brackets == 0} {
               incr chCt
               set char [cindex $line $chCt]

               switch -glob -- $char {
		 \{  {incr brackets 1}

                 \}  {incr brackets -1}

                 "" {
                     # if last character in line is "\" remove it before
                     # appending the line to actSrcFile and poping next line 
                     # from file;
                     #
                     if {[cequal "\\" [cindex $line $chCt-1]]} {
                       if {$chCt >= 2} {
                         append actSrcFile [crange $line 0 $chCt-2] 
                       }
                       append actSrcFile " "
                     }

                     if {[cequal $file ""]} {break}

                     set line [string trim [lvarpop file]]
                     set chCt -1
                    }
                  
		 default {return -code error "'\{' expected at '$line'."}
	       }
	     }

             if {$brackets == -1} {
               return -code error "'\{' expected at '$line'."
             }

             set line [string trimleft [csubstr $line 1 end]]
             set chCt -1

             while {$brackets == 1} {
               incr chCt
               set char [cindex $line $chCt]

               switch -glob -- $char {
		 \{  {incr brackets 1}

                 \}  {incr brackets -1}

                 "" {
                     # if last character in line is "\" remove it before
                     # appending the line to actSrcFile and poping next line 
                     # from file;
                     #
                     if {[cequal "\\" [cindex $line $chCt-1]]} {
                       if {$chCt >= 2} {
                         append actSrcFile [crange $line 0 $chCt-2] 
                       }
                       append actSrcFile " "
                     }

                     if {[cequal $file ""]} {break}

                     set line [string trim [lvarpop file]]
                     set chCt -1
                    }
    
                 #  {set line [string trim [lvarpop file]]
                     set chCt -1}
	       }
	     }

             if {$chCt} {
               append actSrcFile [crange $line 0 $chCt-1]
             } 
    
             if {$brackets} {
	       return -code error "'\}' expected at '$line'."
             }

             break
            }
 
      default {break}
    }
 }

 # if no Action-C-Source is defined use the standard source:
 #
 if {$actSrcFile == ""} {
   set actSrcFile standard.c
   uam_logText "No ACTION-C-SOURCE definition found in '$agtType';\
                using 'standard.c'\n"
 }

 return [string trimright $actSrcFile]
}


