#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# uam_registerAgt
#
#  When an agent has initialized or changed status it sends a statusmesg
#  to his UAM.
#  There is a UAM MesgRule to catch this mesg, which calls this proc 
#  to register the agents status to UAM.
#
# Arguments:
#
#   fid     - filedesc on which the mesg arrived
#   agtIdfk - keyedList that includes the agents status 
#
# Results: None.
# 
proc uam_registerAgt {fid agtIdfk} {
 global agents uam_VarSystem

 # get name out of agents identifikation - includes a format check of keyedList
 #
 if {[catch {set result [keylget agtIdfk NAME agtName]}]} {
   uam_logText "WARNING - received wrong formatted\
                status-message.\nMessage ignored!\n"
   return
 }

 if {$result == 0} {
   uam_logText "WARNING - received wrong formatted status-message:\nKey\
                'NAME' is missing in identifikation.\nMessage ignored!\n"
   return
 }


 if {[info exists agents($agtName)]} {

   # the UAM knows this agent already, so this is a status change
   #
   if {![cequal [keylget agents($agtName) FID] $fid]} {
     uam_logText "WARNING - received status-message on foreign FID:\
                  $fid\nMessage ignored!\n"
     return
   }

   if {[keylget agtIdfk STATUS agtStatus] == 0} {
     uam_logText "WARNING - received wrong formatted status-message:\nKey\
                  'STATUS' is missing in identifikation.\nMessage ignored!\n"
     return
   }

   # save the new status and actualize the listbox entry for this agent
   #
   keylset agents($agtName) STATUS $agtStatus

   # search the listbox entry

   set size [expr [.frame1.list size]-1]

   for {set i 0} {$i <= $size} {incr i} {
     set listItem [.frame1.list get $i]

     if {[string match $agtName* $listItem]} {
       set listItem [lreplace $listItem 4 4 $agtStatus]
       .frame1.list insert $i $listItem
       .frame1.list delete [expr $i + 1]

       # select the entry after all
       .frame1.list selection set $i

       return
     }
   }

   return
 }
       
 # the agent is unknown to UAM till now
 # but then there is an entry in array 'agents' for this name without
 # '.portno' because the agent has been inited by this UAM.
 # So remove this old entry and register the new one.
 #
 set oldAgtName [crange $agtName 0 [string last . $agtName]-1]

 keylget agents($oldAgtName) FID agtFid

 if {![cequal $agtFid $fid]} {
   uam_logText "WARNING - received message on foreign FID:\
                $fid\nMessage ignored!\n"
   return
 }

 # Re-set communication control on pipe to/from agent to full name
 #
 fileevent $agtFid readable "uam_commCtrl $agtFid $agtName"

 set agents($agtName) $agents($oldAgtName)
 unset agents($oldAgtName)

 keylset agents($agtName) NAME $agtName

 keylget agents($agtName) TYPE typeFile

 # insert new agent in agentslist
 #
 .frame1.list insert end "$agtName / $typeFile / deact"

  if {[.frame1.list size] == 1} {
   .frame1.list selection set 0
  }

 # let the activate, delete and mesg-to button appear normal
 #
 .frame2.actB  configure -state normal
 .frame2.delB  configure -state normal
 .frame2.mesgB configure -state normal

 bind . <Alt-Key-a> {.frame2.actB flash;.frame2.actB invoke}
 bind . <Alt-Key-d> {.frame2.delB flash;.frame2.delB invoke}
 bind . <Alt-Key-m> {.frame2.mesgB flash;.frame2.mesgB invoke}

 uam_logText "Registered: $agents($agtName)\n"


 # if the agent is generated in an auto-built MAS remove it from
 # global array uam_VarSystem to remember that it has been initialized

 set oldAgtName [crange $agtName 0 [string first . $agtName]-1]

 if {[info exists uam_VarSystem($oldAgtName)]} {
   unset uam_VarSystem($oldAgtName)
 }

 # if all agents which were auto-built are initialized, we can send
 # handshakes to set the communication paths of the MAS. 
 # The information about the path is stored in uam_VarSystem(CON).
 # (see uam_startMAS for details)
 #
 if {[array size uam_VarSystem] == 1} {

   if ![lempty $uam_VarSystem(CON)] {
     uam_logText "Trying to connect the MAS now...\n"
   }

   # get all the existing local agents
   #
   set existAgts [array names agents]

   foreach con $uam_VarSystem(CON) {
     set addrList ""


     # get the receiver of the handshake and search for his real name,
     # i.e. his specified name with host and portno
     #
     set agtName [lindex $con 0]
     set receiver [lindex $existAgts [lsearch $existAgts $agtName.*]]

     if [lempty $receiver] {
       uam_logText "MAS-connect-Error: Unknown local agent '$agtName'!\n"
       uam_conMASError $agtName
       continue
     }

     # do the same for the agents to connect to
     #
     foreach agtName [lreplace $con 0 0] {
       # get the name of a connection partner
       #
       set conPartner [lindex $existAgts [lsearch $existAgts $agtName.*]]

       if {[lempty $conPartner]} {
         uam_logText "MAS-connect-Error: Unknown local agent\
                      '$agtName'!\n\Cannot connect '$receiver' with\
                      '$agtName'.\n"
         uam_conMASError $agtName
       } else {
         # extract the address from name and append it to connect list
         #
         lappend addrList \
           [crange $conPartner [string first . $conPartner]+1 end]
       }
     }     

     if {![lempty $addrList]} {
       uam_logText "Auto-Connect '$receiver' with '$addrList':\n"

       keylset mesg :RECEIVER $receiver \
                    :TYPE command \
                    :COMMAND handshake \
                    :ADDRESS $addrList

       uam_sendMesg $mesg 
     }
   }

   if ![lempty $uam_VarSystem(CON)] {
     uam_logText "...auto-connecting finished.\n"
   }

   # after initializing the connections remove the uam_VarSystem(CON) entry
   #
   set uam_VarSystem(CON) ""

   # special hack to reset the cursor in mbar after compilation
   # in some special cases... 
  .mbar  configure -cursor {}
   
 }

}

