#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# uam_startMAS
#
#  Reads a Multi-Agent-System description from a file, starts the
#  described agents and provides later sets of communication paths.
#  Called by 'uam_newAgt'.
#
# Arguments:
#
#	systemFile - name of the MAS-description file
#
# Results:
#
#  Side effect: sets global array uam_VarSystem to recognize the agents
#               of the MAS later, when they are initialized 
#
proc uam_startMAS {systemFile} {
global agents uam_VarSystem

 uam_logText "Starting Multi-Agent-System '$systemFile'...\n"

 # try to open the desc file
 #
 if {[catch {set fid [open systems/$systemFile r]} errorMesg]} {
   uam_logText "MAS-Error: $errorMesg\n\nStart of '$systemFile' failed!\n"
   uam_startMASError $systemFile
   return
 }

 # read the desc and divide it in lines
 #
 if {[catch {set desc [split [read $fid] "\n"]} errorMesg]} {
   uam_logText "MAS-Error: $errorMesg\n\nStart of '$systemFile' failed!\n"
   uam_startMASError $systemFile
   return
 }

 close $fid

 # initialize local_VarSystem(CON); it will include a list of lists the
 # described connections
 #
 set local_VarSystem(CON) ""

 # initialize variable previous; it is used to save the content of a line
 # if there is a '\'-character at the end
 #
 set previous ""

 # parse the file
 #
 foreach line $desc {

   # look for a '\'-char at the end of line, so the next line belongs
   # to this one; save this one in previous
   #
   if {[cindex $line end] == "\\"} {
     append previous "[crange $line 0 end-1] "
     continue
   }

   # add previous lines to this one, if there were '\' at the end
   #
   set line ${previous}${line}
   set previous ""

   # get the keyWord of the line: i.e. the first word
   #
   set keyWord [string tolower [lindex $line 0]]

   # ignore if its a remark or line is empty
   #
   if {[regexp ^$|# [cindex $keyWord 0]]} {
     continue
   }

   # check if the line is a agent specification
   #
   if {[cequal $keyWord agent:]} {

     # the it must have 3 elements: 'agent: name type'
     #
     if {[llength $line] != 3} {
       uam_logText "MAS-Syntax-Error in file '$systemFile':\nwrong # args:\
          should be \"Agent: agtName typeName\"\n\nStart of '$systemFile'\
          failed!\n"
       uam_startMASError $systemFile
       return
     }
       
     # get the specified name
     #
     set agtName [string tolower [lindex $line 1]]

     # allow only the beginning part of agtName string without "."
     #
     catch {set agtName [crange $agtName 0 [string first . $agtName]-1]} 

     # test if newName is already existing in real agents or in the
     # list of specified names before;
     # if it exists add '_no' to the name, where 'no' is a number not
     # used at this position untill now (1, 2, ...)
     #
     set existAgts [concat [array names agents] [array names local_VarSystem]]

     set renamed 0
     set old_agtName $agtName

     for {set i 1} \
         {[lsearch -regexp $existAgts ^${agtName}.*$|^${agtName}$] != -1} \
         {incr i} {

       catch {set agtName [crange $agtName 0\
	       [string first _[expr $i-1] $agtName]-1]}
       set agtName ${agtName}_$i
       set renamed 1
     }

     # save the renaming in a array alias, so we can remember when
     # checking for the connections to set
     #
     if {$renamed} {
       set alias($old_agtName) $agtName
     }

     # get the name of the agents type
     #
     set typeFile [string tolower [lindex $line 2]]

     # possibly we have to add the correct extension '.agt'
     #
     if {![cequal [file extension $typeFile] .agt]} {
       append typeFile .agt
     }

     # test if selected filename exists
     #
     if {[lsearch [glob desc/*.agt] desc/$typeFile]==-1} {
       uam_logText "MAS-Error in file '$systemFile':\nType '$typeFile' does\
                    not exists in directory 'systems'.\n\nStart of \
                    '$systemFile' failed!\n"
       uam_startMASError $systemFile
       return
     }

     # save the correct parsed agent name and type in the local array
     #
     set local_VarSystem($agtName) $typeFile

    continue
   }

   # test if the line includes a connection description
   #
   if {[cequal $keyWord con:]} {
     # then it must have 3 elements minimum: 'con: recName agtName ...'
     #
     if {[llength $line] < 3} {
       uam_logText "MAS-Syntax-Error in file '$systemFile':\nwrong # args:\
                    should be \"Con: agtName agtName ?agtName ...?\"\n\nStart\
                    of '$systemFile' failed!\n"
      uam_startMASError $systemFile
      return
     }


     set conList ""

     # get the list of the specified agents in con: and replace names
     # if they were renamed, i.e. saved in alias array
     #
     foreach agtName [lreplace $line 0 0] {
       set agtName [string tolower $agtName]
       if {[info exists alias($agtName)]} {
         lappend conList $alias($agtName)
       } else {
         lappend conList $agtName
       }
     }

     # append the list as an element to local array in field CON
     #
     lappend local_VarSystem(CON) $conList
     continue
   }

   # in this case there is a keyWord not agent: or cont:, so error
   #   
   uam_logText "MAS-Syntax-Error in file '$systemFile':\nunknown command\
                'keyWord': should be 'Agent:' or 'Con:'\n\nStart of \
                '$systemFile' failed!\n"
   uam_startMASError $systemFile
   return
   
 }

 # at this point parsing is done 

 # if there is after end of file parsing a unevaled rest in variable
 # previous it is an error
 #
 if {$previous != ""} {
   uam_logText "MAS-Syntax-Error in file '$systemFile':\nunexpected end\
                of file\n\nStart of '$systemFile' failed!\n"
   uam_startMASError $systemFile
   return
 }  

 # copy the information from the local array in the global uam_VarSystem;
 # show a remark of renamings in logfile and start the specified agents
 #
 foreach name [array names local_VarSystem] {
   set uam_VarSystem($name) $local_VarSystem($name)

   if {[info exists alias($name)]} {
     uam_logText "MAS-Information:\nThere is existing an agent with name\
                  '$alias($name)'\nCreating agent '$name' instead.\n"
   }

   if {![cequal $name CON]} {
     uam_startAgent $name $uam_VarSystem($name)
   }
 }
 
 uam_logText "...all agents of MAS started.\n"

}

